{  *---------------------------------------------------*
   |                  Gr3D.tpu                         |
   *---------------------------------------------------*
   |                                                   |
   | Author     : MFSomers, 1993.                      |
   | Updates    : 10th. (15 Feb 2000)                  |
   | Use        : To make 3D objects in a system using |
   |               OOP. TVISION compatible ...         |
   |                                                   |
   *---------------------------------------------------*
}

{$O-,B-,F+,A+,I-,N+,D-,P-,Q-,R-,S-,T+,V+,G-,E+,X+}

Unit Gr3D;                              { 3D graphics unit }

Interface

Uses Graph,
     Crt,
     Objects;

Const
     Arrow_Left  = #$4B;                { Handy key defines for routines }
     Arrow_Right = #$4D;
     Arrow_Up    = #$48;
     Arrow_Down  = #$50;
     Alt_P       = #25;
     Alt_X       = #45;
     Alt_Z       = #44;
     Alt_A       = #30;
     Alt_S       = #31;
     Ctrl_Left   = #115;
     Ctrl_Right  = #116;

     NoItemID    = 0;

     nTGrObjectType      = 2100;       { Type index of every object }
     nTGrPointType       = 2110;
     nTGrLineType        = 2120;
     nTGrSystemType      = 2130;
     nTGrGroupType       = 2140;
     nTGrOpenWireType    = 2150;
     nTGrClosedWireType  = 2160;
     nTGrOpenWireFrameType   = 2170;
     nTGrClosedWireFrameType = 2180;

Type
    TGrMatrix = Array[ 1 .. 4, 1 .. 4 ] Of Real;
    TGrStatus = Record
                      Visible,
                      Perspective,
                      LabelAxes,
                      ShowTheAxes : Boolean;
                      Phi,
                      Theta,
                      Rho,
                      RelX,
                      RelY    : Integer;
                      ZoomF,
                      PerspF  : Real;
                      XLabel,
                      YLabel,
                      ZLabel  : String[ 10 ];
                End; { Record }
    PGrStatus  = ^TGrStatus;

    TGrLocation = Record
                        X,
                        Y,
                        Z   : Integer;
                        U,
                        V,
                        W   : Word;
                   End; { Record }
    PGrLocation = ^TGrLocation;

    PGrObject = ^TGrObject;
    TGrObject = Object( TObject )
                      ItemID     : Word;
                      POwner     : PGrObject;

                      Constructor Init;
                      Destructor  Done; Virtual;
                      Procedure   Convert; Virtual;
                      Procedure   Clear; Virtual;
                      Function    GetStatus( Var Status2 : TGrStatus ) : Boolean; Virtual;
                      Function    GetMatrix( Var M : TGrMatrix ) : Boolean; Virtual;
                      Procedure   Draw; Virtual;
                      Procedure   Store( Var S : TStream ); Virtual;
                      Constructor Load( Var S : TStream );
                End; { Object }

    TGrPoint = Object( TGrObject )
                     _Point : TGrLocation;
                     _Color : Word;

                     Constructor Init( X, Y, Z : Integer; Color  : Word );
                     Procedure   Clear; Virtual;
                     Procedure   Draw; Virtual;
                     Procedure   Store( Var S : TStream ); Virtual;
                     Constructor Load( Var S : TStream );
                     Procedure   Convert; Virtual;
               End; { Object }
    PGrPoint = ^TGrPoint;

    TGrLine  = Object( TGrObject )
                     _Point      : Array[ 1 .. 2 ] Of TGrLocation;
                     _Color      : Word;

                     Constructor Init( X, Y, Z, X2, Y2, Z2 : Integer; Color : Word );
                     Procedure   Draw; Virtual;
                     Procedure   Store( Var S : TStream ); Virtual;
                     Constructor Load( Var S : TStream );
                     Procedure   Clear; Virtual;
                     Procedure   Convert; Virtual;
               End; { Object }
    PGrLine  = ^TGrLine;

    TGrGroup = Object( TGrObject )
                      PList       : PCollection;

                      Constructor Init;
                      Destructor  Done; Virtual;
                      Procedure   Store( Var S : TStream ); Virtual;
                      Constructor Load( Var S : TStream );
                      Procedure   Draw; Virtual;
                      Procedure   DrawSelf; Virtual;
                      Procedure   Clear; Virtual;
                      Procedure   ClearSelf; Virtual;
                      Procedure   Convert; Virtual;
                      Procedure   ConvertSelf; Virtual;
                      Procedure   Insert( Id : Word; Link : PGrObject );
                      Procedure   Delete( Id : Word ); Virtual;
                      Procedure   MemError; Virtual;
                End; { Object }
    PGrGroup = ^TGrGroup;

    TGrSystem = Object( TGrGroup )
                      _Matrix    : TGrMatrix;
                      _Axes      : Array [ 1 .. 3 ] Of TGrLocation;
                      _Origin    : TGrLocation;
                      _Color     : Word;
                      _Status    : TGrStatus;

                      Constructor Init( X, Y, Z : Integer; Color : Word );
                      Procedure   Store( Var S : TStream ); Virtual;
                      Constructor Load( Var S : TStream );
                      Procedure   CalculateMatrix; Virtual;
                      Procedure   DrawSelf; Virtual;
                      Procedure   ClearSelf; Virtual;
                      Procedure   Convert; Virtual;
                      Procedure   ConvertSelf; Virtual;
                      Procedure   Insert( Id : Word; Link : PGrObject );
                      Function    GetMatrix( Var M : TGrMatrix ) : Boolean; Virtual;
                      Procedure   SetStatus( Var Status2 : TGrStatus ); Virtual;
                      Function    GetStatus( Var Status2 : TGrStatus ) : Boolean; Virtual;
                      Procedure   KeybRotate( Angle : Integer ); Virtual;
                      Procedure   DemoRotate( Angle : Integer; Wait : Word ); Virtual;
                End; { Object }
    PGrSystem = ^TGrSystem;

    TGrOpenWire = Object( TGrGroup )      { A collection of points connected with lines }
                       _Color : Word;

                       Constructor Init( Color : Word );
                       Procedure   DrawSelf; Virtual;
                       Procedure   ClearSelf; Virtual;
                       Procedure   Insert( Id : Word; Link : PGrPoint );
                       Procedure   Store( Var S : TStream ); Virtual;
                       Constructor Load( Var S : TStream );
                  End; { Object }
     PGrOpenWire = ^TGrOpenWire;

     TGrClosedWire = Object( TGrOpenWire )      { A collection of points connected with lines }
                       Procedure   DrawSelf; Virtual;
                       Procedure   ClearSelf; Virtual;
                     End; { Object }
     PGrClosedWire = ^TGrClosedWire;

     TGrOpenWireFrame = Object( TGrOpenWire )   { A collection of wires connected with lines }
                            Procedure   Insert( Id : Word; Link : PGrOpenWire );
                            Procedure   DrawSelf; Virtual;
                            Procedure   ClearSelf; Virtual;
                        End; { Object }
     PGrOpenWireFrame = ^TGrOpenWireFrame;

     TGrClosedWireFrame = Object( TGrOpenWireFrame )
                                Procedure   DrawSelf; Virtual;
                                Procedure   ClearSelf; Virtual;
                          End; { Object }
     PGrClosedWireFrame = ^TGrClosedWireFrame;

Var
   _Gr3D_AuthorMessage : Boolean;

{ ------------------------------------------------- All Functions Etc. - }

 Function GrCos( Phi : Integer ) : Real;
 Function GrSin( Phi : Integer ) : Real;
 Function GrTan( Phi : Integer ) : Real;

 Procedure Transform( Var F : TGrMatrix; Var Point : TGrLocation; Var Status : TGrStatus );
 Procedure GrSetAspectFactor;
 Procedure Register3DSystem;

{ ----------------------------- PRIVATE ---------------------------- }

Implementation

Const
     RGrObject : TStreamRec = (
                               ObjType: nTGrObjectType;
                               VmtLink: Ofs( TypeOf( TGrObject )^ );
                               Load:    @TGrObject.Load;
                               Store:   @TGrObject.Store
                              );
     RGrLine : TStreamRec = (
                              ObjType: nTGrLineType;
                              VmtLink: Ofs( TypeOf( TGrLine )^ );
                              Load:    @TGrLine.Load;
                              Store:   @TGrLine.Store
                             );
     RGrPoint : TStreamRec = (
                              ObjType: nTGrPointType;
                              VmtLink: Ofs( TypeOf( TGrPoint )^ );
                              Load:    @TGrPoint.Load;
                              Store:   @TGrPoint.Store
                             );
     RGrGroup : TStreamRec = (
                              ObjType: nTGrGroupType;
                              VmtLink: Ofs( TypeOf( TGrGroup )^ );
                              Load:    @TGrGroup.Load;
                              Store:   @TGrGroup.Store
                             );
     RGrSystem : TStreamRec = (
                               ObjType: nTGrSystemType;
                               VmtLink: Ofs( TypeOf( TGrSystem )^ );
                               Load:    @TGrSystem.Load;
                               Store:   @TGrSystem.Store
                              );
     RGrOpenWire : TStreamRec = (
                                 ObjType: nTGrOpenWireType;
                                 VmtLink: Ofs( TypeOf( TGrOpenWire )^ );
                                 Load:    @TGrOpenWire.Load;
                                 Store:   @TGrOpenWire.Store
                                );
     RGrClosedWire : TStreamRec = (
                                   ObjType: nTGrClosedWireType;
                                   VmtLink: Ofs( TypeOf( TGrClosedWire )^ );
                                   Load:    @TGrClosedWire.Load;
                                   Store:   @TGrClosedWire.Store
                                  );
     RGrOpenWireFrame : TStreamRec = (
                                      ObjType: nTGrOpenWireFrameType;
                                      VmtLink: Ofs( TypeOf( TGrOpenWireFrame )^ );
                                      Load:    @TGrOpenWireFrame.Load;
                                      Store:   @TGrOpenWireFrame.Store
                                     );
     RGrClosedWireFrame : TStreamRec = (
                                        ObjType: nTGrClosedWireFrameType;
                                        VmtLink: Ofs( TypeOf( TGrClosedWireFrame )^ );
                                        Load:    @TGrClosedWireFrame.Load;
                                        Store:   @TGrClosedWireFrame.Store
                                       );

Var
   OldExitProc  : Pointer;
   AspectFactor : Real;

{ ---------------------------------------------- A Faster Cosinus - }

Function GRCos( Phi : Integer ) : Real;
Const
     Tabel : Array[ 0 .. 90 ] Of Real = (  10.000000000E-01, 9.9984769515E-01, 9.9939082702E-01,
                                           9.9862953475E-01, 9.9756405026E-01, 9.9619469809E-01,
                                           9.9452189537E-01, 9.9254615164E-01, 9.9026806874E-01,
                                           9.8768834059E-01, 9.8480775301E-01, 9.8162718345E-01,
                                           9.7814760073E-01, 9.7437006479E-01, 9.7029572628E-01,
                                           9.6592582629E-01, 9.6126169594E-01, 9.5630475596E-01,
                                           9.5105651630E-01, 9.4551857560E-01, 9.3969262078E-01,
                                           9.3358042650E-01, 9.2718385457E-01, 9.2050485345E-01,
                                           9.1354545764E-01, 9.0630778703E-01, 8.9879404630E-01,
                                           8.9100652419E-01, 8.8294759286E-01, 8.7461970714E-01,
                                           8.6602540378E-01, 8.5716730070E-01, 8.4804809616E-01,
                                           8.3867056794E-01, 8.2903757255E-01, 8.1915204429E-01,
                                           8.0901699437E-01, 7.9863551004E-01, 7.8801075360E-01,
                                           7.7714596146E-01, 7.6604444312E-01, 7.5470958022E-01,
                                           7.4314482548E-01, 7.3135370162E-01, 7.1933980034E-01,
                                           7.0710678118E-01, 6.9465837046E-01, 6.8199836006E-01,
                                           6.6913060636E-01, 6.5605902899E-01, 6.4278760968E-01,
                                           6.2932039105E-01, 6.1566147532E-01, 6.0181502315E-01,
                                           5.8778525229E-01, 5.7357643635E-01, 5.5919290347E-01,
                                           5.4463903501E-01, 5.2991926423E-01, 5.1503807491E-01,
                                           5.0000000000E-01, 4.8480962024E-01, 4.6947156278E-01,
                                           4.5399049973E-01, 4.3837114679E-01, 4.2261826174E-01,
                                           4.0673664307E-01, 3.9073112849E-01, 3.7460659341E-01,
                                           3.5836794954E-01, 3.4202014332E-01, 3.2556815445E-01,
                                           3.0901699437E-01, 2.9237170472E-01, 2.7563735581E-01,
                                           2.5881904510E-01, 2.4192189560E-01, 2.2495105434E-01,
                                           2.0791169081E-01, 1.9080899537E-01, 1.7364817766E-01,
                                           1.5643446503E-01, 1.3917310096E-01, 1.2186934340E-01,
                                           1.0452846326E-01, 8.7155742741E-02, 6.9756473738E-02,
                                           5.2335956238E-02, 3.4899496696E-02, 1.7452406431E-02,
                                           0.0000000000E-00 );

 Begin
      Phi := Phi Mod 360;
      Phi := Abs( Phi );

      If( Phi > 180 ) Then
       Phi := 360 - Phi;

      If( Phi > 90 ) Then
       GrCos := -( Tabel[ 180 - Phi ] )
      Else
       GrCos := Tabel[ Phi ];
 End; { Function }

{ --------------------------------------------------------- A Faster Sinus - }

Function GrSin( Phi : Integer ) : Real;
 Begin
       GrSin := GrCos( 90 - Phi );
 End; { Function }

{ ---------------------------------------------------------- A Faster Tanges - }

Function GrTan( Phi : Integer ) : Real;
 Begin
      GrTan := ( GrSin( Phi ) / GrCos( Phi ) );
 End; { Function }

{ ------------------------------------------------------------------ }

Procedure GrSetAspectFactor;
Var
   X, Y : Word;
 Begin
       GetAspectRatio( X, Y );
       AspectFactor := ( X / Y );
 End; { Procedure }

{ ------------------------------------------------------------------ }

Procedure Register3DSystem;
 Begin
      RegisterType( RGrObject );         { Register all objects }
      RegisterType( RGrPoint );
      RegisterType( RGrLine );
      RegisterType( RGrSystem );
      RegisterType( RGrGroup );
      RegisterType( RGrOpenWire );
      RegisterType( RGrClosedWire );
      RegisterType( RGrOpenWireFrame );
      RegisterType( RGrClosedWireFrame );
 End; { Procedure }

{ ------------------------------------------------------------------ }
                   { Transform Point acording to matrix F and status }

Procedure Transform( Var F : TGrMatrix; Var Point : TGrLocation; Var Status : TGrStatus );
Var WR, UR, VR : Real;
 Begin
      With Point Do
       Begin
             UR := ( X*F[ 1, 1 ] +  Y*F[ 1, 2 ] );
             VR := ( ( X*F[ 2, 1 ] + Y*F[ 2, 2 ] + Z*F[ 2, 3 ] ) );

             If( Status.Perspective ) Then
              Begin
                   WR := ( X*F[ 3, 1 ] + Y*F[ 3, 2 ] + Z*F[ 3, 3 ] + F[ 3, 4 ] );

                   If( WR >= 0 ) Then
                    Begin
                         UR := UR / ( WR * Status.PerspF + 1 );
                         VR := VR / ( WR * Status.PerspF + 1 );
                    End
                   Else
                    Begin
                         UR := -UR * ( WR * Status.PerspF - 1 );
                         VR := -VR * ( WR * Status.PerspF - 1 );
                    End; { If If Then Else }

                   W := Round( WR );
              End; { If then }

             U := Round( Status.ZoomF * UR + Status.RelX );
             V := Round( -1 * AspectFactor * VR * Status.ZoomF + Status.RelY );
        End; { With Do }
 End; { Procedure }

{ ------------------------------------------------------------------ }

Procedure ExitRoutine; Far;
 Begin
      If( _Gr3D_AuthorMessage ) Then
       Writeln( '3D OOP routines by: MFSomers, 15-02-2000, Holland ...' );
      ExitProc := OldExitProc;
 End; { Procedure }

{ -------------------------- TGrObject ---------------------------- }

Constructor TGrObject.Init;
 Begin
      ItemID := NoItemID;
      POwner := Nil;
 End; { Constructor }

{ ----------------------------------------------------------------- }

Destructor TGrObject.Done;
Var TheStatus : TGrStatus;
 Begin
      If( GetStatus( TheStatus ) ) Then
       If( TheStatus.Visible ) Then
        Clear;
 End; { Destructor }

{ ----------------------------------------------------------------- }

Procedure TGrObject.Convert;
 Begin
 End; { Procedure }

{ ------------------------------------------------------------------ }

Function TGrObject.GetStatus( Var Status2 : TGrStatus ) : Boolean;
 Begin
      If( POwner <> Nil ) Then
       GetStatus := POwner^.GetStatus( Status2 )
      Else
       GetStatus := False;
 End; { Procedure }

{ ------------------------------------------------------------------ }

Function TGrObject.GetMatrix( Var M : TGrMatrix ) : Boolean;
 Begin
      If( POwner <> Nil ) Then
       GetMatrix := POwner^.GetMatrix( M )
      Else
       GetMatrix := False;
 End; { Function }

{ ------------------------------------------------------------------ }

Procedure TGrObject.Clear;
 Begin
 End; { Procedure }

{ ------------------------------------------------------------------ }

Procedure TGrObject.Draw;
 Begin
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrObject.Store( Var S : TStream );
 Begin
      S.Write( ItemID, SizeOf( ItemID ) );
 End; { Procedure }

{ ----------------------------------------------------------------- }

Constructor TGrObject.Load( Var S : TStream );
 Begin
      S.Read( ItemID, SizeOf( ItemID ) );
 End; { Constructor }

{ ----------------------------- TGrPoint --------------------------- }

Constructor TGrPoint.Init( X, Y, Z : Integer; Color : Word );
 Begin
      TGrObject.Init;
      _Point.X := X;
      _Point.Y := Y;
      _Point.Z := Z;
      _Color := Color;
 End; { Constructor }

{ ----------------------------------------------------------------- }

Procedure TGrPoint.Draw;
 Begin
      PutPixel( _Point.U, _Point.V, _Color );
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrPoint.Clear;
 Begin
      PutPixel( _Point.U, _Point.V, GetBkColor );
 End; { Procedure }

{ ------------------------------------------------------------------ }

Procedure TGrPoint.Convert;
Var
   Status2 : TGrStatus;
   Matrix2 : TGrMatrix;
 Begin
      GetStatus( Status2 );
      GetMatrix( Matrix2 );
      Transform( Matrix2, _Point, Status2 );
 End; { Procedure }

{ ------------------------------------------------------------------ }

Constructor TGrPoint.Load( Var S : TStream );
 Begin
      TGrObject.Load( S );
      S.Read( _Point, SizeOf( _Point ) );
      S.Read( _Color, SizeOf( _Color ) );
 End; { Constructor }

{ ------------------------------------------------------------------ }

Procedure TGrPoint.Store( Var S : TStream );
 Begin
      TGrObject.Store( S );
      S.Write( _Point, SizeOf( _Point ) );
      S.Write( _Color, SizeOf( _Color ) );
 End; { Procedure }

{ ------------------------------------------ TGrLine --------------- }

Constructor TGRLine.Init( X, Y, Z, X2, Y2, Z2 : Integer; Color : Word );
 Begin
      TGrObject.Init;
      _Point[ 1 ].X := X;
      _Point[ 1 ].Y := Y;
      _Point[ 1 ].Z := Z;
      _Point[ 2 ].X := X2;
      _Point[ 2 ].Y := Y2;
      _Point[ 2 ].Z := Z2;
      _Color := Color;
 End; { Constructor }

{ ----------------------------------------------------------------- }

Procedure TGrLine.Draw;
Var
   OldColor : Word;
 Begin
      OldColor := GetColor;
      SetColor( _Color );
      Line( _Point[ 1 ].U, _Point[ 1 ].V, _Point[ 2 ].U, _Point[ 2 ].V );
      SetColor( OldColor );
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrLine.Clear;
Var
   OldColor : Word;
   Status2  : TGrStatus;
 Begin
      OldColor := GetColor;
      SetColor( GetBKColor );
      Line( _Point[ 1 ].U, _Point[ 1 ].V, _Point[ 2 ].U, _Point[ 2 ].V );
      SetColor( OldColor );
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrLine.Convert;
Var
   Status2 : TGrStatus;
   Matrix2 : TGrMatrix;
 Begin
      GetStatus( Status2 );
      GetMatrix( Matrix2 );
      Transform( Matrix2, _Point[ 1 ], Status2 );
      Transform( Matrix2, _Point[ 2 ], Status2 );
 End; { Procedure }

{ ----------------------------------------------------------------- }

Constructor TGrLine.Load( Var S : TStream );
 Begin
      TGrObject.Load( S );
      S.Read( _Point, SizeOf( _Point ) );
      S.Read( _Color, SizeOf( _Color ) );
 End; { Constructor }

{ ------------------------------------------------------------------ }

Procedure TGrLine.Store( Var S : TStream );
 Begin
      TGrObject.Store( S );
      S.Write( _Point, SizeOf( _Point ) );
      S.Write( _Color, SizeOf( _Color ) );
 End; { Procedure }

{ ----------------------------------------------- TGrGroup ------- }

Constructor TGrGroup.Init;
 Begin
      TGrObject.Init;

      PList := New( PCollection, Init( 5, 5 ) );
      If( PList = Nil ) Then
       MemError;
 End; { Constructor }

{ ----------------------------------------------------------------- }

Destructor TGrGroup.Done;
Var TheStatus : TGrStatus;
 Begin
      If( GetStatus( TheStatus ) ) Then
       If( TheStatus.Visible ) Then
        ClearSelf;
      If( PList <> Nil ) Then
       Dispose( PList, Done );
 End; { Destructor }

{ ----------------------------------------------------------------- }

Procedure TGrGroup.Draw;

 Procedure DrawProc( PObj : PGrObject ); Far;
  Begin
       PObj^.Draw;
  End; { Procedure }

 Begin
      PList^.ForEach( @DrawProc );
      DrawSelf;
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrGroup.DrawSelf;
 Begin
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrGroup.ClearSelf;
 Begin
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrGroup.Clear;

 Procedure ClearProc( PObj : PGrObject ); Far;
  Begin
       PObj^.Clear;
  End; { Procedure }

 Begin
       ClearSelf;
       PList^.ForEach( @ClearProc );
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrGroup.Convert;

 Procedure ConvertProc( PObj : PGrObject ); Far;
  Begin
       PObj^.Convert;
  End; { Procedure }

 Begin
      PList^.ForEach( @ConvertProc );
      ConvertSelf;
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrGroup.ConvertSelf;
 Begin
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrGroup.MemError;
 Begin
      Done;
      WriteLn( #10, #13, 'GROUP OBJECT: Memory error in 3d oop routines ...', #10, #13 );
      Halt( 1 );
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrGroup.Insert( Id : Word; Link : PGrObject );
Var
   Status2 : TGrStatus;
   HasOwner : Boolean;
 Begin
      HasOwner := GetStatus( Status2 );             { get old status }
      If( HasOwner And Status2.Visible ) Then
       Clear;                            { everything from screen }

      Link^.ItemID := Id;                { mark the id }
      Link^.POwner := PGrObject( @Self ); { set the owner to group itself }
      PList^.Insert( Link );

      If( HasOwner ) Then
       Link^.Convert;

      If( HasOwner And Status2.Visible ) Then
       Draw;
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrGroup.Delete( Id : Word );
Var  PObj    : PGrObject;
     Status2 : TGrStatus;
     HasOwner: Boolean;

 Function TestProc( PObj : PGrObject ) : Boolean; Far;
  Begin
       If( PObj^.ItemID = Id ) Then
        TestProc := True
       Else
        TestProc := False;
  End; { Function }

{ - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }

 Begin
      HasOwner := GetStatus( Status2 );

      If( HasOwner And Status2.Visible ) Then
       Clear;                            { everything from screen }

      PObj := PGrObject( PList^.FirstThat( @TestProc ) );
      If( PObj <> Nil ) Then
       PList^.Delete( PObj );

     If( HasOwner And Status2.Visible ) Then         { draw again if was visible }
      Draw;
 End; { Procedure }

{ ------------------------------------------------------------------ }

Constructor TGrGroup.Load( Var S : TStream );

 Procedure SetOwners( PObj : PGrObject ); Far;
  Begin
       PObj^.POwner := PGrObject( @Self );
  End; { Procedure }

 Begin
      TGrObject.Load( S );
      PList := PCollection( S.Get );
      PList^.ForEach( @SetOwners );
 End; { Constructor }

{ ------------------------------------------------------------------ }

Procedure TGrGroup.Store( Var S : TStream );
 Begin
      TGrObject.Store( S );
      S.Put( PCollection( PList ) );
 End; { Procedure }

{ -------------------------------------- TGrOpenWire -------- }

Constructor TGrOpenWire.Init( Color : Word );
 Begin
      TGrGroup.Init;
      _Color := Color;
 End; { Constructor }

{ ----------------------------------------------------------- }

Procedure TGrOpenWire.DrawSelf;
Var Index,
    OldColor : Word;
 Begin
      If( PList^.Count > 1 ) Then
       Begin
            OldColor := GetColor;
            SetColor( _Color );
            For Index := 1 To PList^.Count - 1 Do
             Line( PGrPoint( PList^.At( Index - 1 ) )^._Point.U, PGrPoint( PList^.At( Index - 1 ) )^._Point.V,
                   PGrPoint( PList^.At( Index ) )^._Point.U, PGrPoint( PList^.At( Index ) )^._Point.V );
            SetColor( OldColor );
       End; { If Then }
 End; { Procedure }

{ ----------------------------------------------------------- }

Procedure TGrOpenWire.ClearSelf;
Var Index,
    OldColor : Word;
 Begin
      If( PList^.Count > 1 ) Then
       Begin
            OldColor := GetColor;
            SetColor( GetBkColor );
            For Index := 1 To PList^.Count - 1 Do
             Line( PGrPoint( PList^.At( Index - 1 ) )^._Point.U, PGrPoint( PList^.At( Index - 1 ) )^._Point.V,
                   PGrPoint( PList^.At( Index ) )^._Point.U, PGrPoint( PList^.At( Index ) )^._Point.V );
            SetColor( OldColor );
       End; { If Then }
 End; { Procedure }

{ ------------------------------------------------------------------ }

Procedure TGrOpenWire.Store( Var S : TStream );
 Begin
      TGrGroup.Store( S );
      S.Write( _Color, SizeOf( _Color ) );
 End; { Procedure }

{ ----------------------------------------------------------------- }

Constructor TGrOpenWire.Load( Var S : TStream );
 Begin
      TGrGroup.Load( S );
      S.Read( _Color, SizeOf( _Color ) );
 End; { Constructor }

{ ------------------------------------------------------------------ }

Procedure TGrOpenWire.Insert( Id : Word; Link : PGrPoint );
 Begin
      TGrGroup.Insert( Id, Link );
 End; { Procedure }

{ ------------------------------------------------- TGrClosedWire -- }

Procedure TGrClosedWire.DrawSelf;
Var Index,
    OldColor : Word;
 Begin
      If( PList^.Count > 1 ) Then
       Begin
            OldColor := GetColor;
            SetColor( _Color );
            For Index := 1 To PList^.Count - 1 Do
             Line( PGrPoint( PList^.At( Index - 1 ) )^._Point.U, PGrPoint( PList^.At( Index - 1 ) )^._Point.V,
                   PGrPoint( PList^.At( Index ) )^._Point.U, PGrPoint( PList^.At( Index ) )^._Point.V );
            Line( PGrPoint( PList^.At( Index ) )^._Point.U, PGrPoint( PList^.At( Index ) )^._Point.V,
                  PGrPoint( PList^.At( 0 ) )^._Point.U, PGrPoint( PList^.At( 0 ) )^._Point.V );
            SetColor( OldColor );
       End; { If Then }
 End; { Procedure }

{ ------------------------------------------------------------------- }

Procedure TGrClosedWire.ClearSelf;
Var Index,
    OldColor : Word;
 Begin
      If( PList^.Count > 1 ) Then
       Begin
            OldColor := GetColor;
            SetColor( GetBkColor );
            For Index := 1 To PList^.Count - 1 Do
             Line( PGrPoint( PList^.At( Index - 1 ) )^._Point.U, PGrPoint( PList^.At( Index - 1 ) )^._Point.V,
                   PGrPoint( PList^.At( Index ) )^._Point.U, PGrPoint( PList^.At( Index ) )^._Point.V );
            Line( PGrPoint( PList^.At( Index ) )^._Point.U, PGrPoint( PList^.At( Index ) )^._Point.V,
                  PGrPoint( PList^.At( 0 ) )^._Point.U, PGrPoint( PList^.At( 0 ) )^._Point.V );
            SetColor( OldColor );
       End; { If Then }
 End; { Procedure }

{ --------------------------------------- TGrOpenWireFrame -- }

Procedure TGrOpenWireFrame.DrawSelf;
Var OldColor,
    LineNr,
    PointNr,
    MinNrOfPoints,
    MaxNrOfPoints  : Word;
    TempLine,
    TheSmallestLine,
    TheBiggestLine : PGrOpenWire;
    X1, X2, Y1, Y2 : Integer;
 Begin
      If( PList^.Count > 1 ) Then
       Begin
            OldColor := GetColor;
            SetColor( _Color );

            For LineNr := 1 To PList^.Count - 1 Do
             Begin
                  TheBiggestLine := PList^.At( LineNr );
                  TheSmallestLine := PList^.At( LineNr - 1 );
                  MinNrOfPoints := TheSmallestLine^.PList^.Count;
                  MaxNrOfPoints := TheBiggestLine^.PList^.Count;

                  If( MaxNrOfPoints < MinNrOfPoints ) Then   { swap if needed }
                   Begin
                        PointNr := MaxNrOfPoints;
                        MaxNrOfPoints := MinNrOfPoints;
                        MinNrOfPoints := PointNr;
                        TempLine := TheBiggestLine;
                        TheBiggestLine := TheSmallestLine;
                        TheSmallestLine := TempLine;
                   End; { If Then }

                  If( MinNrOfPoints > 0 ) Then
                   Begin
                        For PointNr := 0 To MinNrOfPoints - 1 Do
                         Begin
                              X1 := PGrPoint( TheBiggestLine^.PList^.At( PointNr ) )^._Point.U;
                              Y1 := PGrPoint( TheBiggestLine^.PList^.At( PointNr ) )^._Point.V;
                              X2 := PGrPoint( TheSmallestLine^.PList^.At( PointNr ) )^._Point.U;
                              Y2 := PGrPoint( TheSmallestLine^.PList^.At( PointNr ) )^._Point.V;
                              Line( X1, Y1, X2, Y2 );
                         End; { For Do }

                        If( MaxNrOfPoints > MinNrOfPoints ) Then
                         For PointNr := MinNrOfPoints + 1 To MaxNrOfPoints - 1 Do
                          Begin
                               X1 := PGrPoint( TheBiggestLine^.PList^.At( PointNr ) )^._Point.U;
                               Y1 := PGrPoint( TheBiggestLine^.PList^.At( PointNr ) )^._Point.V;
                               Line( X1, Y1, X2, Y2 );
                          End; { For Do }
                   End; { If Then }
             End; { For Do }

            SetColor( OldColor );
       End; { If Then }
 End; { Procedure }

{ ----------------------------------------------------------- }

Procedure TGrOpenWireFrame.ClearSelf;
Var OldColor,
    LineNr,
    PointNr,
    MinNrOfPoints,
    MaxNrOfPoints  : Word;
    TempLine,
    TheSmallestLine,
    TheBiggestLine : PGrOpenWire;
    X1, X2, Y1, Y2 : Integer;
 Begin
      If( PList^.Count > 1 ) Then
       Begin
            OldColor := GetColor;
            SetColor( GetBkColor );

            For LineNr := 1 To PList^.Count - 1 Do
             Begin
                  TheBiggestLine := PList^.At( LineNr );
                  TheSmallestLine := PList^.At( LineNr - 1 );
                  MinNrOfPoints := TheSmallestLine^.PList^.Count;
                  MaxNrOfPoints := TheBiggestLine^.PList^.Count;

                  If( MaxNrOfPoints < MinNrOfPoints ) Then   { swap if needed }
                   Begin
                        PointNr := MaxNrOfPoints;
                        MaxNrOfPoints := MinNrOfPoints;
                        MinNrOfPoints := PointNr;
                        TempLine := TheBiggestLine;
                        TheBiggestLine := TheSmallestLine;
                        TheSmallestLine := TempLine;
                   End; { If Then }

                  If( MinNrOfPoints > 0 ) Then
                   Begin
                        For PointNr := 0 To MinNrOfPoints - 1 Do
                         Begin
                              X1 := PGrPoint( TheBiggestLine^.PList^.At( PointNr ) )^._Point.U;
                              Y1 := PGrPoint( TheBiggestLine^.PList^.At( PointNr ) )^._Point.V;
                              X2 := PGrPoint( TheSmallestLine^.PList^.At( PointNr ) )^._Point.U;
                              Y2 := PGrPoint( TheSmallestLine^.PList^.At( PointNr ) )^._Point.V;
                              Line( X1, Y1, X2, Y2 );
                         End; { For Do }

                        If( MaxNrOfPoints > MinNrOfPoints ) Then
                         For PointNr := MinNrOfPoints + 1 To MaxNrOfPoints - 1 Do
                          Begin
                               X1 := PGrPoint( TheBiggestLine^.PList^.At( PointNr ) )^._Point.U;
                               Y1 := PGrPoint( TheBiggestLine^.PList^.At( PointNr ) )^._Point.V;
                               Line( X1, Y1, X2, Y2 );
                          End; { For Do }
                   End; { If Then }
             End; { For Do }

            SetColor( OldColor );
       End; { If Then }
 End; { Procedure }

{ ------------------------------------------------------------------- }

Procedure TGrOpenWireFrame.Insert( Id : Word; Link : PGrOpenWire );
 Begin
      TGrGroup.Insert( Id, Link );
 End; { Procedure }

{ ---------------------------------------------- TGrClosedWireFrame -- }

Procedure TGrClosedWireFrame.DrawSelf;
Var OldColor,
    PointNr,
    MinNrOfPoints,
    MaxNrOfPoints  : Word;
    TempLine,
    TheSmallestLine,
    TheBiggestLine : PGrOpenWire;
    X1, X2, Y1, Y2 : Integer;
 Begin
      If( PList^.Count > 1 ) Then
       Begin
            TGrOpenWireFrame.DrawSelf;

            OldColor := GetColor;
            SetColor( _Color );

            TheBiggestLine := PList^.At( 0 );
            TheSmallestLine := PList^.At( PList^.Count - 1 );
            MinNrOfPoints := TheSmallestLine^.PList^.Count;
            MaxNrOfPoints := TheBiggestLine^.PList^.Count;

            If( MaxNrOfPoints < MinNrOfPoints ) Then   { swap if needed }
             Begin
                  PointNr := MaxNrOfPoints;
                  MaxNrOfPoints := MinNrOfPoints;
                  MinNrOfPoints := PointNr;
                  TempLine := TheBiggestLine;
                  TheBiggestLine := TheSmallestLine;
                  TheSmallestLine := TempLine;
             End; { If Then }

            If( MinNrOfPoints > 0 ) Then
             Begin
                  For PointNr := 0 To MinNrOfPoints - 1 Do
                   Begin
                        X1 := PGrPoint( TheBiggestLine^.PList^.At( PointNr ) )^._Point.U;
                        Y1 := PGrPoint( TheBiggestLine^.PList^.At( PointNr ) )^._Point.V;
                        X2 := PGrPoint( TheSmallestLine^.PList^.At( PointNr ) )^._Point.U;
                        Y2 := PGrPoint( TheSmallestLine^.PList^.At( PointNr ) )^._Point.V;
                        Line( X1, Y1, X2, Y2 );
                   End; { For Do }

                  If( MaxNrOfPoints > MinNrOfPoints ) Then
                   For PointNr := MinNrOfPoints + 1 To MaxNrOfPoints - 1 Do
                    Begin
                         X1 := PGrPoint( TheBiggestLine^.PList^.At( PointNr ) )^._Point.U;
                         Y1 := PGrPoint( TheBiggestLine^.PList^.At( PointNr ) )^._Point.V;
                         Line( X1, Y1, X2, Y2 );
                    End; { For Do }
             End; { If Then }

            SetColor( OldColor );
       End; { If Then }
 End; { Procedure }

{ ------------------------------------------------------------------- }

Procedure TGrClosedWireFrame.ClearSelf;
Var OldColor,
    PointNr,
    MinNrOfPoints,
    MaxNrOfPoints  : Word;
    TempLine,
    TheSmallestLine,
    TheBiggestLine : PGrOpenWire;
    X1, X2, Y1, Y2 : Integer;
 Begin
      If( PList^.Count > 1 ) Then
       Begin
            OldColor := GetColor;
            SetColor( GetBkColor );

            TheBiggestLine := PList^.At( 0 );
            TheSmallestLine := PList^.At( PList^.Count - 1 );
            MinNrOfPoints := TheSmallestLine^.PList^.Count;
            MaxNrOfPoints := TheBiggestLine^.PList^.Count;

            If( MaxNrOfPoints < MinNrOfPoints ) Then   { swap if needed }
             Begin
                  PointNr := MaxNrOfPoints;
                  MaxNrOfPoints := MinNrOfPoints;
                  MinNrOfPoints := PointNr;
                  TempLine := TheBiggestLine;
                  TheBiggestLine := TheSmallestLine;
                  TheSmallestLine := TempLine;
             End; { If Then }

            If( MinNrOfPoints > 0 ) Then
             Begin
                  For PointNr := 0 To MinNrOfPoints - 1 Do
                   Begin
                        X1 := PGrPoint( TheBiggestLine^.PList^.At( PointNr ) )^._Point.U;
                        Y1 := PGrPoint( TheBiggestLine^.PList^.At( PointNr ) )^._Point.V;
                        X2 := PGrPoint( TheSmallestLine^.PList^.At( PointNr ) )^._Point.U;
                        Y2 := PGrPoint( TheSmallestLine^.PList^.At( PointNr ) )^._Point.V;
                        Line( X1, Y1, X2, Y2 );
                   End; { For Do }

                  If( MaxNrOfPoints > MinNrOfPoints ) Then
                   For PointNr := MinNrOfPoints + 1 To MaxNrOfPoints - 1 Do
                    Begin
                         X1 := PGrPoint( TheBiggestLine^.PList^.At( PointNr ) )^._Point.U;
                         Y1 := PGrPoint( TheBiggestLine^.PList^.At( PointNr ) )^._Point.V;
                         Line( X1, Y1, X2, Y2 );
                    End; { For Do }
             End; { If Then }

            SetColor( OldColor );
            TGrOpenWireFrame.ClearSelf;
       End; { If Then }
 End; { Procedure }

{ ------------------------------------------------ TGrSystem ------- }

Constructor TGrSystem.Init( X, Y, Z : Integer; Color : Word );
 Begin
      TGrGroup.Init;

      _Status.Phi := 10;                     { Def angles for a xyz system }
      _Status.Theta := 80;
      _Status.Rho := 0;
      _Status.RelX := 200;
      _Status.RelY := 200;
      _Status.ZoomF := 1.0;
      _Status.PerspF := 0.0025;
      _Status.Visible := False;
      _Status.Perspective := False;
      _Status.LabelAxes := True;
      _Status.ShowTheAxes := True;
      _Status.XLabel := 'x';
      _Status.YLabel := 'y';
      _Status.ZLabel := 'z';

      _Origin.X := 0;
      _Origin.Y := 0;
      _Origin.Z := 0;
      _Axes[ 1 ].X := X;
      _Axes[ 1 ].Y := 0;
      _Axes[ 1 ].Z := 0;
      _Axes[ 2 ].X := 0;
      _Axes[ 2 ].Y := Y;
      _Axes[ 2 ].Z := 0;
      _Axes[ 3 ].X := 0;
      _Axes[ 3 ].Y := 0;
      _Axes[ 3 ].Z := Z;
      _Color := Color;

      Convert;
 End; { Constructor }

{ ----------------------------------------------------------------- }

Procedure TGrSystem.CalculateMatrix;
 Begin
      _Matrix[ 1, 1 ] := ( -1 * GrSin( _Status.Phi ) );          { building up the matrix }
      _Matrix[ 1, 2 ] := ( GrCos( _Status.Phi ) );
      _Matrix[ 1, 3 ] := ( 0 );
      _Matrix[ 1, 4 ] := ( 0 );
      _Matrix[ 2, 1 ] := ( -1 * GrCos( _Status.Phi ) * GrCos( _Status.Theta ) );
      _Matrix[ 2, 2 ] := ( -1 * GrSin( _Status.Phi ) * GrCos( _Status.Theta ) );
      _Matrix[ 2, 3 ] := ( GrSin( _Status.Theta ) );
      _Matrix[ 2, 4 ] := ( 0 );
      _Matrix[ 3, 1 ] := ( -1 * GrCos( _Status.Phi ) * GrSin( _Status.Theta ) );
      _Matrix[ 3, 2 ] := ( -1 * GrSin( _Status.Phi ) * GrSin( _Status.Theta ) );
      _Matrix[ 3, 3 ] := ( -1 * GrCos( _Status.Theta ) );
      _Matrix[ 3, 4 ] := ( _Status.Rho );
      _Matrix[ 4, 1 ] := ( 0 );
      _Matrix[ 4, 2 ] := ( 0 );
      _Matrix[ 4, 3 ] := ( 0 );
      _Matrix[ 4, 4 ] := ( 1 );
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrSystem.Insert( Id : Word; Link : PGrObject );
 Begin
      If( TypeOf( Link^ ) = TypeOf( TGrSystem ) ) Then
       Begin
            Done;
            Writeln( #10, #13, 'SYSTEM OBJECT: Cannot insert a system ...', #10, #13 );
            Halt( 1 );
       End { If Then }
      Else
       TGrGroup.Insert( Id, Link );
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrSystem.DrawSelf;
Var
   OldColor   : Word;
 Begin
      If( _Status.ShowTheAxes ) Then
       Begin
            OldColor := GetColor;
            SetColor( _Color );

            Line( _Origin.U, _Origin.V, _Axes[ 1 ].U, _Axes[ 1 ].V );
            Line( _Origin.U, _Origin.V, _Axes[ 2 ].U, _Axes[ 2 ].V );
            Line( _Origin.U, _Origin.V, _Axes[ 3 ].U, _Axes[ 3 ].V );

            If( _Status.LabelAxes ) Then
             Begin
                  OutTextXY( _Axes[ 1 ].U, _Axes[ 1 ].V, _Status.XLabel );
                  OutTextXY( _Axes[ 2 ].U, _Axes[ 2 ].V, _Status.YLabel );
                  OutTextXY( _Axes[ 3 ].U, _Axes[ 3 ].V, _Status.ZLabel );
             End; { If Then }

            SetColor( OldColor );
       End; { If Then }
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrSystem.ClearSelf;
Var
   OldColor : Integer;

 Begin
      If( _Status.ShowTheAxes ) Then
       Begin
            OldColor := GetColor;
            SetColor( GetBkColor );

            Line( _Origin.U, _Origin.V, _Axes[ 1 ].U, _Axes[ 1 ].V );
            Line( _Origin.U, _Origin.V, _Axes[ 2 ].U, _Axes[ 2 ].V );
            Line( _Origin.U, _Origin.V, _Axes[ 3 ].U, _Axes[ 3 ].V );

            If( _Status.LabelAxes ) Then
             Begin
                  OutTextXY( _Axes[ 1 ].U, _Axes[ 1 ].V, _Status.XLabel );
                  OutTextXY( _Axes[ 2 ].U, _Axes[ 2 ].V, _Status.YLabel );
                  OutTextXY( _Axes[ 3 ].U, _Axes[ 3 ].V, _Status.ZLabel );
             End; { If Then }

            SetColor( OldColor );
       End; { If Then }
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrSystem.ConvertSelf;
 Begin
      If( _Status.ShowTheAxes ) Then
       Begin
            TransForm( _Matrix, _Origin, _Status );
            TransForm( _Matrix, _Axes[ 1 ], _Status );
            TransForm( _Matrix, _Axes[ 2 ], _Status );
            TransForm( _Matrix, _Axes[ 3 ], _Status );
       End; { If Then }
 End; { Procedure }

{ ----------------------------------------------------------------- }

Function TGrSystem.GetMatrix( Var M : TGrMatrix ) : Boolean;
 Begin
      M := _Matrix;
      GetMatrix := True;
 End; { Function }

{ ----------------------------------------------------------------- }

Function TGrSystem.GetStatus( Var Status2 : TGrStatus ) : Boolean;
 Begin
      Status2 := _Status;
      GetStatus := True;
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrSystem.Convert;
 Begin
      CalculateMatrix;
      TGrGroup.Convert;
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrSystem.SetStatus( Var Status2 : TGrStatus );
 Begin
      If( _Status.Visible ) Then
       Clear;

      _Status := Status2;

      If( _Status.Rho < 0 ) Then
       _Status.Rho := 0;

      If( _Status.PerspF <= 0 ) Then
       _Status.PerspF := 0.0025;

      If( _Status.ZoomF <= 0 ) Then
       _Status.ZoomF := 1.0;

      Convert;

      If( _Status.Visible ) Then
       Draw;
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrSystem.KeybRotate( Angle : Integer );
Var
   C : Char;
   Modified : Boolean;
   Status2 : TGrStatus;
 Begin
      _Status.Visible := True;
      Draw;
      Repeat
            C := ReadKey;

            If( C = Chr( 0 ) ) Then
             Begin
                  C := ReadKey;
                  GetStatus( Status2 );
                  Modified := True;
                  Case C Of
                       Arrow_Up    : Status2.Theta := ( Status2.Theta - Angle ) Mod 360;
                       Arrow_Down  : Status2.Theta := ( Status2.Theta + Angle ) Mod 360;
                       Arrow_Left  : Status2.Phi := ( Status2.Phi - Angle ) Mod 360;
                       Arrow_Right : Status2.Phi := ( Status2.Phi + Angle ) Mod 360;
                       Alt_P       : If( Status2.Perspective ) Then
                                      Status2.Perspective := False
                                     Else
                                      Status2.Perspective := True;
                       Alt_A       : If( Status2.LabelAxes ) Then
                                      Status2.LabelAxes := False
                                     Else
                                      Status2.LabelAxes := True;
                       Alt_Z       : If( Status2.ZoomF > 0.1 ) Then
                                       Status2.ZoomF := Status2.ZoomF - 0.1;
                       Alt_X       : If( Status2.ZoomF < 100 ) Then
                                       Status2.ZoomF := Status2.ZoomF + 0.1;
                       Ctrl_Left   : If( Status2.Rho < 10000 ) Then
                                       Status2.Rho := Status2.Rho + 10;
                       Ctrl_Right  : If( Status2.Rho > 10 ) Then
                                       Status2.Rho := Status2.Rho - 10;
                  Else
                      Modified := False;
                  End; { Case Of }
                  If( Modified ) Then
                   SetStatus( Status2 );
             End; { If Then }

      Until( C = #27 );
 End; { Procedure }

{ ----------------------------------------------------------------- }

Procedure TGrSystem.DemoRotate( Angle : Integer; Wait : Word );
Var
   Counter   : Word;
   Status2   : TGrStatus;
   Modified,
   Waiting   : Boolean;
   C         : Char;
 Begin
      Counter := 0;
      Waiting := False;
      _Status.Visible := True;
      Draw;
      Repeat
            If( Not Waiting ) Then
             Inc( Counter );

            If( KeyPressed ) Then
             C := ReadKey
            Else
             C := #1;

            If( C = Chr( 0 ) ) Then
             Begin
                  C := ReadKey;
                  GetStatus( Status2 );
                  Modified := True;
                  Case C Of
                       Arrow_Up    : Status2.Theta := ( Status2.Theta - Angle ) Mod 360;
                       Arrow_Down  : Status2.Theta := ( Status2.Theta + Angle ) Mod 360;
                       Alt_P       : If( Status2.Perspective ) Then
                                      Status2.Perspective := False
                                     Else
                                      Status2.Perspective := True;
                       Alt_A       : If( Status2.LabelAxes ) Then
                                      Status2.LabelAxes := False
                                     Else
                                      Status2.LabelAxes := True;
                       Alt_S       : If( Waiting ) Then
                                      Waiting := False
                                     Else
                                      Waiting := True;
                       Alt_Z       : If( Status2.ZoomF > 0.1 ) Then
                                       Status2.ZoomF := Status2.ZoomF - 0.1;
                       Alt_X       : If( Status2.ZoomF < 100 ) Then
                                       Status2.ZoomF := Status2.ZoomF + 0.1;
                       Ctrl_Left   : If( Status2.Rho < 10000 ) Then
                                       Status2.Rho := Status2.Rho + 10;
                       Ctrl_Right  : If( Status2.Rho > 10 ) Then
                                       Status2.Rho := Status2.Rho - 10;
                  Else
                      Modified := False;
                  End; { Case Of }
                  If( Modified ) Then
                   SetStatus( Status2 );
             End; { If Then }

            If( Counter >= Wait ) Then
             Begin
                  GetStatus( Status2 );
                  Status2.Phi := ( Status2.Phi + Angle ) Mod 360;
                  SetStatus( Status2 );
                  Counter := 0;
             End; { If Then }

      Until( C = #27 );

 End; { Procedure }

{ ----------------------------------------------------------------- }

Constructor TGrSystem.Load( Var S : TStream );
 Begin
      TGrGroup.Load( S );
      S.Read( _Status, SizeOf( _Status ) );
      S.Read( _Matrix, SizeOf( _Matrix ) );
      S.Read( _Origin, SizeOf( _Origin ) );
      S.Read( _Axes, SizeOf( _Axes ) );
      S.Read( _Color, SizeOf( _Color ) );
 End; { Constructor }

{ ------------------------------------------------------------------ }

Procedure TGrSystem.Store( Var S : TStream );
 Begin
      TGrGroup.Store( S );
      S.Write( _Status, SizeOf( _Status ) );
      S.Write( _Matrix, SizeOf( _Matrix ) );
      S.Write( _Origin, SizeOf( _Origin ) );
      S.Write( _Axes, SizeOf( _Axes ) );
      S.Write( _Color, SizeOf( _Color ) );
 End; { Procedure }

{ ----------------------------------------------------------------- }

Begin                 { Main }
     OldExitProc := ExitProc;
     ExitProc := Addr( ExitRoutine );
     _Gr3D_AuthorMessage := True;
     AspectFactor := 1;
End. { Unit }