{       []--------------------------------------------[]
         |                MATHOBJ.pas                 |
        []--------------------------------------------[]
         |                                            |
         | AUTHOR:    MFSomers, 1998. Holland.        |
         | UPDATES:   None. Feb. 1998.                |
         | USE:       A simple calculus utility which |
         |            uses the stackparser and 3D     |
         |            system objects to plot and      |
         |            calculate on functions ...      |
         |                                            |
        []--------------------------------------------[]
}

{$O-,B-,F+,A+,I-,N+,P-,Q-,R-,S-,T+,V+,G-,E+,X+}
{ must be set or else input check will fail !!}

Unit MATHOBJ;

Interface

{ All usable units summarized for the easy way ... }

Uses DOS, CRT, GRAPH, GDRIVERS, GR3D, PARSER, APP, GADGETS,
     DIALOGS, DRIVERS, HISTLIST, MEMORY, MENUS, MSGBOX,
     OBJECTS, STDDLG, VIEWS;

{ --------------- INTERFACE OBJECT DEFINITIONS ---------------------- }

Type TMyApplication = Object( TApplication )                   { My application }
                            MemMonitor  : PHeapView;
                            TheParser   : PStackParser;

                            Constructor Init;
                            Destructor  Done; Virtual;
                            Procedure   InitMenuBar; Virtual;
                            Procedure   InitStatusLine; Virtual;
                            Procedure   HandleEvent( Var Event : TEvent ); Virtual;
                            Procedure   Idle; Virtual;
                            Procedure   SetVar( I : Integer; V : Extended ); Virtual;
                            Function    GetVar( I : Integer ) : Extended; Virtual;
                            Function    Differentiate( I : Integer; Var D : Extended ) : Boolean; Virtual;
                            Function    ItterateRoot( I : Integer; X1 : Extended; Var X2 : Extended ) : Boolean; Virtual;
                            Procedure   Plot2DFunc( X, Y, X2, Y2, UX, UY : Extended; I : Integer ); Virtual;
                            Procedure   Plot3DFunc( R, X, Y, X2, Y2, XI, YI : Extended; I1, I2 : Integer ); Virtual;
                            Procedure   PlotDifEqu( FY, X, Y, X2, Y2, UX, UY : Extended; I : Integer ); Virtual;
                            Procedure   PlotDerivative( X, Y, X2, Y2, UX, UY : Extended; I : Integer ); Virtual;
                            Procedure   BuildGrPlotSystem( X, Y, X2, Y2, UX, UY : Extended; I : Integer;
                                                           S1, S2 : String; Var L, W : Integer;
                                                           Var OX, OY : LongInt ); Virtual;
                            Function    IntoGraphics : Boolean; Virtual;
                            Procedure   OutofGraphics; Virtual;
                            Function    GetResult( Var P : Extended ) : Boolean; Virtual;
                      End; { Object }
     PMyApplication = ^TMyApplication;

{ -------------- IMPLEMENTATION PART OF UNIT ------------------------ }

Implementation

Const Extention   = '*.MTH';                    { default extention for math files }
      FileID      = 'MATH FILE' + #26;          { math file ID }
      HistId      = 0;                          { filedialog history ID }
      VarSize     = 19;                         { max length of variable string }
      ValSize     = 30;                         { max length of extended string }
      MinInterval = 10;                         { minimum nr of intervals }
      DifferentialIV = 1E-6;                    { interval used during differentation }
      NrPlotPixels = 0.05;                      { inc of pixels during plots }
      MinMemFor3D = 90 * 1024;                  { need at least 90 kb mem left during 3D }
      PixelsPer3DUnit = 20;                     { nr pixels per unit in 3D plot }
      PrimInterV = 0.001;                       { interval during plotting dif. equ. }

Const                                   { the used commands }
      cmNoCommand    = 2000;            { no command }
      cmAbout        = 2005;            { about dialog command }
      cmNewDefines   = 2010;            { new defines command }
      cmLoadDefines  = 2020;            { load defines command }

      cmDefVar       = 2030;            { define variable command }
      cmDeleteVar    = 2040;            { delete variable command }

      cmPlot2D       = 100;             { plot a function }
      cmPlot3D       = 110;             { plot a function of 2 vars }
      cmIntegrate    = 120;             { integrate a function }
      cmDifferentiate = 130;            { differentiate a function }
      cmScanRoots    = 140;             { find possible roots of function }
      cmSaveDefines  = 150;             { save the entered defines }
      cmEvaluate     = 160;             { evaluate function command }
      cmPlotDerivative = 170;           { plot derivative command }
      cmPlotDifEqu   = 180;             { plot primitive command }

Type
     TMyStatusLine = Object( TStatusLine ) { a new statusline }
                           TheMessage : PString;        { a new statusline message to show the formula }

                           Destructor Done; Virtual;
                           Function   Hint( HCTX : Word ) : String; Virtual;
                           Procedure  MessageStr( P : PString ); Virtual;
                           Procedure  Draw; Virtual;
                     End; { Object }
     PMyStatusLine = ^TMyStatusLine;

     TMyVarList  = Object( TListBox ) { a listbox that shows the variables in a variable list }
                         DontShowVal : Boolean;

                         Function GetText( Item : Integer; L : Integer ) : String; Virtual;
                   End; { Object }
     PMyVarList = ^TMyVarList;

     TMyDefineDlg = Object( TDialog ) { the define input dialog }
                          ListBox : PMyVarList;   { contains a listbox }

                          Constructor Init;
                          Procedure HandleEvent( Var Event : TEvent ); Virtual;
                    End; { Object }
     PMyDefineDlg = ^TMyDefineDlg;

     TMyVarInput = Object( TDialog )    { the variable edit dialog }
                        Constructor Init;
                   End; { Object }
     PMyVarInput = ^TMyVarInput;

     TMyIntervalDlg = Object( TDialog )   { the interval dialog for roots or integration }
                            Constructor Init( S : String );
                      End; { Object }
     PMyIntervalDlg = ^TMyIntervalDlg;

     TMyMessageDlg = Object( TDialog )    { a message dialog }
                          Constructor Init( S : String );
                          Procedure SizeLimits( Var Min, Max : TPoint ); Virtual;
                          Procedure HandleMouseEvents; Virtual;
                     End; { Object }
     PMyMessageDlg = ^TMyMessageDlg;

     TMyDifferentDlg = Object( TDialog )   { the derivative input dialog }
                            Constructor Init;
                       End; { Object }
     PMyDifferentDlg = ^TMyDifferentDlg;

     TMyPlotDlg   = Object( TDialog )       { the plot range dialog }
                         Constructor Init( S : String );
                    End; { Object }
     PMyPlotDlg   = ^TMyPlotDlg;

     TMy3DPlotDlg   = Object( TDialog )       { the 3D plot range dialog }
                         Constructor Init;
                    End; { Object }
     PMy3DPlotDlg   = ^TMy3DPlotDlg;

     TMyPrimPlotDlg = Object( TDialog )       { the primitive plot range dialog }
                          Constructor Init;
                      End; { Object }
     PMyPrimPlotDlg = ^TMyPrimPlotDlg;

{ --------------- RECORDS ------------------------------------------------ }

     InpDataRec = Record         { the variable edit dialog data record }
                        Value : String[ ValSize ];
                        Id   : String[ VarSize ];
                   End; { Record }

     DefDataRec = Record         { the define dialog data record }
                        List  : PCollection;
                        Focus : Integer;
                        Formula : String;
                  End; { Record }

     IntervDataRec = Record   { the interval dialog data record }
                           StartV : String[ ValSize ];
                           EndV   : String[ ValSize ];
                           Interv : String[ ValSize ];
                           List   : PCollection;
                           Focus  : Integer;
                     End; { Record }

     DifferentDataRec = Record   { the different dialog object data record }
                             List  : PCollection;
                             Focus : Integer;
                        End; { Record }

     PlotRangeDataRec = Record     { the my plot function object data record }
                            List : PCollection;
                            Item : Integer;
                            XS, XE, XU : String[ ValSize ];
                            YS, YE, YU : String[ ValSize ];
                        End; { Record }

     PrimPlotRangeDataRec = Record     { the my plot primitive object data record }
                                List : PCollection;
                                Item : Integer;
                                XS, XE, XU : String[ ValSize ];
                                YS, YE, YU : String[ ValSize ];
                                FY         : String[ ValSize ];
                            End; { Record }

     Plot3DRangeDataRec = Record     { the my plot 3D function object data record }
                              List1 : PCollection;
                              Item1 : Integer;
                              List2 : PCollection;
                              Item2 : Integer;
                              XS, XE, XI : String[ ValSize ];
                              YS, YE, YI : String[ ValSize ];
                          End; { Record }

{ ------------------- OBJECT ROUTINES --------------------------------- }

 Function TMyStatusLine.Hint( HCTX : Word ) : String;
  Begin
       Hint := 'MATH.exe 2.00 -- By: MFSomers, 1999.';

       If( TheMessage <> Nil ) Then            { use message if exists one }
        Hint := TheMessage^;
  End; { Function }

{ --------------------------------------------------------------------- }

 Procedure TMyStatusLine.Draw;
 Var R : TRect;
  Begin
       GetExtent( R );               { draw the statusline without a | char }
       WriteChar( 0, 0, #32, 1, 2 );
       WriteStr( 1, 0, Hint( HelpCtx ), 2 );
       WriteChar( Length( Hint( HelpCtx ) ) + 1, 0, #32, 2, R.B.X - Length( Hint( HelpCtx ) ) - 1 );
  End; { Procedure }

{ --------------------------------------------------------------------- }

 Procedure TMyStatusLine.MessageStr( P : PString );
  Begin
       If( TheMessage <> Nil ) Then     { dispose old message }
        DisposeStr( TheMessage );

       TheMessage := P;                  { set new message }

       Draw;                             { and draw new statusline }
  End; { Procedure }

{ -------------------------------------------------------------------- }

 Destructor TMyStatusLine.Done;
  Begin
       If( TheMessage <> Nil ) Then     { dispose message if exists }
        DisposeStr( TheMessage );
       TStatusLine.Done;
  End; { Destructor }

{ -------------------------------------------------------------------- }

 Constructor TMyMessageDlg.Init( S : String );

 Function NrLines( S : String ) : Integer;     { Get Nmbr lines in S }
 Var Count : Integer;
     Nr    : Integer;
  Begin
       Nr := 1;

       For Count := 1 To Ord( S[ 0 ] ) Do      { count nr of #13 }
        If( S[ Count ] = #13 ) Then
         Nr := Nr + 1;

       NrLines := Nr;
  End; { Function }

 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }

 Function StrLngth( S : String ) : Integer;     { To get max width in S }
 Var Count : Integer;
     L     : Integer;
     MaxL  : Integer;
  Begin
       L    := 0;
       MaxL := 0;

       For Count := 1 To Ord( S[ 0 ] ) Do       { scan string }
        If( S[ Count ] = #13 ) Then            { for new lines }
         Begin
              If( L > MaxL ) Then              { if wider line found }
               Begin
                    MaxL := L;
                    L := 0;
               End; { If Then }
         End { If Then }
        Else
         If( S[ Count ] <> #3 ) Then          { ignore #3 chars, not visible }
          L := L + 1;                       { count length of current line }

       If( MaxL > L ) Then                  { if last scanned line is smaller }
        StrLngth := MaxL
       Else
        StrLngth := L;
  End; { Function }

 { - - - - - - - - MAIN TMYMESSAGE.INIT - - - - - - - - - - - - - }

 Var R  : TRect;
     ST : PStaticText;
  Begin                                  { build size using S }
       R.Assign( 3, 3, StrLngth( S ) + 7, 5 + NrLines( S ) );
       TDialog.Init( R, 'Message' );
       Options := Options Or ofCentered;    { center in desktop }
       Flags := Flags XOr wfClose;          { no close icon }
                                            { insert static text }
       R.Assign( 2, 1, StrLngth( S ) + 2, 1 + NrLines( S ) );
       ST := New( PStaticText, Init( R, S ) );
       TDialog.Insert( ST );
  End; { Constructor }

{ ------------------------------------------------------------------- }

 Procedure TMyMessageDlg.SizeLimits( Var Min, Max : TPoint );
 Var R : TRect;
  Begin
       TDialog.SizeLimits( Min, Max );     { minimum size is current size }
       GetExtent( R );
       Min := R.B;
  End; { Procedure }

{ -------------------------------------------------------------------- }

 Procedure TMyMessageDlg.HandleMouseEvents;
 Var E : TEvent;
  Begin                              { regulary called during calculations }
       GetEvent( E );                { to make mouse events not ignored }
       If( E.What And evMouse <> 0 ) Then
        HandleEvent( E );            { any other event is ignored }
  End; { Procedure }

{ -------------------------------------------------------------------- }

 Constructor TMyPlotDlg.Init( S : String );
 Var R : TRect;
     L : PLabel;
     B : PButton;
     I : PInputLine;
     SCR : PScrollBar;
     LB : PMyVarList;
  Begin
       R.Assign( 5, 5, 57, 25 );
       TDialog.Init( R, S );
       Options := Options Or ofCentered;

       R.Assign( 49, 3, 50, 8 );
       SCR := New( PScrollBar, Init( R ) );
       TDialog.Insert( SCR );
       R.Assign( 2, 3, 49, 8 );
       LB := New( PMyVarList, Init( R, 1, SCR ) );
       LB^.DontShowVal := False;
       TDialog.Insert( LB );
       R.Assign( 2, 2, 49, 3 );
       L := New( PLabel, Init( R, '~V~ariable', LB ) );
       TDialog.Insert( L );

       R.Assign( 2, 10, 16, 11 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 2, 9, 16, 10 );
       L := New( PLabel, Init( R, '~X~ Start', I ) );
       TDialog.Insert( L );

       R.Assign( 19, 10, 33, 11 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 19, 9, 33, 10 );
       L := New( PLabel, Init( R, '~X~ End', I ) );
       TDialog.Insert( L );

       R.Assign( 36, 10, 50, 11 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 36, 9, 50, 10 );
       L := New( PLabel, Init( R, '~X~ Unit', I ) );
       TDialog.Insert( L );

       R.Assign( 2, 13, 16, 14 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 2, 12, 16, 13 );
       L := New( PLabel, Init( R, '~Y~ Start', I ) );
       TDialog.Insert( L );

       R.Assign( 19, 13, 33, 14 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 19, 12, 33, 13 );
       L := New( PLabel, Init( R, '~Y~ End', I ) );
       TDialog.Insert( L );

       R.Assign( 36, 13, 50, 14 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 36, 12, 50, 13 );
       L := New( PLabel, Init( R, '~Y~ Unit', I ) );
       TDialog.Insert( L );

       R.Assign( 6, 16, 21, 18 );
       B := New( PButton, Init( R, '~O~k', cmOk, bfDefault ) );
       TDialog.Insert( B );

       R.Assign( 31, 16, 46, 18 );
       B := New( PButton, Init( R, '~C~ancel', cmCancel, bfNormal ) );
       TDialog.Insert( B );

       SelectNext( False );

  End; { Constructor }

{ -------------------------------------------------------------------- }

 Constructor TMyPrimPlotDlg.Init;
 Var R : TRect;
     L : PLabel;
     B : PButton;
     I : PInputLine;
     SCR : PScrollBar;
     LB : PMyVarList;
  Begin
       R.Assign( 5, 5, 57, 25 );
       TDialog.Init( R, 'Primitive plot range' );
       Options := Options Or ofCentered;

       R.Assign( 49, 3, 50, 8 );
       SCR := New( PScrollBar, Init( R ) );
       TDialog.Insert( SCR );
       R.Assign( 2, 3, 49, 8 );
       LB := New( PMyVarList, Init( R, 1, SCR ) );
       LB^.DontShowVal := False;
       TDialog.Insert( LB );
       R.Assign( 2, 2, 49, 3 );
       L := New( PLabel, Init( R, '~V~ariable', LB ) );
       TDialog.Insert( L );

       R.Assign( 2, 10, 16, 11 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 2, 9, 16, 10 );
       L := New( PLabel, Init( R, '~X~ Start', I ) );
       TDialog.Insert( L );

       R.Assign( 19, 10, 33, 11 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 19, 9, 33, 10 );
       L := New( PLabel, Init( R, '~X~ End', I ) );
       TDialog.Insert( L );

       R.Assign( 36, 10, 50, 11 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 36, 9, 50, 10 );
       L := New( PLabel, Init( R, '~X~ Unit', I ) );
       TDialog.Insert( L );

       R.Assign( 2, 13, 16, 14 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 2, 12, 16, 13 );
       L := New( PLabel, Init( R, '~Y~ Start', I ) );
       TDialog.Insert( L );

       R.Assign( 19, 13, 33, 14 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 19, 12, 33, 13 );
       L := New( PLabel, Init( R, '~Y~ End', I ) );
       TDialog.Insert( L );

       R.Assign( 36, 13, 50, 14 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 36, 12, 50, 13 );
       L := New( PLabel, Init( R, '~Y~ Unit', I ) );
       TDialog.Insert( L );

       R.Assign( 2, 16, 16, 17 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 2, 15, 16, 16 );
       L := New( PLabel, Init( R, '~P~rimitive Y', I ) );
       TDialog.Insert( L );

       R.Assign( 18, 16, 33, 18 );
       B := New( PButton, Init( R, '~O~k', cmOk, bfDefault ) );
       TDialog.Insert( B );

       R.Assign( 35, 16, 50, 18 );
       B := New( PButton, Init( R, '~C~ancel', cmCancel, bfNormal ) );
       TDialog.Insert( B );

       SelectNext( False );

  End; { Constructor }

{ -------------------------------------------------------------------- }

 Constructor TMy3DPlotDlg.Init;
 Var R : TRect;
     L : PLabel;
     B : PButton;
     I : PInputLine;
     SCR : PScrollBar;
     LB : PMyVarList;
  Begin
       R.Assign( 5, 5, 57, 25 );
       TDialog.Init( R, '3D plot range' );
       Options := Options Or ofCentered;

       R.Assign( 24, 3, 25, 8 );
       SCR := New( PScrollBar, Init( R ) );
       TDialog.Insert( SCR );
       R.Assign( 2, 3, 24, 8 );
       LB := New( PMyVarList, Init( R, 1, SCR ) );
       LB^.DontShowVal := True;
       TDialog.Insert( LB );
       R.Assign( 2, 2, 24, 3 );
       L := New( PLabel, Init( R, 'X ~V~ariable', LB ) );
       TDialog.Insert( L );

       R.Assign( 49, 3, 50, 8 );
       SCR := New( PScrollBar, Init( R ) );
       TDialog.Insert( SCR );
       R.Assign( 27, 3, 49, 8 );
       LB := New( PMyVarList, Init( R, 1, SCR ) );
       LB^.DontShowVal := True;
       TDialog.Insert( LB );
       R.Assign( 27, 2, 49, 3 );
       L := New( PLabel, Init( R, 'Y ~V~ariable', LB ) );
       TDialog.Insert( L );

       R.Assign( 2, 10, 16, 11 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 2, 9, 16, 10 );
       L := New( PLabel, Init( R, '~X~ Start', I ) );
       TDialog.Insert( L );

       R.Assign( 19, 10, 33, 11 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 19, 9, 33, 10 );
       L := New( PLabel, Init( R, '~X~ End', I ) );
       TDialog.Insert( L );

       R.Assign( 36, 10, 50, 11 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 36, 9, 50, 10 );
       L := New( PLabel, Init( R, '~X~ Interval', I ) );
       TDialog.Insert( L );

       R.Assign( 2, 13, 16, 14 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 2, 12, 16, 13 );
       L := New( PLabel, Init( R, '~Y~ Start', I ) );
       TDialog.Insert( L );

       R.Assign( 19, 13, 33, 14 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 19, 12, 33, 13 );
       L := New( PLabel, Init( R, '~Y~ End', I ) );
       TDialog.Insert( L );

       R.Assign( 36, 13, 50, 14 );
       I := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( I );
       R.Assign( 36, 12, 50, 13 );
       L := New( PLabel, Init( R, '~Y~ Interval', I ) );
       TDialog.Insert( L );

       R.Assign( 6, 16, 21, 18 );
       B := New( PButton, Init( R, '~O~k', cmOk, bfDefault ) );
       TDialog.Insert( B );

       R.Assign( 31, 16, 46, 18 );
       B := New( PButton, Init( R, '~C~ancel', cmCancel, bfNormal ) );
       TDialog.Insert( B );

       SelectNext( False );

  End; { Constructor }

{ -------------------------------------------------------------------- }

 Constructor TMyDifferentDlg.Init;
 Var R : TRect;
     L  : PLabel;
     B  : PButton;
     LB : PMyVarList;
     SC : PScrollBar;
  Begin
        R.Assign( 5, 5, 57, 17 );
        TDialog.Init( R, 'Differentiate' );
        Options := Options Or ofCentered;

        R.Assign( 49, 3, 50, 8 );
        SC := New( PScrollBar, Init( R ) );
        TDialog.Insert( SC );
        R.Assign( 2, 3, 49, 8 );
        LB := New( PMyVarList, Init( R, 1, SC ) );
        LB^.DontShowVal := False;
        TDialog.Insert( LB );
        R.Assign( 2, 2, 49, 3 );
        L := New( PLabel, Init( R, '~V~ariable', LB ) );
        TDialog.Insert( L );

        R.Assign( 3, 9, 18, 11 );
        B := New( PButton, Init( R, '~O~k', cmOk, bfDefault ) );
        TDialog.Insert( B );
        R.Assign( 34, 9, 48, 11 );
        B := New( PButton, Init( R, '~C~ancel', cmCancel, bfNormal ) );
        TDialog.Insert( B );

        LB^.Select;                   { select listbox }

  End; { Constructor }

{ -------------------------------------------------------------------- }

 Constructor TMyIntervalDlg.Init( S : String );
 Var B  : PButton;
     IL : PInputLine;
     LB : PMyVarList;
     SC : PScrollBar;
     L  : PLabel;
     R  : TRect;
  Begin
       R.Assign( 6, 5, 60, 20 );     { define dialog and subviews }
       TDialog.Init( R, S );
       Options := Options Or ofCentered;

       R.Assign( 3, 3, 17, 4 );
       IL := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( IL );
       R.Assign( 3, 2, 17, 3 );
       L := New( PLabel, Init( R, '~S~tart', IL ) );
       TDialog.Insert( L );

       R.Assign( 20, 3, 34, 4 );
       IL := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( IL );
       R.Assign( 20, 2, 34, 3 );
       L := New( PLabel, Init( R, '~E~nd', IL ) );
       TDialog.Insert( L );

       R.Assign( 37, 3, 51, 4 );
       IL := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( IL );
       R.Assign( 37, 2, 51, 3 );
       L := New( PLabel, Init( R, '~I~nterval', IL ) );
       TDialog.Insert( L );

       R.Assign( 50, 6, 51, 11 );
       SC := New( PScrollBar, Init( R ) );
       TDialog.Insert( SC );
       R.Assign( 3, 6, 50, 11 );
       LB := New( PMyVarList, Init( R, 1, SC ) );
       LB^.DontShowVal := False;
       TDialog.Insert( LB );
       R.Assign( 3, 5, 50, 6 );
       L := New( PLabel, Init( R, '~V~ariable', LB ) );
       TDialog.Insert( L );

       R.Assign( 3, 12, 18, 14 );
       B := New( PButton, Init( R, '~O~k', cmOk, bfDefault ) );
       TDialog.Insert( B );

       R.Assign( 35, 12, 50, 14 );
       B := New( PButton, Init( R, '~C~ancel', cmCancel, bfNormal ) );
       TDialog.Insert( B );

       SelectNext( False );                 { select start field }
  End; { Constructor }

{ -------------------------------------------------------------------- }

 Function TMyVarList.GetText( Item : Integer; L : Integer ) : String;
 Var P : PVariable;
     S, R : String;
     C : Integer;                         { build up text: 'var' = 'value' }
  Begin
       P := PVariable( List^.At( Item ) );
       If( Not DontShowVal ) Then
        Begin
             If( P^.Kind = aNumber ) Then
              Str( PExtended( P^.Value )^, S )
             Else
              S := 'no number';
             GetText := Copy( PString( P^.ID )^, 1, L - Length( S ) - 6 ) + ' = ' + S;
        End
       Else
        GetText := Copy( PString( P^.ID )^, 1, L - 2 );
  End; { Function }

{ -------------------------------------------------------------------- }

 Constructor TMyVarInput.Init;
 Var R  : TRect;
     IL : PInputLine;
     L  : PLabel;
     B  : PButton;
  Begin
       R.Assign( 6, 5, 70, 14 );          { define dialog and subviews }
       TDialog.Init( R, 'Define variable' );

       R.Assign( 2, 6, 45, 7 );
       IL := New( PInputLine, Init( R, ValSize ) );
       TDialog.Insert( IL );

       R.Assign( 2, 5, 45, 6 );
       L := New( PLabel, Init( R, '~V~alue of var', IL ) );
       TDialog.Insert( L );

       R.Assign( 47, 3, 60, 5 );
       B := New( PButton, Init( R, '~O~k', cmOk, bfDefault ) );
       TDialog.Insert( B );

       R.Assign( 47, 6, 60, 8 );
       B := New( PButton, Init( R, '~C~ancel', cmCancel, bfNormal ) );
       TDialog.Insert( B );

       R.Assign( 2, 3, 45, 4 );
       IL := New( PInputLine, Init( R, VarSize ) );
       TDialog.Insert( IL );

       R.Assign( 2, 2, 45, 3 );
       L := New( PLabel, Init( R, '~S~tring ID', IL ) );
       TDialog.Insert( L );

       IL^.Select;
  End; { Constructor }

{ -------------------------------------------------------------------- }

 Constructor TMyDefineDlg.Init;
 Var B   : PButton;
     D   : PDialog;
     Scr : PScrollBar;
     L   : PLabel;
     IL  : PInputLine;
     R   : TRect;
   Begin
        R.Assign( 5, 2, 75, 20 );       { define dialog and subviews }
        TDialog.Init( R, 'Defines' );
        Flags := Flags XOr wfClose;          { no close icon }
        EventMask := EventMask Or evBroadCast;  { check broadcast events too }

        R.Assign( 50, 6, 51, 16 );
        Scr := New( PScrollBar, Init( R ) );
        TDialog.Insert( Scr );

        R.Assign( 3, 6, 50, 16 );
        ListBox := New( PMyVarList, Init( R, 1, Scr ) );
        ListBox^.DontShowVal := False;
        TDialog.Insert( ListBox );

        R.Assign( 3, 5, 50, 6 );
        L := New( PLabel, Init( R, '~V~ariables', ListBox ) );
        TDialog.Insert( L );

        R.Assign( 3, 3, 67, 4 );
        IL := New( PInputLine, Init( R, SizeOf( String ) - 1 ) );
        TDialog.Insert( IL );

        R.Assign( 52, 7, 67, 9 );
        B := New( PButton, Init( R, '~O~k', cmOk, bfDefault ) );
        TDialog.Insert( B );

        R.Assign( 52, 10, 67, 12 );
        B := New( PButton, Init( R, 'Define ~V~ar', cmDefVar, bfNormal ) );
        TDialog.Insert( B );

        R.Assign( 52, 12, 67, 14 );
        B := New( PButton, Init( R, '~D~elete Var', cmDeleteVar, bfNormal ) );
        TDialog.Insert( B );

        R.Assign( 3, 2, 67, 3 );
        L := New( PLabel, Init( R, '~F~ormula', IL ) );
        TDialog.Insert( L );

        IL^.Select;          { select formula field }

    End; { Constructor }

{ -------------------------------------------------------------------- }

 Procedure TMyDefineDlg.HandleEvent( Var Event : TEvent );

  Procedure DefVar;                   { define a var using TMyVarInput dialog }
   Var
      Inp     : PMyVarInput;
      DataRec : InpDataRec;
      P       : PExtended;
      C       : Integer;
      TheVar  : PVariable;
   Begin                             { use current data if any }
        If( PMyApplication( Application )^.TheParser^.Variables^.Count > 0 ) Then
         Begin
              TheVar := PVariable( PMyApplication( Application )^.TheParser^.Variables^.At( ListBox^.Focused ) );
              DataRec.Id := PString( TheVar^.ID )^;
              If( TheVar^.Kind = aNumber ) Then
               Str( PExtended( TheVar^.Value )^, DataRec.Value )
              Else
               DataRec.Value := '0';
         End { If Then }
        Else
         Begin                     { else default data }
              DataRec.Id := 'x';
              DataRec.Value := '0';
         End; { IF Then Else }

        Inp := New( PMyVarInput, Init );
        Inp^.SetData( DataRec );

        C := 1;               { check value string for errors }

        If( Application^.ValidView( Inp ) <> Nil ) Then
         While ( ( C <> 0 ) And ( DeskTop^.ExecView( Inp ) = cmOk ) ) Do
          Begin
               Inp^.GetData( DataRec );
               P := New( PExtended );
               Val( DataRec.Value, PExtended( P )^, C );

               If( C = 0 ) Then             { if no error }
                Begin
                     PMyApplication( Application )^.TheParser^.SetVar( DataRec.Id, aNumber, P );
                     ListBox^.SetRange( PMyApplication( Application )^.TheParser^.Variables^.Count );
                     ListBox^.FocusItem( PMyApplication( Application )^.TheParser^.Variables^.Count - 1 );
                     ListBox^.Draw;
                End { If Then }
               Else
                Begin                      { else error message and again }
                     Dispose( P );
                     MessageBox( #3 +  'Invalid value entered !', Nil, mfError Or mfOkButton );
                End; { If Then Else }
          End; { If While Do }

        Dispose( Inp, Done );
   End; { Sub Procedure }

{ - - - - - - - - - - MAIN DEFINE DIALOG HANDLEEVENT  - - - - - - - - - - }

  Begin
       If( Event.What And evMessage <> 0 ) Then
        Case ( Event.Command ) Of
            cmListItemSelected,         { when var is selected or button is hit }
            cmDefVar: Begin
                           DefVar;
                           ClearEvent( Event );
                      End; { DefVar }
            cmDeleteVar: Begin                { if var available, then delete it }
                              If( PMyApplication( Application )^.TheParser^.Variables^.Count > 0 ) Then
                               Begin
                                    PMyApplication( Application )^.TheParser^.Variables^.AtFree( ListBox^.Focused );
                                    ListBox^.SetRange( ListBox^.Range - 1 );
                                    ListBox^.FocusItem( 0 );
                                    ListBox^.Draw;
                               End
                              Else
                                Begin           { else erro message }
                                    MessageBox( #3 + 'No variables to delete !', Nil, mfError Or mfOkButton );
                                End; { If Then Else }

                              ClearEvent( Event );
                         End; { DeleteVar }
        End; { Case Of }

       TDialog.HandleEvent( Event );
  End; { Procedure }

{ -------------------------------------------------------------------- }

 Procedure TMyApplication.InitMenuBar;
 Var R : TRect;                    { define new menubar }
  Begin
       GetExtent( R );
       R.B.Y := R.A.Y + 1;

       MenuBar := New( PMenuBar, Init( R, NewMenu( NewSubMenu(
                  '~D~efines', hcNocontext, NewMenu( NewItem(
                     '~N~ew defines', 'F2', kbF2, cmNewDefines, hcNoContext, NewItem(
                     '~L~oad defines', 'F3', kbF3, cmLoadDefines, hcNoContext, NewItem(
                     '~S~ave defines', 'F4', kbF4, cmSaveDefines, hcNoContext, Nil ) ) ) ), NewSubMenu(
                  '~P~lots', hcNoContext, NewMenu( NewItem(
                     '~2~D Plot', 'F5', kbF5, cmPlot2D, hcNoContext, NewItem(
                     '~3~D Plot', 'F6', kbF6, cmPlot3D, hcNoContext, NewItem(
                     'Plot ~d~erivative', 'ALT-F2', kbAltF2, cmPlotDerivative, hcNoContext, NewItem(
                     'Plot ~p~rimitive', 'ALT-F3', kbAltF3, cmPlotDifEqu, hcNoContext, Nil ) ) ) ) ), NewSubMenu(
                  '~C~alculus', hcNoContext, NewMenu( NewItem(
                     '~I~ntegrate', 'F7', kbF7, cmIntegrate, hcNoContext, NewItem(
                     '~D~ifferentiate', 'F8', kbF8, cmDifferentiate, hcNoContext, NewItem(
                     '~S~can for roots', 'F9', kbF9, cmScanRoots, hcNoContext, NewItem(
                     '~E~valuate', 'F10', kbF10, cmEvaluate, hcNoContext, Nil ) ) ) ) ), NewItem(
                  '~A~bout', '', kbAltA, cmAbout, hcNoContext, NewItem(
                  '~Q~uit', '', kbAltQ, cmQuit, hcNoContext, Nil ) ) ) ) ) ) ) );
  End; { Procedure }

{ --------------------------------------------------------------------- }

 Procedure TMyApplication.InitStatusLine;
 Var R : TRect;                { define new statusline, mine !! }
  Begin
      GetExtent( R );
      R.A.Y := R.B.Y - 1;

      StatusLine := New( PMyStatusLine, Init( R, Nil ) );
  End; { Procedure }

{ -------------------------------------------------------------------- }

 Procedure TMyApplication.SetVar( I : Integer; V : Extended );
 Var TheVar : String;
     P      : PExtended;
  Begin
       TheVar := PVariable( TheParser^.Variables^.At( I ) )^.ID^;
       P := New( PExtended );
       P^ := V;
       If( TheParser^.Error <> No_Error ) Then
        TheParser^.Reset;
       TheParser^.SetVar( TheVar, aNumber, P );
  End; { Procedure }

{ -------------------------------------------------------------------- }

 Function TMyApplication.GetVar( I : Integer ) : Extended;
 Var TheVar : PVariable;
  Begin
       TheVar := PVariable( TheParser^.Variables^.At( I ) );
       If( TheVar^.Kind = aNumber ) Then
        GetVar := Extended( TheVar^.Value^ )
       Else
        GetVar := 0;
  End; { Function }

{ ------------------------------------------------------------------- }

 Function TMyApplication.GetResult( Var P : Extended ) : Boolean;
 Var
     P2 : Pointer;
     K  : ElementKind;
     Okay : Boolean;
  Begin
       Okay := TheParser^.GetResult( P2, K );

       If( Okay ) Then
        Case( K ) Of
           aBoole: If( PBoolean( P2 )^ ) Then
                    P := 1
                   Else
                    P := -1;
           aString: P := 0;
        Else
         P := PExtended( P2 )^;
        End; { Case }

      GetResult := Okay;
  End; { Function }

{ -------------------------------------------------------------------- }

 Function TMyApplication.Differentiate( I : Integer; Var D : Extended ) : Boolean;
 Var IV   : Extended;
     R, R2 : Extended;
     OldV : Extended;
     K    : ElementKind;
     E    : Integer;
  Begin
       OldV := GetVar( I );                            { save old value }
       Differentiate := True;

       If( Not GetResult( R ) ) Then     { if function not defined ? }
        Begin
             Differentiate := False;
             Exit;
        End; { If Then }

       IV := Abs( 2 * OldV * DifferentialIV );          { calculate interval }

       If( IV < DifferentialIV ) Then                   { min interval check }
        IV := DifferentialIV;

       SetVar( I, OldV - IV / 2 );                     { set var }

       If( Not GetResult( R ) ) Then
        Begin
             E := TheParser^.Error;               { get result }
             SetVar( I, OldV );
             TheParser^.Error := E;
             Differentiate := False;
             Exit;
        End; { If Then }

       SetVar( I, OldV + IV / 2 );           { get 2nd result }

       If( Not GetResult( R2 ) ) Then
        Begin
             E := TheParser^.Error;
             SetVar( I, OldV );
             TheParser^.Error := E;
             Differentiate := False;
             Exit;
        End; { If Then }                          { calculate derivative }

       D := ( R2 - R ) / IV;
       SetVar( I, OldV );           { restore var }
  End; { Function }

{ -------------------------------------------------------------------- }

 Function TMyApplication.ItterateRoot( I : Integer; X1 : Extended; Var X2 : Extended ) : Boolean;
 Var C  : Word;
     PV : Extended;
     K  : ElementKind;
     V1, VM, M : Extended;
  Begin
       C := 1;
       ItterateRoot := False;

       While ( C <> 0 ) Do       { max of times to itterate 65535 }
        Begin
             M := ( X2 + X1 ) / 2.00000000;    { get mean of the two vars }

             SetVar( I, X1 );             { get f(x1) }
             If( Not GetResult( PV ) ) Then
              Exit;
             V1 := PV;
                                          { if converge => root found }
             If( ( M = X1 ) Or ( M = X2 ) ) Then
              Begin
                   ItterateRoot := True;
                   X2 := X1;
                   Exit;
              End; { If Then }

             SetVar( I, M );              { get f(m) }
             If( Not GetResult( PV ) ) Then
              Exit;
             VM := PV;
                                       { if root between X1 and M }
             If( V1 * VM <= 0 ) Then
              X2 := M
             Else                                                   { root must be between M and X2 }
              X1 := M;

             Inc( C );
        End; { While Do }

       X2 := M;    { if no root found after itteration => use mean as result }
  End; { Function }

{ -------------------------------------------------------------------- }

 Function TMyApplication.IntoGraphics : Boolean;
 Var Driver, Mode : Integer;
  Begin
       DoneSysError;                    { deactivate all system stuff }
       DoneEvents;
       DoneVideo;
       DoneMemory;

       Driver := DETECT;                { into graphics }
       InitGraph( Driver, Mode, '' );
       If( GraphResult <> 0 ) Then
        Begin
             InitMemory;                { error ?? }
             InitVideo;
             InitEvents;
             InitSysError;
             PMyApplication( Application )^.ReDraw;
             MessageBox( #3 + 'Can''t load graphics ...', Nil, mfError Or mfOkButton );
             IntoGraphics := False;
             Exit;
        End; { If Then }
       IntoGraphics := True;
  End; { Function }

{ -------------------------------------------------------------------- }

 Procedure TMyApplication.OutofGraphics;
  Begin
       CloseGraph;          { close graphics and restore system stuff }
       InitMemory;
       InitVideo;
       InitEvents;
       InitSysError;
       PMyApplication( Application )^.ReDraw;
  End; { Procedure }

{ -------------------------------------------------------------------- }

 Procedure TMyApplication.BuildGrPlotSystem( X, Y, X2, Y2, UX, UY : Extended; I : Integer; S1, S2 : String; Var
  L, W : Integer; Var OX, OY : LongInt );
 Var SX, SY, S    : String;
     Resolution   : Extended;     { nr pixels per unit }
     C            : Extended;
  Begin
       L := GetMaxY - TextHeight( 'M' ) - 3;   { get max vert. range }
       W := GetMaxX - 1;                       { get max horiz. range }

       OX := Round( -1 * X * W / ( X2 - X ) );    { default is not visible }
       OY := Round( L + Y * L / ( Y2 - Y ) );     { but calculate the origin }

       SetColor( DARKGRAY );                      { set color }

       Resolution := UX * W / ( X2 - X );         { draw grids }
       C := Round( X / UX );
       While( C <= Round( X2 / UX ) ) Do
        Begin
             Line( Round( OX + C * Resolution ), 0, Round( OX + C * Resolution ), L );
             C := C + 1;
        End; { While Do }

       Resolution := UY * L / ( Y2 - Y );
       C := Round( Y / UY );
       While( C <= Round( Y2 / UY ) ) Do
        Begin
             Line( 0, Round( OY - C * Resolution ), W, Round( OY - C * Resolution ) );
             C := C + 1;
        End; { While Do }

       SetColor( LIGHTGRAY );              { draw window }
       Line( 0, L, W, L );
       Line( 0, 0, W, 0 );
       Line( 0, 0, 0, L );
       Line( W, 0, W, L );
       OutTextXY( 1, L + 2, S1 );          { and message S1 }

       SetColor( LIGHTCYAN );           { display window values }

       Str( X:2:1, SX );
       Str( Y2:2:1, SY );
       OutTextXY( 0, 0, '('+ SX + ',' + SY + ')' );

       Str( X:2:1, SX );
       Str( Y:2:1, SY );
       OutTextXY( 0, L - TextHeight( 'M' ), '('+ SX + ',' + SY + ')' );

       Str( X2:2:1, SX );
       Str( Y:2:1, SY );
       S := '(' + SX + ',' + SY + ')';
       OutTextXY( W - TextWidth( S ) - 1, L - TextHeight( S ), S );

       Str( X2:2:1, SX );
       Str( Y2:2:1, SY );
       S := '(' + SX + ',' + SY + ')';
       OutTextXY( W - TextWidth( S ) - 1, 0, S );

       SetColor( YELLOW );                      { show origin }
       OutTextXY( OX + 1, OY + 1, 'O' );

       If( ( X < 0 ) And ( X2 > 0 ) ) Then                { plot y axis }
        Begin                                       { if visible }
             Line( OX, 0, OX, L );
             OutTextXY( OX + 2, 0, S2 );            { plot function name }

             Resolution := UY * L / ( Y2 - Y );      { plot unit markers }
             C := Round( Y / UY );
             While( C <= Round( Y2 / UY ) ) Do
              Begin
                   Line( OX - 3, Round( OY - C * Resolution ), OX + 3, Round( OY - C * Resolution ) );
                   C := C + 1;
              End; { While Do }

        End; { If Then }

       If( ( Y < 0 ) And ( Y2 > 0 ) ) Then                { plot x axis }
        Begin                                    { if visible }
             Line( 0, OY, W, OY );                { plot var string }
             OutTextXY( W - TextWidth( PVariable( TheParser^.Variables^.At( I ) )^.ID^  ) - 1, OY + 2,
                        PVariable( TheParser^.Variables^.At( I ) )^.ID^ );

             Resolution := UX * W / ( X2 - X );        { plot unit markers }
             C := Round( X / UX );
             While( C <= Round( X2 / UX ) ) Do
              Begin
                   Line( Round( OX + C * Resolution ), OY + 3, Round( OX + C * Resolution ), OY - 3 );
                   C := C + 1;
              End; { While Do }

        End; { If Then }

       SetViewPort( 0, 0, W, L, True );             { set active window }
  End; { Procedure }

{ ------------------------------------------------------------------------- }

 Procedure TMyApplication.Plot2DFunc( X, Y, X2, Y2, UX, UY : Extended; I : Integer );
 Var L, W         : Integer;
     OX, OY       : LongInt;
     Resolution   : Extended;     { nr pixels per unit }
     C, OldV, V   : Extended;
     PV           : Extended;
     K            : ElementKind;
     S            : String;
  Begin
       OldV := GetVar( I );

       If( Not IntoGraphics ) Then
        Exit;                            { build up screen and return some values }

       BuildGrPlotSystem( X, Y, X2, Y2, UX, UY, I,
                          'Plotting press ESC to stop ...',
                          'f(' + PVariable( TheParser^.Variables^.At( I ) )^.ID^ + ')',
                          L, W, OX, OY );

       C := 0;                        { plot function }
       Resolution := L / ( Y2 - Y );

       While( C <= W ) Do
        Begin
             If( KeyPressed ) Then
              If( ReadKey = #27 ) Then
               C := W;

             SetVar( I, X + ( X2 - X ) / W * C );     { set var }

             If( Not GetResult( PV ) ) Then     { get result }
              Begin
                   If( ( TheParser^.Error <> Memory_Error ) And
                       ( TheParser^.Error <> Major_Internal_Error ) And
                       ( TheParser^.Error <> List_Eval_Error ) ) Then
                    TheParser^.Error := 0
                   Else
                    TheParser^.Reset;
                   TheParser^.ReEvaluate := True;
                   V := -1;
              End { If Then }
             Else
              If( ( PV <= Y2 ) And ( PV >= Y ) ) Then
               V := ( OY - PV * Resolution )
              Else
               V := -1;
                                                            { if visible, then plot point }
             If( Round( V ) >= 0 ) Then
              PutPixel( Round( C ), Round( V ), WHITE );

             C := C + NrPlotPixels;          { next pixel(s) }
        End; { While Do }

       SetVar( I, OldV );                     { restore var }
                                              { display final message }
       SetViewPort( 0, L + 1, W, GetMaxY, True );
       ClearViewPort;
       S := 'f(' + PVariable( TheParser^.Variables^.At( I ) )^.ID^ + ') = ' + TheParser^.Formula^;
       SetColor( LIGHTGRAY );
       OutTextXY( 0, 1, S );
       S := 'Press any key to return!';
       OutTextXY( W - TextWidth( S ) - 1, 1, S );

       While( KeyPressed ) Do              { clear key buffer }
        ReadKey;
       ReadKey;                            { wait for key }

       OutofGraphics;                      { back to tvision }
  End; { Procedure }

{ -------------------------------------------------------------------- }

 Procedure TMyApplication.Plot3DFunc( R, X, Y, X2, Y2, XI, YI : Extended; I1, I2 : Integer );
 Var Point : PGrPoint;
     C1, C2 : Extended;
     PV    : Extended;
     K     : ElementKind;
     S     : String;
     L, W  : Integer;
     Status : TGrStatus;
     M     : PMyMessageDlg;
     AWire : PGrOpenWire;
     AFrame : PGrOpenWireFrame;
     The3DSystem : PGrSystem;
     OldV1, OldV2 : Extended;
  Begin
       The3DSystem := New( PGrSystem, Init( 50, 50, 50, YELLOW ) );

       If( The3DSystem = Nil ) Then
        Begin
            MessageBox( #3 + 'Not enough memory to start 3D ...', Nil, mfError Or mfOkButton );
            Exit;
        End; { If Then }

       The3DSystem^.GetStatus( Status );
       Status.Visible := False;
       The3DSystem^.SetStatus( Status );
       AFrame := New( PGrOpenWireFrame, Init( LIGHTGREEN ) );
       The3DSystem^.Insert( 0, AFrame );

       M := New( PMyMessageDlg, Init( 'Creating 3D system, press ESC to stop ...' ) );
       Desktop^.Insert( M );

       OldV1 := GetVar( I1 );
       OldV2 := GetVar( I2 );

       C1 := X;
       While( C1 < X2 ) Do
        Begin
             SetVar( I1, C1 );
             AWire := New( PGrOpenWire, Init( LIGHTGREEN ) );
             AFrame^.Insert( 0, AWire );

             C2 := Y;
             While( C2 < Y2 ) Do
              Begin
                   If( MemAvail < MinMemFor3D ) Then
                    Begin
                         DeskTop^.Delete( M );
                         Dispose( M, Done );
                         Dispose( The3DSystem, Done );
                         SetVar( I1, OldV1 );
                         SetVar( I2, OldV2 );
                         MessageBox( #3 + 'Not enough memory to create 3D ...', Nil, mfError Or mfOkButton );
                         Exit;
                    End; { If Then }

                   If( ( KeyPressed ) And ( ReadKey = #27 ) ) Then
                    Begin
                         DeskTop^.Delete( M );
                         Dispose( M, Done );
                         Dispose( The3DSystem, Done );
                         SetVar( I1, OldV1 );
                         SetVar( I2, OldV2 );
                         MessageBox( #3 + '3D plot aborted ...', Nil, mfInformation Or mfOkButton );
                         Exit;
                    End
                   Else
                    M^.HandleMouseEvents;

                   SetVar( I2, C2 );

                   If( Not GetResult( PV ) ) Then
                    Begin
                         If( ( TheParser^.Error <> Memory_Error ) And
                             ( TheParser^.Error <> Major_Internal_Error ) And
                             ( TheParser^.Error <> List_Eval_Error ) ) Then
                          TheParser^.Error := 0
                         Else
                          TheParser^.Reset;
                         TheParser^.ReEvaluate := True;
                    End    { If Then }
                   Else
                    If( ( C1 * R < MaxInt - 1 ) And ( C2 * R < MaxInt - 1 )
                        And ( PV * R < MaxInt - 10 ) ) Then
                     Begin
                          Point := New( PGrPoint, Init( Round( C1 * R ), Round( C2 * R ),
                                        Round( PV * R ), WHITE ) );
                          AWire^.Insert( 0, Point );
                     End;   { If If Then Else }

                   C2 := C2 + YI;
              End; { While Do }
             C1 := C1 + XI;
        End; { While Do }

       Desktop^.Delete( M );
       Dispose( M, Done );

       SetVar( I1, OldV1 );
       SetVar( I2, OldV2 );

       If( Not IntoGraphics ) Then
        Begin
            Dispose( The3DSystem, Done );
            Exit;
        End;

       L := GetMaxY - TextHeight( 'M' ) - 3;
       W := GetMaxX;
       SetColor( LIGHTGRAY );
       Line( 0, L + 1, W, L + 1 );
       OutTextXY( 0, L + 2, 'Use cursor keys to rotate, press ESC to return ...' );
       SetViewPort( 0, 0, W, L, True );

       Status.Visible := True;
       Status.RelX := W Div 2;
       Status.RelY := L Div 2;
       The3DSystem^.SetStatus( Status );

       The3DSystem^.KeybRotate( 5 );

       Status.Visible := False;
       The3DSystem^.SetStatus( Status );

       OutofGraphics;

       Dispose( The3DSystem, Done );
  End; { Procedure }

{ -------------------------------------------------------------------- }

 Procedure TMyApplication.PlotDifEqu( FY, X, Y, X2, Y2, UX, UY : Extended; I : Integer );
 Var L, W         : Integer;
     OX, OY       : LongInt;
     Resolution   : Extended;     { nr pixels per unit }
     C, OldV, V, T : Extended;
     PV           : Extended;
     K            : ElementKind;
     S            : String;
  Begin
       OldV := GetVar( I );

       If( Not IntoGraphics ) Then
        Exit;

       BuildGrPlotSystem( X, Y, X2, Y2, UX, UY, I,
                          'Plotting press ESC to stop ...',
                          'F(' + PVariable( TheParser^.Variables^.At( I ) )^.ID^ + ')',
                          L, W, OX, OY );


      Resolution := L / ( Y2 - Y );

      C := W / ( X2 - X ) * ( OldV - X );

      While( C <= W ) Do
       Begin
             If( KeyPressed ) Then
              If( ReadKey = #27 ) Then
               C := W;

             SetVar( I, X + ( X2 - X ) / W * C );     { set var }

             If( Not GetResult( PV ) ) Then     { get result }
              Begin
                   If( ( TheParser^.Error <> Memory_Error ) And
                       ( TheParser^.Error <> Major_Internal_Error ) And
                       ( TheParser^.Error <> List_Eval_Error ) ) Then
                    TheParser^.Error := 0
                   Else
                    TheParser^.Reset;
                   TheParser^.ReEvaluate := True;
                   V := -1;
              End { If Then }
             Else
              Begin
                   FY := FY + PV * NrPlotPixels * ( X2 - X ) / W;
                   V := ( OY - FY * Resolution );
                   If( ( V < 0 ) Or ( V > L ) ) Then
                    V := -1;
              End; { If Then Else }
                                                            { if visible, then plot point }
             If( Round( V ) >= 0  ) Then
              PutPixel( Round( C ), Round( V ), WHITE );

             C := C + NrPlotPixels;          { next pixel(s) }
       End; { While Do }


       SetVar( I, OldV );

       SetViewPort( 0, L + 1, W, GetMaxY, True );
       ClearViewPort;
       S := 'F''(' + PVariable( TheParser^.Variables^.At( I ) )^.ID^ + ') = ' + TheParser^.Formula^;
       SetColor( LIGHTGRAY );
       OutTextXY( 0, 1, S );
       S := 'Press any key to return!';
       OutTextXY( W - TextWidth( S ) - 1, 1, S );

       While( KeyPressed ) Do
        ReadKey;
       ReadKey;

       OutofGraphics;
  End; { Procedure }

{ -------------------------------------------------------------------- }

 Procedure TMyApplication.PlotDerivative( X, Y, X2, Y2, UX, UY : Extended; I : Integer );
 Var L, W         : Integer;
     OX, OY       : LongInt;
     Resolution   : Extended;     { nr pixels per unit }
     C, OldV, V   : Extended;
     S            : String;
  Begin
       OldV := GetVar( I );

       If( Not IntoGraphics ) Then
        Exit;

       BuildGrPlotSystem( X, Y, X2, Y2, UX, UY, I,
                          'Plotting press ESC to stop ...',
                          'f(' + PVariable( TheParser^.Variables^.At( I ) )^.ID^ + ')',
                          L, W, OX, OY );

       C := 0;                        { plot function }
       Resolution := L / ( Y2 - Y );

       While( C <= W ) Do
        Begin
             If( KeyPressed ) Then
              If( ReadKey = #27 ) Then
               C := W;

             SetVar( I, X + ( X2 - X ) / W * C );

             If( Not Differentiate( I, V ) ) Then
              Begin
                   If( ( TheParser^.Error <> Memory_Error ) And
                       ( TheParser^.Error <> Major_Internal_Error ) And
                       ( TheParser^.Error <> List_Eval_Error ) ) Then
                    TheParser^.Error := 0
                   Else
                    TheParser^.Reset;
                   TheParser^.ReEvaluate := True;
                   V := -1;
              End { If Then }
             Else
              If( ( V  <= Y2 ) And ( V >= Y ) ) Then
               V := OY - V * Resolution
              Else
               V := -1;

             If( Round( V ) >= 0 ) Then
              PutPixel( Round( C ), Round( V ), WHITE );

             C := C + NrPlotPixels;
        End; { While Do }

       SetVar( I, OldV );

       SetViewPort( 0, L + 1, W, GetMaxY, True );
       ClearViewPort;
       S := 'f(' + PVariable( TheParser^.Variables^.At( I ) )^.ID^ + ') = [ ' + TheParser^.Formula^ + ' ]''';
       SetColor( LIGHTGRAY );
       OutTextXY( 0, 1, S );
       S := 'Press any key to return!';
       OutTextXY( W - TextWidth( S ) - 1, 1, S );

       While( KeyPressed ) Do
        ReadKey;
       ReadKey;

       OutofGraphics;
  End; { Procedure }

{ -------------------------------------------------------------------- }

 Procedure TMyApplication.HandleEvent( Var Event : TEvent );

 Function GetVal( ErrorStr, S : String; Var V : Extended ) : Boolean;
 Var C : Integer;
  Begin
       Val( S, V, C );
       If( C <> 0 ) Then
        Begin
             MessageBox( #3 + ErrorStr, Nil, mfError Or mfOkButton );
             GetVal := False;
             Exit;
        End
       Else
        GetVal := True;
  End; { Sub Function }

 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }

  Procedure DoDifferentiate;
  Var Dialog : PMyDifferentDlg;
      Data   : DifferentDataRec;
      Result : Word;
      C      : Integer;
      RS     : String;
      R      : Extended;
   Begin
         Dialog := New( PMyDifferentDlg, Init );

         Data.List := TheParser^.Variables;
         Data.Focus := 0;
         Dialog^.SetData( Data );

         Result := cmCancel;

         If( ValidView( Dialog ) <> Nil ) Then
          Result := Desktop^.ExecView( Dialog );

         If( Result <> cmCancel ) Then      { if not canceled }
          Begin
               Dialog^.GetData( Data );
               Dispose( Dialog, Done );

               If( Not Differentiate( Data.Focus, R ) ) Then
                Begin
                     MessageBox( #3 + 'Differential error:' + #13 + #13 + #3 + TheParser^.GetErrorStr,
                                 Nil, mfError Or mfOkButton );
                     TheParser^.Reset;
                     Exit;
                End; { If Then }

               Str( R, RS );
               MessageBox( #3 + 'Differential result:' + #13 + #13 + #3 + RS, Nil, mfInformation Or mfOkButton );
          End { If Then }
         Else
          Dispose( Dialog, Done );

   End; { Sub Procedure }

 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }

   Procedure FindRoots;
   Var Dialog : PMyIntervalDlg;         { interval dialog }
       Result : Word;                   { the result returned from execview }
       Data   : IntervDataRec;          { data record for interval dialog }
       PV     : Extended;                { pointer to the value of a variable }
       SV     : String;
       V, R1, R2,
       StrtV,                           { StartValue, EndValue, IntervalValue }
       EndV,
       IntrV  : Extended;
       OldV   : Extended;               { the old value of the variable }
       FirstScan : Boolean;
       NrRoots : LongInt;
       K      : ElementKind;            { kind of result from the parser }
       M      : PMyMessageDlg;          { my message view }
    Begin
         Data.StartV := '-10';             { default data }
         Data.EndV   := '10';
         Data.Interv := '0.001';
         Data.List   := TheParser^.Variables;
         Data.Focus  := 0;

         Dialog := New( PMyIntervalDlg, Init( 'Find roots' ) );

         Dialog^.SetData( Data );      { set/get interval dialog data }

         Result := cmCancel;

         If( ValidView( Dialog ) <> Nil ) Then
          Result := Desktop^.ExecView( Dialog );

         If( Result <> cmCancel ) Then       { if ok }
          Begin
               Dialog^.GetData( Data );      { check fields }
               Dispose( Dialog, Done );

               If( Not GetVal( 'Invalid start value !', Data.StartV, StrtV ) ) Then
                Exit;

               If( Not GetVal( 'Invalid end value !', Data.EndV, EndV ) ) Then
                Exit;

               If( Not GetVal( 'Invalid interval value !', Data.Interv, IntrV ) ) Then
                Exit;

               If( ( IntrV <= 0 ) Or ( Abs( EndV - StrtV ) / IntrV < MinInterval ) ) Then
                Begin
                     MessageBox( #3 + 'Invalid interval !', Nil, mfError Or mfOkButton );
                     Exit;
                End; { If Then }

               If( EndV < StrtV ) Then      { swap end and start so they're in order }
                Begin
                     R1 := StrtV;
                     StrtV := EndV;
                     EndV := R1;
                End; { If Then }
                                                { define message view }
               M := New( PMyMessageDlg, Init( #3 + 'Scanning for roots, press ESC to stop ...' ) );
               DeskTop^.Insert( M );

                                          { set var value }
               OldV := GetVar( Data.Focus );
               V := StrtV;
               NrRoots := 0;
               FirstScan := True;

               While( V < EndV ) Do           { continue rest of interval range }
                Begin
                     If( KeyPressed ) Then     { abort on ESC hit }
                      If( ReadKey = #27 ) Then
                       Begin
                            DeskTop^.Delete( M );
                            Dispose( M, Done );
                            Str( V, SV );
                            MessageBox( #3 + 'Scanning aborted at:' + #13 + #13 + #3 + SV,
                                        Nil, mfInformation Or mfOkButton );
                            SetVar( Data.Focus, OldV );
                            Exit;
                       End; { If If Then }
                  
                     M^.HandleMouseEvents;      { to handle message views mouse events }

                     SetVar( Data.Focus, V );

                     If( Not GetResult( PV ) ) Then
                      Begin
                           If( ( TheParser^.Error <> Memory_Error ) And
                               ( TheParser^.Error <> Major_Internal_Error ) And
                               ( TheParser^.Error <> List_Eval_Error ) ) Then
                            TheParser^.Error := 0
                           Else
                            TheParser^.Reset;
                           TheParser^.ReEvaluate := True;
                           FirstScan := True;
                      End { If Then }
                     Else
                      Begin
                           R2 := PV;        { remember result }

                           If( ( Not FirstScan ) And ( R1 * R2 < 0 ) )  Then
                            If( Not ItterateRoot( Data.Focus, V - IntrV, V ) ) Then
                             Begin
                                  If( TheParser^.GetError < 0 ) Then
                                   TheParser^.Reset;

                                  M^.SetState( sfSelected, False );
                                  Str( V, SV );
                                  MessageBox( #3 + 'Change of sign found at:' + #13 + #13 + #3 + SV,
                                              Nil, mfInformation Or mfOkButton );
                                  NrRoots := NrRoots + 1;
                                  M^.SetState( sfSelected, True );
                                  FirstScan := True;
                                  R2 := 0.00;
                             End { If If Then }
                            Else
                             Begin
                                  M^.SetState( sfSelected, False );
                                  Str( V, SV );
                                  MessageBox( #3 + 'Root found at:' + #13 + #13 + #3 + SV,
                                              Nil, mfInformation Or mfOkButton );
                                  NrRoots := NrRoots + 1;
                                  M^.SetState( sfSelected, True );
                                  FirstScan := True;
                                  R2 := 0.00;
                             End; { If If Then Else }

                           R1 := R2;
                           FirstScan := False;
                      End; { If Then Else }

                     V := V + IntrV;                { next step }
                End; { While Do }

               DeskTop^.Delete( M );            { remove view }
               Dispose( M, Done );
               SetVar( Data.Focus, OldV );
               MessageBox( #3 + 'Number of possible roots found:' + #13 + #13 + #3 + '%d', @NrRoots,
                           mfInformation Or mfOkButton );
          End { If Then }
         Else
          Dispose( Dialog, Done );         { if cancel hit, quit }

    End; { Sub Procedure }

 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }

   Procedure Integrate;                 { called when integrate command selected }
   Var Dialog : PMyIntervalDlg;         { interval dialog }
       Result : Word;                   { the result returned from execview }
       Data   : IntervDataRec;          { data record for interval dialog }
       PV     : Extended;                { pointer to e value of a variable }
       SI     : String;                 { result of integration in string form }
       V, Accum,                        { Variable, Integral, Delta Var, Rest }
       Delta,
       StrtV, IntrV,                    { StartValue, EndValue, # of Intervals }
       EndV   : Extended;
       N : LongInt;
       OldV   : Extended;               { the old value of the variable }
       C      : Integer;                { code argument for val procedure }
       K      : ElementKind;            { kind of result from the parser }
       M      : PMyMessageDlg;          { my message view }
    Begin
         Data.StartV := '0';             { default data }
         Data.EndV   := '0';
         Data.Interv := '100';
         Data.List   := TheParser^.Variables;
         Data.Focus  := 0;

         Dialog := New( PMyIntervalDlg, Init( 'Integrate' ) );

         Dialog^.SetData( Data );      { set/get interval dialog data }

         Result := cmCancel;

         If( ValidView( Dialog ) <> Nil ) Then
          Result := Desktop^.ExecView( Dialog );

         If( Result <> cmCancel ) Then       { if ok }
          Begin
               Dialog^.GetData( Data );      { check fields }
               Dispose( Dialog, Done );

               If( Not GetVal( 'Invalid number of intervals value !', Data.Interv, IntrV ) ) Then
                Exit;

               If( Not GetVal( 'Invalid start value !', Data.StartV, StrtV ) ) Then
                Exit;

               If( Not GetVal( 'Invalid end value !', Data.EndV, EndV ) ) Then
                Exit;

               If( EndV = StrtV ) Then
                Begin
                     MessageBox( #3 + 'Start and end value the same !', Nil, mfError Or mfOkButton );
                     Exit;
                End; { If Then }

               If( EndV < StrtV ) Then      { swap end and start so they're in order }
                Begin
                     Accum := StrtV;
                     StrtV := EndV;
                     EndV := Accum;
                End; { If Then }

               If( ( IntrV < 1 ) And ( Intrv > 0 ) ) Then
                IntrV := Round( ( EndV - StrtV ) / IntrV )
               Else
                IntrV := Round( IntrV );

               If( IntrV < MinInterval ) Then
                Begin
                     MessageBox( #3 + 'Invalid number of intervals !', Nil, mfError Or mfOkButton );
                     Exit;
                End; { If Then }

                                                { define message view }
               M := New( PMyMessageDlg, Init( #3 + 'Integrating, press ESC to stop ...' ) );
               DeskTop^.Insert( M );

               Delta := ( EndV - StrtV ) / IntrV;
               OldV := GetVar( Data.Focus );  { save old var value }
               SetVar( Data.Focus, StrtV );   { get f(start) }
               If( GetResult( PV ) ) Then
                Begin            { into Rest }
                     Accum := PV;
                     SetVar( Data.Focus, EndV );    { get f(end) }
                End; { If Then }
               If( Not GetResult( PV ) ) Then
                Begin                     { check errors }
                     DeskTop^.Delete( M );
                     Dispose( M, Done );
                     MessageBox( #3 + 'Integrate error:' + #13 + #13 +
                                 #3 + TheParser^.GetErrorStr, Nil, mfError Or mfOkButton );
                     SetVar( Data.Focus, OldV );
                     Exit;
                End;

               Accum := ( Accum + PV ) / 2.0; { calc rest part }

               For N := Round( IntrV - 1 ) DownTo 1 Do      { calculate accum part }
                Begin
                     If( KeyPressed ) Then     { abort on ESC hit }
                      If( ReadKey = #27 ) Then
                       Begin
                            DeskTop^.Delete( M );
                            Dispose( M, Done );
                            Str( N, SI );
                            MessageBox( #3 + 'Integration aborted at:' + #13 + #13 + #3 + SI,
                                        Nil, mfInformation Or mfOkButton );
                            SetVar( Data.Focus, OldV );
                            Exit;
                       End; { If If Then }

                     If( N Mod 50 = 0 ) Then
                      M^.HandleMouseEvents;      { to handle message views mouse events }

                     SetVar( Data.Focus, ( StrtV + Delta * N ) );
                     If( Not GetResult( PV ) ) Then
                      Begin                       { check on error }
                           DeskTop^.Delete( M );
                           Dispose( M, Done );
                           MessageBox( #3 + 'Integrate error:' + #13 + #13 +
                                       #3 + TheParser^.GetErrorStr, Nil, mfError Or mfOkButton );
                           SetVar( Data.Focus, OldV );
                           Exit;
                      End;

                     Accum := Accum + PV;  { add to accumulate part }
                End; { For Do }

               DeskTop^.Delete( M );            { remove view }
               Dispose( M, Done );

               Accum := Delta * Accum;    { calculate the true integral }
               Str( Accum, SI );                     { display result }

               MessageBox( #3 + 'Result of integration:' + #13 + #13 + #3 + SI, Nil, mfInformation Or mfOkButton );
               SetVar( Data.Focus, OldV );
          End { If Then }
         Else
          Dispose( Dialog, Done );         { if cancel hit, quit }

    End; { Sub Procedure }

 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }

   Procedure NewDefines;                 { get new defines }
   Var D    : PMyDefineDlg;
       Data : DefDataRec;
       P    : Pointer;
       K    : ElementKind;
       Okay : Boolean;
    Begin
          D := New( PMyDefineDlg, Init );
                                                { set defines data }
          Data.List := TheParser^.Variables;
          Data.Focus := 0;
          Data.Formula := PString( TheParser^.Formula )^;

          D^.SetData( Data );
          Okay := False;

          If( ValidView( D ) <> Nil ) Then
           While ( Not Okay ) Do
            Begin
                 TheParser^.Reset;

                 DeskTop^.ExecView( D );     { get data }
                 D^.GetData( Data );
                                             { check data }
                 TheParser^.SetFormula( Data.Formula );
                 TheParser^.BuildUpList;         { error => again }

                 If( TheParser^.GetError < 0  ) Then
                  Begin
                       MessageBox( #3 + TheParser^.GetErrorStr, Nil, mfError Or mfOkButton );
                       Okay := False;
                  End
                 Else
                  Okay := True;
            End; { If While Do }

          Dispose( D, Done );           { activate corresponding commands }

          DisableCommands( [ cmEvaluate, cmSaveDefines,
                             cmPlot2D, cmPlot3D, cmIntegrate,
                             cmDifferentiate, cmScanRoots, cmPlotDifEqu,
                             cmPlotDerivative ] );

          EnableCommands( [ cmEvaluate, cmSaveDefines ] );

          If( TheParser^.Variables^.Count > 0 ) Then
           EnableCommands( [ cmPlot2D, cmIntegrate, cmDifferentiate, cmScanRoots,
                            cmPlotDifEqu, cmPlotDerivative ] );

          If( TheParser^.Variables^.Count > 1 ) Then
           EnableCommands( [ cmPlot3D ] );
                                                { set new status message to formula }
          PMyStatusLine( StatusLine )^.MessageStr( NewStr( 'Formula: ' + PString( TheParser^.Formula )^ ) );

    End; { Sub Procedure }

 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }

  Procedure Evaluate;                 { evaluate formula }
  Var  R  : Extended;
       RS : String;
   Begin
        If( GetResult( R ) ) Then
         Begin
              Str( R, RS );
              MessageBox( #3 + 'Result of evaluation:' + #13+ #13 + #3 + RS, Nil, mfInformation Or mfOkButton );
         End { If Then }
        Else
         Begin
              MessageBox( #3 + 'Evaluation error:' + #13 + #13 + #3 + TheParser^.GetErrorStr, Nil, mfError Or mfOkButton );
              TheParser^.Reset;
         End; { If Then Else }
   End; { Sub Procedure }

 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }

   Procedure LoadDefines;             { load defines from math file }
   Var AStream : PDosStream;
       FName   : PathStr;
       Dialog  : PFileDialog;
       Return  : Word;
       Buffer  : String;
    Begin
        Dialog := New( PFileDialog, Init( Extention, 'Select file to load', '', fdOkButton, HistId ) );

        Return := cmCancel;

        If( ValidView( Dialog ) <> Nil ) Then
         Return := DeskTop^.ExecView( Dialog );      { get filename }

        Dialog^.GetFileName( FName );
        Dispose( Dialog, Done );

        If( Return <> cmCancel ) Then           { if not canceled }
         Begin                                  { load and check file }
              AStream := New( PDosStream, Init( FName, stOpenRead ) );

              If( AStream^.ErrorInfo <> stOk ) Then
               Begin
                    MessageBox( #3 + 'Error loading the file ...', Nil, mfError Or mfOkButton );
                    Dispose( AStream, Done );
                    Exit;
               End; { If Then }

              AStream^.Read( Buffer[ 1 ], Length( FileId ) );
              Buffer[ 0 ] := Char( Length( FileId ) );

              If( Buffer <> FileId ) Then
               Begin
                    MessageBox( #3 + 'File is not a MATH file ...', Nil, mfError Or mfOkButton );
                    Dispose( AStream, Done );
                    Exit;
               End; { If Then }

              Dispose( TheParser, Done );
              TheParser := PStackParser( AStream^.Get );
              TheParser^.Reset;

              Dispose( AStream, Done );
                                              { update statusline message }
              PMyStatusLine( StatusLine )^.MessageStr( NewStr( 'Formula: ' + TheParser^.Formula^ ) );
                                               { activate corresponing commands }
              DisableCommands( [ cmEvaluate, cmSaveDefines,
                          cmPlot2D, cmPlot3D, cmIntegrate,
                          cmDifferentiate, cmScanRoots, cmPlotDifEqu,
                          cmPlotDerivative ] );

              EnableCommands( [ cmEvaluate, cmSaveDefines ] );

              If( TheParser^.Variables^.Count > 0 ) Then
              EnableCommands( [ cmPlot2D, cmIntegrate, cmDifferentiate, cmScanRoots,
                                cmPlotDifEqu, cmPlotDerivative ] );

              If( TheParser^.Variables^.Count > 1 ) Then
               EnableCommands( [ cmPlot3D ] );
         End; { If Then }

    End; { Sub Procedure }

 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }

  Procedure SaveDefines;            { save defines to a math file }
  Var AStream : PDosStream;
      FName   : PathStr;
      Dialog  : PFileDialog;
      Return  : Word;
      Buffer  : String;
   Begin
        Dialog := New( PFileDialog, Init( Extention, 'Select file to save', '', fdOkButton, HistId ) );

        Return := cmCancel;

        If( ValidView( Dialog ) <> Nil ) Then
         Return := DeskTop^.ExecView( Dialog ); { get new filename }

        Dialog^.GetFileName( FName );
        Dispose( Dialog, Done );

        If( Return <> cmCancel ) Then       { if not canceled, save shit }
         Begin
              AStream := New( PDosStream, Init( FName, stCreate ) );

              Buffer := FileId;
              AStream^.Write( Buffer[ 1 ], Length( FileId ) );
              AStream^.Put( TheParser );

              Dispose( AStream, Done );
         End; { If Then }

   End; { Sub Procedure }

 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }

  Procedure PlotFunction;
  Var Dialog : PMyPlotDlg;
      Data   : PlotRangeDataRec;
      Result : Word;
      XS, XE, XU,
      YS, YE, YU, T : Extended;
   Begin
        Data.List := TheParser^.Variables;
        Data.Item := 0;
        Data.XS := '-10';
        Data.YS := '-10';
        Data.XE := '10';
        Data.YE := '10';
        Data.XU := '1';
        Data.YU := '1';

        Dialog := New( PMyPlotDlg, Init( 'Plot range for 2D plot' ) );
        Dialog^.SetData( Data );

        Result := cmCancel;

        If( ValidView( Dialog ) <> Nil ) Then
         Result := Desktop^.ExecView( Dialog );

        Dialog^.GetData( Data );
        Dispose( Dialog, Done );

        If( Result <> cmCancel ) Then
         Begin
               If( Not GetVal( 'Invalid X start !', Data.XS, XS ) ) Then
                Exit;
               If( Not GetVal( 'Invalid Y start !', Data.YS, YS ) ) Then
                Exit;
               If( Not GetVal( 'Invalid X end !', Data.XE, XE ) ) Then
                Exit;
               If( Not GetVal( 'Invalid Y end !', Data.YE, YE ) ) Then
                Exit;
               If( Not GetVal( 'Invalid X unit !', Data.XU, XU ) ) Then
                Exit;
               If( Not GetVal( 'Invalid Y unit !', Data.YU, YU ) ) Then
                Exit;

               If( XS > XE ) Then
                Begin
                     T := XS;
                     XS := XE;
                     XE := T;
                End; { If Then }

               If( YS > YE ) Then
                Begin
                     T := YS;
                     YS := YE;
                     YE := T;
                End; { If Then }

               If( ( YS = YE ) Or ( XS = XE ) Or ( XU <= 0 ) Or ( YU <= 0 ) ) Then
                Begin
                     MessageBox( #3 + 'Invalid interval ...', Nil, mfError Or mfOkButton );
                     Exit;
                End; { If Then }

               Plot2DFunc( XS, YS, XE, YE, XU, YU, Data.Item );

         End; { If Then }

   End; { Sub Procedure }

 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }

  Procedure PlotPrim;
  Var Dialog : PMyPrimPlotDlg;
      Data   : PrimPlotRangeDataRec;
      Result : Word;
      XS, XE, XU,
      YS, YE, YU,
      FY, T      : Extended;
   Begin
        Data.List := TheParser^.Variables;
        Data.Item := 0;
        Data.XS := '-10';
        Data.YS := '-10';
        Data.XE := '10';
        Data.YE := '10';
        Data.XU := '1';
        Data.YU := '1';
        Data.FY := '0';

        Dialog := New( PMyPrimPlotDlg, Init );
        Dialog^.SetData( Data );

        Result := cmCancel;

        If( ValidView( Dialog ) <> Nil ) Then
         Result := Desktop^.ExecView( Dialog );

        Dialog^.GetData( Data );
        Dispose( Dialog, Done );

        If( Result <> cmCancel ) Then
         Begin
               If( Not GetVal( 'Invalid X start !', Data.XS, XS ) ) Then
                Exit;
               If( Not GetVal( 'Invalid Y start !', Data.YS, YS ) ) Then
                Exit;
               If( Not GetVal( 'Invalid X end !', Data.XE, XE ) ) Then
                Exit;
               If( Not GetVal( 'Invalid Y end !', Data.YE, YE ) ) Then
                Exit;
               If( Not GetVal( 'Invalid X unit !', Data.XU, XU ) ) Then
                Exit;
               If( Not GetVal( 'Invalid Y unit !', Data.YU, YU ) ) Then
                Exit;
               If( Not GetVal( 'Invalid primitive value !', Data.FY, FY ) ) Then
                Exit;

               If( XS > XE ) Then
                Begin
                     T := XS;
                     XS := XE;
                     XE := T;
                End; { If Then }

               If( YS > YE ) Then
                Begin
                     T := YS;
                     YS := YE;
                     YE := T;
                End; { If Then }

               If( ( YS = YE ) Or ( XS = XE ) Or ( XU <= 0 ) Or ( YU <= 0 ) ) Then
                Begin
                     MessageBox( #3 + 'Invalid interval ...', Nil, mfError Or mfOkButton );
                     Exit;
                End; { If Then }

               PlotDifEqu( FY, XS, YS, XE, YE, XU, YU, Data.Item );

         End; { If Then }

   End; { Sub Procedure }

 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }

  Procedure PlotADerivative;
  Var Dialog : PMyPlotDlg;
      Data   : PlotRangeDataRec;
      Result : Word;
      XS, XE, XU,
      YS, YE, YU, T : Extended;
   Begin
        Data.List := TheParser^.Variables;
        Data.Item := 0;
        Data.XS := '-10';
        Data.YS := '-10';
        Data.XE := '10';
        Data.YE := '10';
        Data.XU := '1';
        Data.YU := '1';

        Dialog := New( PMyPlotDlg, Init( 'Plot range for derivative plot' ) );
        Dialog^.SetData( Data );

        Result := cmCancel;

        If( ValidView( Dialog ) <> Nil ) Then
         Result := Desktop^.ExecView( Dialog );

        Dialog^.GetData( Data );
        Dispose( Dialog, Done );

        If( Result <> cmCancel ) Then
         Begin
               If( Not GetVal( 'Invalid X start !', Data.XS, XS ) ) Then
                Exit;
               If( Not GetVal( 'Invalid Y start !', Data.YS, YS ) ) Then
                Exit;
               If( Not GetVal( 'Invalid X end !', Data.XE, XE ) ) Then
                Exit;
               If( Not GetVal( 'Invalid Y end !', Data.YE, YE ) ) Then
                Exit;
               If( Not GetVal( 'Invalid X unit !', Data.XU, XU ) ) Then
                Exit;
               If( Not GetVal( 'Invalid Y unit !', Data.YU, YU ) ) Then
                Exit;

               If( XS > XE ) Then
                Begin
                     T := XS;
                     XS := XE;
                     XE := T;
                End; { If Then }

               If( YS > YE ) Then
                Begin
                     T := YS;
                     YS := YE;
                     YE := T;
                End; { If Then }

               If( ( YS = YE ) Or ( XS = XE ) Or ( XU <= 0 ) Or ( YU <= 0 ) ) Then
                Begin
                     MessageBox( #3 + 'Invalid interval ...', Nil, mfError Or mfOkButton );
                     Exit;
                End; { If Then }

               PlotDerivative( XS, YS, XE, YE, XU, YU, Data.Item );

         End; { If Then }

   End; { Sub Procedure }

 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

  Procedure Plot3D;
  Var Dialog : PMy3DPlotDlg;
      Data   : Plot3DRangeDataRec;
      Result : Word;
      XS, XE, XI,
      YS, YE, YI, T : Extended;
   Begin
        Data.List1 := TheParser^.Variables;
        Data.Item1 := 0;
        Data.List2 := Data.List1;
        Data.Item2 := 1;
        Data.XS := '0';
        Data.YS := '0';
        Data.XE := '10';
        Data.YE := '10';
        Data.XI := '0.5';
        Data.YI := '0.5';

        Dialog := New( PMy3DPlotDlg, Init );
        Dialog^.SetData( Data );

        Result := cmCancel;

        If( ValidView( Dialog ) <> Nil ) Then
         Result := Desktop^.ExecView( Dialog );

        Dialog^.GetData( Data );
        Dispose( Dialog, Done );

        If( Result <> cmCancel ) Then
         Begin
               If( Not GetVal( 'Invalid X start !', Data.XS, XS ) ) Then
                Exit;
               If( Not GetVal( 'Invalid Y start !', Data.YS, YS ) ) Then
                Exit;
               If( Not GetVal( 'Invalid X end !', Data.XE, XE ) ) Then
                Exit;
               If( Not GetVal( 'Invalid Y end !', Data.YE, YE ) ) Then
                Exit;
               If( Not GetVal( 'Invalid X interval !', Data.XI, XI ) ) Then
                Exit;
               If( Not GetVal( 'Invalid Y interval !', Data.YI, YI ) ) Then
                Exit;

               If( XS > XE ) Then
                Begin
                     T := XS;
                     XS := XE;
                     XE := T;
                End; { If Then }

               If( YS > YE ) Then
                Begin
                     T := YS;
                     YS := YE;
                     YE := T;
                End; { If Then }

               If( ( YS = YE ) Or ( XS = XE ) Or ( XI <= 0 ) Or ( YI <= 0 ) ) Then
                Begin
                     MessageBox( #3 + 'Invalid interval ...', Nil, mfError Or mfOkButton );
                     Exit;
                End; { If Then }

               Plot3DFunc( PixelsPer3DUnit, XS, YS, XE, YE, XI, YI, Data.Item1, Data.Item2 );

         End; { If Then }

   End; { Sub Procedure }

 { - - - - - - - - - - - - MAIN APPLICATION HANDLEEVENT - - - - - - - - }

  Begin
       TApplication.HandleEvent( Event );

       If( Event.What = evCommand ) Then
        Case Event.Command Of                { check commands and run them }
             cmAbout: Begin
                           MessageBox( #3 + 'MATH.exe 2.00' + #13 + #13 +
                                       #3 + 'By: MFSomers, 2000.', Nil,
                                       mfInformation Or mfOkButton );
                           ClearEvent( Event );
                      End; { About command }
             cmNewDefines: Begin
                                NewDefines;
                                Clearevent( Event );
                           End; { New Defines command }
             cmSaveDefines: Begin
                                 SaveDefines;
                                 ClearEvent( Event );
                            End; { SaveDefines }
             cmLoadDefines: Begin
                                 LoadDefines;
                                 ClearEvent( Event );
                            End; { LoadDefines }
             cmEvaluate: Begin
                              Evaluate;
                              ClearEvent( Event );
                         End; { Evaluate }
             cmIntegrate: Begin
                               Integrate;
                               ClearEvent( Event );
                          End; { Integrate }
             cmDifferentiate: Begin
                                   DoDifferentiate;
                                   ClearEvent( Event );
                              End; { Differentiate }
             cmScanRoots: Begin
                               FindRoots;
                               ClearEvent( Event );
                          End; { Roots }
             cmPlot2D:    Begin
                               PlotFunction;
                               ClearEvent( Event );
                          End; { Plot 2D }
             cmPlotDerivative:  Begin
                                     PlotADerivative;
                                     ClearEvent( Event );
                                End; { Plot Derivative }
             cmPlot3D:    Begin
                               Plot3D;
                               ClearEvent( Event );
                          End; { Plot 3D }
             cmPlotDifEqu : Begin
                                    PlotPrim;
                                    ClearEvent( Event );
                               End; { Plot DifEqu }
        Else
            MessageBox( #3 + 'Function not implemented yet ...', Nil, mfError Or mfOkButton );
        End; { Case Of }

  End; { Procedure }

{ -------------------------------------------------------------------- }

 Procedure TMyApplication.Idle;
  Begin
       TApplication.Idle;
       MemMonitor^.UpDate;              { update memory monitor }
  End; { Procedure }

{ -------------------------------------------------------------------- }

 Constructor TMyApplication.Init;
 Var R : TRect;
     P : Pointer;
  Begin
       TApplication.Init;

       _GR3D_AuthorMessage := FALSE;           { init units and graphic stuff }
       _Parser_Author_Message := FALSE;

       If( RegisterBgiDriver( @EGAVGA_PROC ) < 0 ) Then
        Begin
             TApplication.Done;
             Writeln( 'Error registering the EGA & VGA graphic routines ...' );
             Halt( 1 );
        End;

       RegisterObjects;        { register the parser for stream use }
       RegisterParser;
                                         { disable corresponding commands }
       DisableCommands( [ cmEvaluate, cmSaveDefines,
                          cmPlot2D, cmPlot3D, cmIntegrate,
                          cmDifferentiate, cmScanRoots, cmPlotDifEqu,
                          cmPlotDerivative ] );

       GetExtent( R );            { insert memory monitor }
       R.A.X := R.B.X - 8;
       R.B.Y := R.A.Y + 1;
                                  { define memmonitor and parser }
       MemMonitor := New( PHeapView, Init( R ) );
       TheParser   := New( PStackParser, Init );

       If( ( MemMonitor = Nil ) Or ( TheParser = Nil ) ) Then
        Begin
             Done;
             Writeln( 'Not enough memory to start MATH 2.00 ...' );
             Halt( 1 );
        End; { If Then }

       TMyApplication.Insert( MemMonitor );

  End; { Constructor }

{ -------------------------------------------------------------------- }

 Destructor TMyApplication.Done;
  Begin
       If( MemMonitor <> Nil ) Then            { remove all vars }
        Dispose( MemMonitor, Done );
       If( TheParser <> Nil ) Then
        Dispose( TheParser, Done );

       TApplication.Done;
  End; { Destructor }

{ --------------------- MAIN UNIT --------------------------------- }

 Begin
 End. { Unit }
