{$X+}
{****************************************************************************}
{*  This unit was written by Elliot Brown, with much help from his friends  *}
{*  and a set of tutorials by Denthor of Asphyxia.                          *}
{*  Copyright (c), Elliot Brown, May 2001                                   *}
{****************************************************************************}

Program graphics;

uses crt,dos;

const vga = $a000;

Var
   pall,pall2 : Array [0..255,1..3] of byte;

{****************************************************************************}
{*                Sets the video mode to VGA(320x200x256)                   *}
{****************************************************************************}
Procedure SetVGA;
begin
     asm
        mov     ax,0013h
        int     10h
     end;
end;

{****************************************************************************}
{*                 Sets the video mode to text(80x30x16)                    *}
{****************************************************************************}
Procedure Settext;
begin
     asm
        mov     ax,0003h
        int     10h
     end;
end;

{****************************************************************************}
{*           This delays untill the current frame has been drawn            *}
{****************************************************************************}
procedure Waitretrace; assembler;

asm
   mov dx, 3DAh
@l1:
   in al,dx
   and al,08h
   jnz @l1;
@l2:
   in al,dx
   and al,08h
   jz @l2
end;

{****************************************************************************}
{*                   Gets RGB value of selected color                       *}
{****************************************************************************}
procedure getpal (colorNo : Byte; Var R,G,B : byte);
begin
     Port[$3c7] := colorNo;
     R := port[$3c9];
     G := port[$3c9];
     B := port[$3c9];
end;

{****************************************************************************}
{*                   Sets RGB value of selected color                       *}
{****************************************************************************}
procedure setpal (colorNo : byte; R,G,B : byte);
begin
     port[$3c8] := colorNo;
     port[$3c9] := R;
     port[$3c9] := G;
     port[$3c9] := B;
end;

{****************************************************************************}
{*                   The faster version of Putpixel                         *}
{****************************************************************************}
Procedure putpixel(x,y:integer; col:byte);
begin
     mem [vga:x+(y*320)]:=col;
end;

{****************************************************************************}
{*                     Draws a line to the screen                           *}
{****************************************************************************}
procedure line (x1,y1,x2,y2:integer; col:byte);
var
   x,y,xlength,ylength,dx,dy:integer;
   xslope,yslope:real;
begin
     xlength:=abs (x1-x2);
     if (x1-x2) < 0 then dx:=-1;
     if (x1-x2) = 0 then dx:=0;
     if (x1-x2) > 0 then dx:=+1;
     ylength:=abs (y1-y2);
     if (y1-y2) < 0 then dy:=-1;
     if (y1-y2) = 0 then dy:=0;
     if (y1-y2) > 0 then dy:=+1;
     if (dy=0) then
        begin
             if dx < 0 then for x:=x1 to x2 do
                putpixel(x,y1,col);
             if dx > 0 then for x:=x2 to x1 do
                putpixel(x,y1,col);
             exit
        end;
     if (dx=0) then
        begin
             if dy < 0 then for y:=y1 to y2 do
                putpixel (x1,y,col);
             if dy > 0 then for y:=y2 to y1 do
                putpixel (x1,y,col);
             exit;
        end;
     xslope:=xlength/ylength;
     yslope:=ylength/xlength;
     if (yslope/xslope<1) and (yslope/xslope>-1) then
        begin
             if dx < 0 then for x:=x1 to x2 do
                begin
                     y:=round (yslope*x);
                     putpixel (x,y,col);
                end;
             if dx > 0 then for x:=x2 to x1 do
                begin
                     y:=round (yslope*x);
                     putpixel (x,y,col);
                end;
        end
     ELSE
         begin
              if dy < 0 then for y:=y1 to y2 do
                 begin
                      x:=round (xslope*y);
                      putpixel (x,y,col);
                 end;
               if dy > 0 then for y:=y2 to y1 do
                 begin
                      x:=round (xslope*y);
                      putpixel (x,y,col);
                 end;
         end;
     end;

{****************************************************************************}
{*                             Grabs palette                                *}
{****************************************************************************}
procedure grabpalette;
var
   loop:integer;
begin
     for loop:= 0 to 255 do
         getpal (loop,pall2[loop,1],pall2[loop,2],pall2[loop,3]);
end;

{****************************************************************************}
{*                     Sets entire palette to black                         *}
{****************************************************************************}
procedure blackout;
var
   loop:integer;
begin
     for loop:=0 to 255 do
         setpal(loop,0,0,0);
end;

{****************************************************************************}
{*                    Prepares screen while blacked out                     *}
{****************************************************************************}
procedure screenSetup;
var
   loop,loop2   :       integer;
begin
     for loop:=0 to 319 do
         for loop2:= 0 to 199 do
             putpixel (loop,loop2,random (256));
end;

{****************************************************************************}
{*                    Fades screen out at a set speed                       *}
{****************************************************************************}
procedure fadeout(del:integer);
var
   loop,loop2:integer;
   tmp:array [1..3] of byte;
begin
     for loop:=1 to 64 do
         begin
              delay(del);
              waitretrace;
              for loop2:=0 to 255 do
                  begin
                       getpal (loop2,tmp[1],tmp[2],tmp[3]);
                       if tmp[1]>0 then dec (tmp[1]);
                       if tmp[2]>0 then dec (tmp[2]);
                       if tmp[3]>0 then dec (tmp[3]);
                       setpal (loop2,tmp[1],tmp[2],tmp[3]);
                  end;
         end;
end;

{****************************************************************************}
{*                 Restores palette saved by GrabPalette                    *}
{****************************************************************************}
procedure restorepalette;
var
   loop:integer;
begin
     waitretrace;
     for loop:=0 to 255 do
         setpal(loop,pall2[loop,1],pall2[loop,2],pall2[loop,2]);
end;

{****************************************************************************}
{*            Clears the screen in VGA mode to a selected color             *}
{****************************************************************************}
Procedure Cls (col : Byte);
begin
     Fillchar (mem [$a000:0],64000,col);
end;

{****************************************************************************}
{*                            Draws a circle                                *}
{****************************************************************************}
procedure circle(oX,oY,rad:integer; col:byte);
var
   deg:real;
   x,y:integer;
begin
     deg:=0;
     repeat
           x:=round(rad*cos (deg));
           y:=round(rad*sin (deg));
           putpixel (x+ox,y+oy,col);
           deg:=deg+0.005;
     until (deg>6.4);
end;

{****************************************************************************}
{*                            Rotates palette                               *}
{****************************************************************************}
procedure rotpal(col,intensity:byte);
var
   tmp : array [1..3] of byte;
begin
     tmp[1]:=intensity;
     tmp[2]:=0;
     tmp[3]:=0;
     repeat
           begin
                waitretrace;
                if tmp[2] < intensity then inc(tmp[2]);
                setpal (col,tmp[1],tmp[2],tmp[3]);
           end;
     until ((tmp[2] = intensity)or(keypressed));
     repeat
           begin
                waitretrace;
                if tmp[1] > 0 then dec(tmp[1]);
                setpal (col,tmp[1],tmp[2],tmp[3]);
           end;
     until ((tmp[1] = 0)or(keypressed));
     repeat
           begin
                waitretrace;
                if tmp[3] < intensity then inc(tmp[3]);
                setpal (col,tmp[1],tmp[2],tmp[3]);
           end;
     until ((tmp[3] = intensity)or(keypressed));
     repeat
           begin
                waitretrace;
                if tmp[2] > 0 then dec(tmp[2]);
                setpal (col,tmp[1],tmp[2],tmp[3]);
           end;
     until ((tmp[2] = 0)or(keypressed));
     repeat
           begin
                waitretrace;
                if tmp[1] < intensity then inc(tmp[1]);
                setpal (col,tmp[1],tmp[2],tmp[3]);
           end;
     until ((tmp[1] = intensity)or(keypressed));
     repeat
           begin
                waitretrace;
                if tmp[3] > 0 then dec(tmp[3]);
                setpal (col,tmp[1],tmp[2],tmp[3]);
           end;
     until ((tmp[3]=0)or(keypressed));
end;

{============================================================================}
{********************************/ MAIN \**********************************}
{============================================================================}
begin
     setVGA;
     cls(1);
     repeat
           rotpal(1,50);
     until keypressed;
end.