{  *------------------------------------------*
   |                 PARSER.tpu               |
   *------------------------------------------*
   |   DEBUG VERSION OF PARSER.PAS 6.00...    |
   |   If the unit is compiled with D+ then   |
   |   the internal debugging will be active  |
   |   and included into the compilation...   |
   *------------------------------------------*
   |                                          |
   |  Author:      MFSomers, 1996.            |
   |  Updates:     6th. 11 April 2000.        |
   |  Use:         To parse a string formula  |
   |                with the stack parser.    |
   |               Supports strings, booleans |
   |                numerics, constants,      |
   |                variables and lots of     |
   |                operators listed.         |
   |                                          |
   *------------------------------------------*
}

{$O-,B-,F-,A+,I-,N+,P-,Q-,R-,S-,T+,V+,G-,E+,X+}

{$IFOPT D+}
{$DEFINE INCDEBUG}
{$ENDIF}

Unit Parser;

Interface

Uses Objects {$IFDEF INCDEBUG}, CRT {$ENDIF};   { All Units }

Const
      No_Error              = 0;                  { All ok }
      Memory_Error          = -1;
      List_Eval_Error       = -12;
      Major_Internal_Error  = -43;

Type
     PExtended  = ^Extended;              { Default data type pointers }
     PBoolean   = ^Boolean;
     PCharacter = ^Char;
                                          { Kind of objects in a stack }
     ElementKind = ( aBoole, aString, aNumber, aVariable, anOperator, aCounter, anError );

     TStackElement = Object( TObject )          { The general stack or list object }
                           Kind      : ElementKind;         { kind of value }
                           Value     : Pointer;             { pointer to value }
                           FormPos   : Integer;             { position in string }
                           ObjStrLen : Integer;             { string length of object }
                           Index     : Integer;             { an index variable }

                           Constructor Init( K : ElementKind; V : Pointer; I, S, L : Integer );
                           Destructor  Done; Virtual;
                           Procedure   Store( Var S : TStream ); Virtual;
                           Constructor Load( Var S : TStream );
                     End; { Object }
     PStackElement = ^TStackElement;

     { One MUST use the TStackParser object for handleing stacks and variables !!! }
     { Else the correct error checks will fail !!! }

     TVariable = Object( TObject )                       { Variabele object type }
                       ID : PString;                     { string ID }
                       Kind : ElementKind;               { kind of value in variable }
                       Value : Pointer;                  { pointer to value }

                       Constructor Init( Var S : String; K : ElementKind; V : Pointer );
                       Destructor  Done; Virtual;
                       Procedure   Store( Var S : TStream ); Virtual;
                       Constructor Load( Var S : TStream );
                 End; { Object }
     PVariable = ^TVariable;

     TStackParser = Object( TObject )           { StackParser object definition }
                          Variables    : PCollection;         { list of all known vars }
                          RList        : PCollection;         { the compiled RList }
                          EvalStack    : PCollection;         { the evaluation stack }
                          HeapSaver    : PCollection;         { dump of non-disposed stuff }
                          Formula      : PString;             { the formula to parse }
                          Result       : Pointer;             { result of parser }
                          KindResult   : ElementKind;         { kind of result }
                          Error        : Integer;             { error flag and value }
                          ReEvaluate   : Boolean;             { check flags for speed }
                          ReBuildList  : Boolean;
                          OldExit      : Pointer;             { runtime error handleing vars }
                          _ES, _SS, _SP,
                          _DS, _BP, _FLAG,
                          _JMPSEG, _JMPOFS : Word;

                          Constructor Init;
                          Destructor  Done; Virtual;
                          Procedure   Reset; Virtual;
                          Procedure   Store( Var S : TStream ); Virtual;
                          Constructor Load( Var S : TStream );
                          Procedure   SetFormula( F : String ); Virtual;
                          Procedure   SetVar( V : String; K : ElementKind; Value : Pointer ); Virtual;
                          Procedure   RemoveVar( V : String ); Virtual;
                          Procedure   RemoveAllVars; Virtual;
                          Procedure   BuildUpList; Virtual;
                          Procedure   EvaluateList; Virtual;
                          Function    GetFormula: String; Virtual;
                          Function    GetError: Integer; Virtual;
                          Function    GetErrorStr : String; Virtual;
                          Function    GetResult( Var R: Pointer; Var K : ElementKind ): Boolean; Virtual;
                    End; { Object }
     PStackParser = ^TStackParser;

Var
   _Parser_Author_Message : Boolean;

Procedure RegisterParser;

Implementation

{ ------------- INTERNAL CONSTANTS -------------------------------- }

Const
      An_Empty_String : String[ 1 ] = ''; { An empty string }
      Stck_Size      = 15;                { Default stack size }
      Vars_Size      = 5;                 { Default variable list size }
      Str_Value_Size = 30;                { Size of a constant string value }
      ID_Size        = 12;                { Size of ID strings in  all tables }
      Op_Priority_Delta = 200;            { Delta priority level value }

{ ----------------------- INTERNAL TYPES ------------------------- }

Type OpFunc = Function( P : PStackParser ): Integer;

     OpElement = Record                      { Element of Operator table }
                       ID      : String[ ID_Size ];
                       Level   : Word;          { Priority level }
                       Grammar : Word;          { Grammar check word }
                       ParCount: Integer;       { Number of parameters if function, -1 at least 1, 0 none etc. }
                       Func    : OpFunc;        { Routine called }
                 End; { Record }

     CnstElement = Record                    { Element of Constant table }
                         ID    : String[ ID_Size ];
                         Kind  : ElementKind;       { kind of constant }
                         NumValue : Extended;       { Numerical value of constant }
                         BlnValue : Boolean;
                         StrValue : String[ Str_Value_Size ];
                   End; { Record }

{ -------------------- REGISTRATION CONSTANTS ------------------- }

Const RStackElement : TStreamRec = ( ObjType: 54224;
                                     VmtLink: Ofs(TypeOf(TStackElement)^);
                                     Load:    @TStackElement.Load;
                                     Store:   @TStackElement.Store
                                   );
      RVariable  : TStreamRec = ( ObjType: 54230;
                                  VmtLink: Ofs(TypeOf(TVariable)^);
                                  Load:    @TVariable.Load;
                                  Store:   @TVariable.Store
                                );
      RStackParser  : TStreamRec = ( ObjType: 54221;
                                     VmtLink: Ofs(TypeOf(TStackParser)^);
                                     Load:    @TStackParser.Load;
                                     Store:   @TStackParser.Store
                                   );

{ --------------------------- GLOBAL VARS FOR ERROR HANDLEING -------- }

Var     OverFlow     : Boolean;
        OldExitProc  : Pointer;          { These vars are for the overflow hook }
        ErrorCaller  : PStackParser;
        __ES, __SS, __DS,
        __BP, __SP, __FLAG,
        __JMPSEG, __JMPOFS : Word;

{ -------------------- THE RUN TIME CALCULATION FUNCTIONS ------------- }

Procedure FreeThisItem( Kind : ElementKind; P : Pointer );
 Begin
       If( P <> Nil ) Then
        Case Kind Of
             aNumber,
             anOperator,
             anError,
             aCounter  : Dispose( PExtended( P ) );
             aBoole    : Dispose( PBoolean( P ) );
             aString   : DisposeStr( PString( P ) );
             aVariable : Dispose( PVariable( P ), Done );
        End; { Case }
 End; { Function }

{ --------------------------------------------------------------------- }

 Function StrUpr( S : String ): String;
 Var
    I : Integer;
  Begin
       For I := 1 To Ord( S[ 0 ] ) Do
        S[ I ] := UpCase( S[ I ] );

       StrUpr := S;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function StrLwr( S : String ): String;
 Var
    I : Integer;
  Begin
       For I := 1 To Ord( S[ 0 ] ) Do
        If( ( S[ I ] >= 'A' ) And ( S[ I ] <= 'Z' ) ) Then
         S[ I ] := Chr( Ord( S[ I ] ) + 32 );

       StrLwr := S;
  End; { Function }

{ ---------------------------------------------------------------------- }
{ Pops an object from eval stack and puts it in heapsaver...             }

 Function PopObj( Var Obj: PStackElement; P : PStackParser; Var E : Integer ): Boolean;
  Begin
       If( P^.EvalStack^.Count <= 0 ) Then
        Begin
             PopObj := False;
             E := List_Eval_Error;                    { signal stack error }
             Obj := Nil;
        End { If Then }
       Else
        Begin
             PopObj := True;
             Obj := P^.EvalStack^.At( 0 );
             If( Obj^.Kind = anError ) Then             { if poped an error => into E }
              E := Round( Extended( Obj^.Value^ ) );
             P^.EvalStack^.AtDelete( 0 );
             P^.HeapSaver^.AtInsert( 0, Obj );  { save the obj in the heap saver }
        End; { If Then Else }                   { so it will be removed after evaluation }
  End; { Function }                             { automaticaly (speeds up) }

{ -------------------------------------------------------------------- }
{ Creates a new (empty) object and puts it in the heapsaver...         }

 Function NewObj( Var Obj: PStackElement; P : PStackParser; Kind : ElementKind; Var E : Integer ): Boolean;
  Begin
       Obj := New( PStackElement, Init( Kind, Nil, 0, 0, 0 ) );

       If( Obj = Nil ) Then
        Begin
             NewObj := False;
             E := Memory_Error;
             Exit;
        End;

       Case( Kind ) Of                     { allocate mem for specified type }
           aBoole:      Obj^.Value := New( PBoolean );
           aNumber,
           anError,
           aCounter:    Obj^.Value := New( PExtended );
       Else
           Obj^.Value := Nil;            { string default to nil !!! }
       End; { Case Of }

       If( ( Kind = aString ) Or ( Obj^.Value <> Nil ) ) Then
        Begin
             NewObj := True;
             P^.HeapSaver^.AtInsert( 0, Obj );       { also keep track of it }
        End { If Then }
       Else
        Begin
             NewObj := False;
             Dispose( Obj, Done );
             Obj := Nil;
             E := Memory_Error;
        End; { If Then Else }

  End; { Function }

{ -------------------------------------------------------------------- }
{ Pushes an object back onto the eval stack and removes it from the    }
{  heapsaver list...                                                   }

 Procedure PushObj( Obj: PStackElement; P : PStackParser; E : Integer );
  Begin
       If( E = No_Error ) Then                  { if no error }
        Begin
             P^.EvalStack^.AtInsert( 0, Obj );
             P^.HeapSaver^.Delete( Obj );       { then remove the obj from the heapsaver }
        End
       Else
        P^.HeapSaver^.Free( Obj );              { else just dispose from heapsaver }
  End; { Function }

{ -------------------------------------------------------------------- }
{ Disposes an object and removes it from the heapsaver list...         }

 Procedure RemoveObj( Obj : PStackElement; P : PStackParser );
  Begin
       If( Obj <> Nil ) Then
        P^.HeapSaver^.Free( Obj );      { remove from heapsaver and dispose obj }
  End;  { Procedure }

{ --------------  DEFAULT PROCEDURE OF OpTable ( All are FAR ) -------- }

 Function OpDummy( P : PStackParser ) : Integer; Far; { Dummy func for OpTable }
  Begin
       OpDummy := No_Error;                  { no error, dont do anything at all }
  End; { Function }

{ --------------------------------------------------------------------- }
                                                       { + }
 Function OpAdd( P : PStackParser ) : Integer;  Far;
 Var
    Obj1, Obj2 : PStackElement;
    E          : Integer;
    V1, V2     : Extended;
    aP         : Pointer;
    S1, S2     : String;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( E = No_Error ) Then
        Begin
             If( ( Obj1^.Kind = aNumber ) And ( Obj2^.Kind = aNumber ) ) Then
              Begin
                   V1 := Extended( Obj1^.Value^ );
                   V2 := Extended( Obj2^.Value^ );
                   Extended( Obj1^.Value^ ) := V1 + V2;
              End  { If Then }
             Else
              If( ( Obj1^.Kind = aString ) And ( Obj2^.Kind = aString ) ) Then
               Begin
                    If( Obj1^.Value <> Nil ) Then
                     S1 := PString( Obj1^.Value )^
                    Else
                     S1 := An_Empty_String;

                    If( Obj2^.Value <> Nil ) Then
                     S2 := PString( Obj2^.Value )^
                    Else
                     S2 := An_Empty_String;

                    If( ( Length( S1 ) + Length( S2 ) ) > 255 ) Then
                     E := -37
                    Else
                     Begin
                          aP := NewStr( S2 + S1 );
                          If( ( Length( S1 + S2 ) > 1 ) And ( aP = Nil ) ) Then
                           E := Memory_Error;
                     End; { If Then Else }

                    If( E = No_Error ) Then
                     Begin
                          If( Obj1^.Value <> Nil ) Then
                           DisposeStr( PString( Obj1^.Value ) );
                          Obj1^.Value := aP;
                     End; { If Then }
               End
              Else
               E := -21;
        End; { If Then }

       PushObj( Obj1, P, E );
       RemoveObj( Obj2, P );

       OpAdd := E;
  End; { Function }

{ --------------------------------------------------------------------- }
                                                       { - }
 Function OpSubstract( P : PStackParser ) : Integer; Far;
  Var
    Obj1, Obj2 : PStackElement;
    E          : Integer;
    V1, V2     : Extended;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( E = No_Error ) Then
        Begin
             If( ( Obj1^.Kind = aNumber ) And ( Obj2^.Kind = aNumber ) ) Then
              Begin
                   V1 := Extended( Obj1^.Value^ );
                   V2 := Extended( Obj2^.Value^ );

                   Extended( Obj1^.Value^ ) := V2 - V1;
              End  { If Then }
             Else
              E := -22;
        End; { If Then }

       PushObj( Obj1, P, E );
       RemoveObj( Obj2, P );

       OpSubstract := E;
  End; { Function }

{ --------------------------------------------------------------------- }
                                                       { * }
 Function OpMultiply( P : PStackParser ) : Integer; Far;
   Var
    Obj1, Obj2 : PStackElement;
    E          : Integer;
    V1, V2     : Extended;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( E = No_Error ) Then
        Begin
             If( ( Obj1^.Kind = aNumber ) And ( Obj2^.Kind = aNumber ) ) Then
              Begin
                   V1 := Extended( Obj1^.Value^ );
                   V2 := Extended( Obj2^.Value^ );

                   Extended( Obj1^.Value^ ) := V2 * V1;
              End  { If Then }
             Else
              E := -23;
        End; { If Then }

       PushObj( Obj1, P, E );
       RemoveObj( Obj2, P );

       OpMultiply := E;
  End; { Function }

{ --------------------------------------------------------------------- }
                                                       { / }
 Function OpDivide( P : PStackParser ) : Integer; Far;
  Var
    Obj1, Obj2 : PStackElement;
    E          : Integer;
    V1, V2     : Extended;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( E = No_Error ) Then
        Begin
             If( ( Obj1^.Kind = aNumber ) And ( Obj2^.Kind = aNumber ) ) Then
              Begin
                   V1 := Extended( Obj1^.Value^ );
                   V2 := Extended( Obj2^.Value^ );

                   If( V1 <> 0.000 ) Then
                    Extended( Obj1^.Value^ ) := V2 / V1
                   Else
                    E := -13;
              End  { If Then }
             Else
              E := -24;

        End; { If Then }

       PushObj( Obj1, P, E );
       RemoveObj( Obj2, P );

       OpDivide := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpDiv( P : PStackParser ) : Integer; Far;                        { div }
   Var
    Obj1, Obj2 : PStackElement;
    E          : Integer;
    V1, V2     : Extended;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( E = No_Error ) Then
        Begin
             If( ( Obj1^.Kind = aNumber ) And ( Obj2^.Kind = aNumber ) ) Then
              Begin
                   V1 := Extended( Obj1^.Value^ );
                   V2 := Extended( Obj2^.Value^ );

                   If( V1 <> 0.000 ) Then
                    Extended( Obj1^.Value^ ) := Int( V2 / V1 )
                   Else
                    E := -13;
              End  { If Then }
             Else
              E := -24;

        End; { If Then }

       PushObj( Obj1, P, E );
       RemoveObj( Obj2, P );

       OpDiv := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpMod( P : PStackParser ) : Integer; Far;                        { mod }
 Var
    Obj1, Obj2 : PStackElement;
    E          : Integer;
    V1, V2     : Extended;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( E = No_Error ) Then
        Begin
             If( ( Obj1^.Kind = aNumber ) And ( Obj2^.Kind = aNumber ) ) Then
              Begin
                   V1 := Extended( Obj1^.Value^ );
                   V2 := Extended( Obj2^.Value^ );

                   If( V1 <> 0.000 ) Then
                    Extended( Obj1^.Value^ ) := Frac( V2 / V1 )
                   Else
                    E := -13;
              End  { If Then }
             Else
              E := -26;

        End; { If Then }

       PushObj( Obj1, P, E );
       RemoveObj( Obj2, P );

       OpMod := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpPower( P : PStackParser ) : Integer; Far;              { ^ }
 Var
    Obj1, Obj2 : PStackElement;
    E          : Integer;
    V1, V2     : Extended;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( E = No_Error ) Then
        Begin
             If( ( Obj1^.Kind = aNumber ) And ( Obj2^.Kind = aNumber ) ) Then
              Begin
                   V1 := Extended( Obj1^.Value^ );
                   V2 := Extended( Obj2^.Value^ );

                   If( V2 < 0 ) Then
                    Begin
                         If( Int( V1 ) = V1 ) Then
                          Begin
                                If( V1 = 0.0000 ) Then
                                 V1 := 1.000
                                Else
                                 If( Odd( Trunc( V1 ) ) ) Then
                                  V1 := -1 * Exp( V1 * Ln( Abs( V2 ) ) )
                                 Else
                                  V1 := Exp( V1 * Ln( Abs( V2 ) ) );
                          End { If Then }
                         Else
                          E := -14;
                    End
                   Else
                    If( V2 = 0.000 ) Then
                     V1 := 0.000
                    Else
                     V1 := Exp( V1 * Ln( V2 ) );

                   Extended( Obj1^.Value^ ) := V1;
              End { If Then }
             Else
              E := -14;

        End; { If Then }

       PushObj( Obj1, P, E );
       RemoveObj( Obj2, P );

       OpPower := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpXRoot( P : PStackParser ) : Integer; Far;              { xroot }
 Var
    Obj1, Obj2 : PStackElement;
    E          : Integer;
    V1, V2     : Extended;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( E = No_Error ) Then
        Begin
             If( ( Obj1^.Kind = aNumber ) And ( Obj2^.Kind = aNumber ) ) Then
              Begin
                   V1 := Extended( Obj1^.Value^ );
                   V2 := Extended( Obj2^.Value^ );

                   If( V2 = 0 ) Then
                    E := -36
                   Else
                    If( V1 < 0 ) Then
                     If( ( Odd( Round( V2 ) ) ) And ( Int( V2 ) = V2 ) ) Then
                      V1 := -1 * Exp( Ln( Abs( V1 ) ) / V2 )
                     Else
                      E := -36
                    Else
                     If( V1 <> 0 ) Then
                      V1 := Exp( Ln( V1 ) / V2 );

                   Extended( Obj1^.Value^ ) := V1;
              End { If Then }
             Else
              E := -36;

        End; { If Then }

       PushObj( Obj1, P, E );
       RemoveObj( Obj2, P );

       OpXRoot := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpNegate( P : PStackParser ) : Integer; Far;
 Var
    Obj1 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        Begin
             If( Obj1^.Kind = aNumber ) Then
              Extended( Obj1^.Value^ ) := -1 * Extended( Obj1^.Value^ )
             Else
              E := -27;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpNegate := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpInt( P : PStackParser ) : Integer; Far;
 Var
    Obj1 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        Begin
             If( Obj1^.Kind = aNumber ) Then
              Extended( Obj1^.Value^ ) := Int( Extended( Obj1^.Value^ ) )
             Else
              E := -28;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpInt := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpAbs( P : PStackParser ) : Integer; Far;
 Var
    Obj1 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        Begin
             If( Obj1^.Kind = aNumber ) Then
              Extended( Obj1^.Value^ ) := Abs( Extended( Obj1^.Value^ ) )
             Else
              E := -29;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpAbs := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpRnd( P : PStackParser ) : Integer; Far;
 Var
    Obj1 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       If( NewObj( Obj1, P, aNumber, E ) ) Then
        Begin
             Extended( Obj1^.Value^ ) := Random( 65534 ) / ( Random( 65534 ) + 1 );

             If( Random( 20 ) > 10 ) Then
              Extended( Obj1^.Value^ ) := Extended( Obj1^.Value^ ) * -1;

             PStackParser( P )^.ReEvaluate := True;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpRnd := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpSin( P : PStackParser ) : Integer; Far;
 Var
    Obj1 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        Begin
             If( Obj1^.Kind = aNumber ) Then
              Extended( Obj1^.Value^ ) := Sin( Extended( Obj1^.Value^ ) )
             Else
              E := -30;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpSin := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpCos( P : PStackParser ) : Integer; Far;
 Var
    Obj1 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        Begin
             If( Obj1^.Kind = aNumber ) Then
              Extended( Obj1^.Value^ ) := Cos( Extended( Obj1^.Value^ ) )
             Else
              E := -31;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpCos := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpTan( P : PStackParser ) : Integer; Far;
 Var
    Obj1 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        Begin
             If( Obj1^.Kind = aNumber ) Then
              Extended( Obj1^.Value^ ) := Sin( Extended( Obj1^.Value^ ) ) / Cos( Extended( Obj1^.Value^ ) )
             Else
              E := -32;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpTan := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpATan( P : PStackParser ) : Integer; Far;
 Var
    Obj1 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        Begin
             If( Obj1^.Kind = aNumber ) Then
              Extended( Obj1^.Value^ ) := ArcTan( Extended( Obj1^.Value^ ) )
             Else
              E := -33;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpATan := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpASin( P : PStackParser ) : Integer; Far;
 Var
    Obj1 : PStackElement;
    E    : Integer;
    V1   : Extended;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        Begin
             If( Obj1^.Kind = aNumber ) Then
              Begin
                   V1 := Extended( Obj1^.Value^ );

                   If( Abs( V1 ) > 1 ) Then
                    E := -19
                   Else
                    If( Abs( V1 ) = 1 ) Then
                     V1 := V1 / 2 * Pi
                    Else
                     V1 := ArcTan( V1 / Sqrt( 1 - Sqr( V1 ) ) );

                   Extended( Obj1^.Value^ ) := V1;
               End
              Else
               E := -19;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpASin := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpACos( P : PStackParser ) : Integer; Far;
 Var
    Obj1 : PStackElement;
    E    : Integer;
    V1   : Extended;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        Begin
             If( Obj1^.Kind = aNumber ) Then
              Begin
                   V1 := Extended( Obj1^.Value^ );

                   If( Abs( V1 ) > 1 ) Then
                    E := -20
                   Else
                    If( V1 = 0 ) Then
                     V1 := Pi / 2
                    Else
                     If( V1 < 0 ) Then
                      V1 := Pi + ArcTan( Sqrt( 1 - Sqr( V1 ) ) / V1 )
                     Else
                      V1 := ArcTan( Sqrt( 1 - Sqr( V1 ) ) / V1 );

                   Extended( Obj1^.Value^ ) := V1;
              End
             Else
              E := -20;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpACos := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpLn( P : PStackParser ) : Integer; Far;
 Var
    Obj1 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        Begin
             If( ( Obj1^.Kind = aNumber ) And ( Extended( Obj1^.Value^ ) > 0 ) ) Then
              Extended( Obj1^.Value^ ) := Ln( Extended( Obj1^.Value^ ) )
             Else
              E := -15;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpLn := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpExp( P : PStackParser ) : Integer; Far;
 Var
    Obj1 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        Begin
             If( Obj1^.Kind = aNumber ) Then
              Extended( Obj1^.Value^ ) := Exp( Extended( Obj1^.Value^ ) )
             Else
              E := -14;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpExp := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpLog( P : PStackParser ) : Integer; Far;
 Var
    Obj1 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        Begin
             If( ( Obj1^.Kind = aNumber ) And ( Extended( Obj1^.Value^ ) > 0 ) ) Then
              Extended( Obj1^.Value^ ) := Ln( Extended( Obj1^.Value^ ) ) / Ln( 10 )
             Else
              E := -15;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpLog := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpSqrt( P : PStackParser ) : Integer; Far;
 Var
    Obj1 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        Begin
             If( ( Obj1^.Kind = aNumber ) And ( Extended( Obj1^.Value^ ) >= 0 ) ) Then
              Extended( Obj1^.Value^ ) := Sqrt( Extended( Obj1^.Value^ ) )
             Else
              E := -16;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpSqrt := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpFac( P : PStackParser ) : Integer; Far;
 Var
    Obj1 : PStackElement;
    E    : Integer;
    C    : LongInt;
    V, V1 : Extended;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        Begin
             If( Obj1^.Kind = aNumber ) Then
              Begin
                   V1 := Extended( Obj1^.Value^ );

                   If( ( Int( V1 ) = V1 ) And ( V1 > 0 ) ) Then
                    Begin
                         V := 1;

                         For C := 1 To Round( V1 ) Do
                          V := V * C;
                    End { If Then }
                   Else
                    E := -17;

                   Extended( Obj1^.Value^ ) := V;
              End
             Else
              E := -17;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpFac := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpSum( P : PStackParser ) : Integer; Far;
 Var
    Obj1, Obj2   : PStackElement;
    E      : Integer;
    X      : Longint;
    Cnt    : Longint;
    V1     : Extended;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        If( Obj1^.Kind <> aCounter ) Then
         E := List_Eval_Error
        Else
         Begin
              Cnt := Round( Extended( Obj1^.Value^ ) );
              V1 := 0;

              For X := 1 To Cnt Do
               Begin
                    If( PopObj( Obj2, P, E ) ) Then
                     Begin
                          If( Obj2^.Kind = aNumber ) Then
                           V1 := Extended( Obj2^.Value^ ) + V1
                          Else
                           E := -33;

                          If( Obj2^.Kind = anError ) Then
                           E := Round( Extended( Obj2^.Value^ ) );

                          RemoveObj( Obj2, P );
                     End { If Then }
                    Else
                     E := List_Eval_Error;
               End; { For Do }

              Extended( Obj1^.Value^ ) := V1;
              Obj1^.Kind := aNumber;
         End; { If Then }

       PushObj( Obj1, P, E );

       OpSum := E;
  End; { Function }

{ ----------------------------------------------------------------------- }

 Function OpSqrSum( P : PStackParser ) : Integer; Far;
  Var
    Obj1, Obj2 : PStackElement;
    E      : Integer;
    X      : Longint;
    Cnt    : Longint;
    V1     : Extended;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        If( Obj1^.Kind <> aCounter ) Then
         E := List_Eval_Error
        Else
         Begin
              Cnt := Round( Extended( Obj1^.Value^ ) );
              V1 := 0;

              For X := 1 To Cnt Do
               Begin
                    If( PopObj( Obj2, P, E ) ) Then
                     Begin
                          If( Obj2^.Kind = aNumber ) Then
                           V1 := Sqr( Extended( Obj2^.Value^ ) ) + V1
                          Else
                           E := -34;

                          If( Obj2^.Kind = anError ) Then
                           E := Round( Extended( Obj2^.Value^ ) );

                          RemoveObj( Obj2, P );
                     End { If Then }
                    Else
                     E := List_Eval_Error;
               End; { For Do }

              Extended( Obj1^.Value^ ) := V1;
              Obj1^.Kind := aNumber;
         End; { If Then }

       PushObj( Obj1, P, E );

       OpSqrSum := E;
  End; { Function }

{ ----------------------------------------------------------------------- }

 Function OpMean( P : PStackParser ) : Integer; Far;
  Var
    Obj1, Obj2 : PStackElement;
    E      : Integer;
    X      : Longint;
    Cnt    : Longint;
    V1     : Extended;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        If( Obj1^.Kind <> aCounter ) Then
         E := List_Eval_Error
        Else
         Begin
              Cnt := Round( Extended( Obj1^.Value^ ) );
              V1 := 0;

              For X := 1 To Cnt Do
               Begin
                    If( PopObj( Obj2, P, E ) ) Then
                     Begin
                          If( Obj2^.Kind = aNumber ) Then
                           V1 := Extended( Obj2^.Value^ ) + V1
                          Else
                           E := -35;

                          If( Obj2^.Kind = anError ) Then
                           E := Round( Extended( Obj2^.Value^ ) );

                          RemoveObj( Obj2, P );
                     End { If Then }
                    Else
                     E := List_Eval_Error;
               End; { For Do }

              Extended( Obj1^.Value^ ) := V1 / Cnt;
              Obj1^.Kind := aNumber;
         End; { If Then }

       PushObj( Obj1, P, E );

       OpMean := E;
  End; { Function }

{ -------------------------------------------------------------------------- }

 Function OpNelem( P : PStackParser ) : Integer; Far;
 Var
    Obj1, Obj2   : PStackElement;
    E      : Integer;
    X      : Longint;
    Cnt    : Longint;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        If( Obj1^.Kind <> aCounter ) Then
         E := List_Eval_Error
        Else
         Begin
              Cnt := Round( Extended( Obj1^.Value^ ) );
              Obj1^.Kind := aNumber;

              For X := 1 To Cnt Do
               If( PopObj( Obj2, P, E ) ) Then
                RemoveObj( Obj2, P )
               Else
                E := List_Eval_Error;
         End; { If Then }

       PushObj( Obj1, P, E );

       OpNelem := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpBT( P : PStackParser ) : Integer;  Far;
 Var
    Obj1, Obj2, Obj3 : PStackElement;
    E          : Integer;
    S1, S2     : String;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( NewObj( Obj3, P, aBoole, E ) ) Then
        Begin
             If( ( Obj1^.Kind = aNumber ) And ( Obj2^.Kind = aNumber ) ) Then
              Begin
                   If( Extended( Obj2^.Value^ ) > Extended( Obj1^.Value^ ) ) Then
                    PBoolean( Obj3^.Value )^ := True
                   Else
                    PBoolean( Obj3^.Value )^ := False;
              End
             Else
              If( ( Obj1^.Kind = aString ) And ( Obj2^.Kind = aString ) ) Then
               Begin
                    If( Obj1^.Value <> Nil ) Then
                     S1 := PString( Obj1^.Value )^
                    Else
                     S1 := An_Empty_String;

                    If( Obj2^.Value <> Nil ) Then
                     S2 := PString( Obj2^.Value )^
                    Else
                     S2 := An_Empty_String;

                    If( S2 > S1 ) Then
                     PBoolean( Obj3^.Value )^ := True
                    Else
                     PBoolean( Obj3^.Value )^ := False;
               End
              Else
               E := -39;
        End; { If Then }

       PushObj( Obj3, P, E );
       RemoveObj( Obj1, P );
       RemoveObj( Obj2, P );

       OpBT := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpST( P : PStackParser ) : Integer;  Far;
 Var
    Obj1, Obj2, Obj3 : PStackElement;
    E          : Integer;
    S1, S2     : String;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( NewObj( Obj3, P, aBoole, E ) ) Then
        Begin
             If( ( Obj1^.Kind = aNumber ) And ( Obj2^.Kind = aNumber ) ) Then
              Begin
                   If( Extended( Obj2^.Value^ ) < Extended( Obj1^.Value^ ) ) Then
                    PBoolean( Obj3^.Value )^ := True
                   Else
                    PBoolean( Obj3^.Value )^ := False;
              End
             Else
              If( ( Obj1^.Kind = aString ) And ( Obj2^.Kind = aString ) ) Then
               Begin
                    If( Obj1^.Value <> Nil ) Then
                     S1 := PString( Obj1^.Value )^
                    Else
                     S1 := An_Empty_String;

                    If( Obj2^.Value <> Nil ) Then
                     S2 := PString( Obj2^.Value )^
                    Else
                     S2 := An_Empty_String;

                    If( S2 < S1 ) Then
                     PBoolean( Obj3^.Value )^ := True
                    Else
                     PBoolean( Obj3^.Value )^ := False;
               End
              Else
               E := -39;
        End; { If Then }

       PushObj( Obj3, P, E );
       RemoveObj( Obj1, P );
       RemoveObj( Obj2, P );

       OpST := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpSame( P : PStackParser ) : Integer;  Far;
 Var
    Obj1, Obj2, Obj3 : PStackElement;
    E          : Integer;
    S1, S2     : String;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( NewObj( Obj3, P, aBoole, E ) ) Then
        Begin
             If( ( Obj1^.Kind = aNumber ) And ( Obj2^.Kind = aNumber ) ) Then
              Begin
                   If( Extended( Obj2^.Value^ ) = Extended( Obj1^.Value^ ) ) Then
                    PBoolean( Obj3^.Value )^ := True
                   Else
                    PBoolean( Obj3^.Value )^ := False;
              End
             Else
              If( ( Obj1^.Kind = aString ) And ( Obj2^.Kind = aString ) ) Then
               Begin
                    If( Obj1^.Value <> Nil ) Then
                     S1 := PString( Obj1^.Value )^
                    Else
                     S1 := An_Empty_String;

                    If( Obj2^.Value <> Nil ) Then
                     S2 := PString( Obj2^.Value )^
                    Else
                     S2 := An_Empty_String;

                    If( S2 = S1 ) Then
                     PBoolean( Obj3^.Value )^ := True
                    Else
                     PBoolean( Obj3^.Value )^ := False;
               End
              Else
               If( ( Obj1^.Kind = aBoole ) And ( Obj2^.Kind = aBoole ) ) Then
                Begin
                     If( Boolean( Obj2^.Value^ ) = Boolean( Obj1^.Value^ ) ) Then
                      PBoolean( Obj3^.Value )^ := True
                     Else
                      PBoolean( Obj3^.Value )^ := False;
                End
               Else
                E := -39;
        End; { If Then }

       PushObj( Obj3, P, E );
       RemoveObj( Obj1, P );
       RemoveObj( Obj2, P );

       OpSame := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpBTSame( P : PStackParser ) : Integer;  Far;
 Var
    Obj1, Obj2, Obj3 : PStackElement;
    E          : Integer;
    S1, S2     : String;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( NewObj( Obj3, P, aBoole, E ) ) Then
        Begin
             If( ( Obj1^.Kind = aNumber ) And ( Obj2^.Kind = aNumber ) ) Then
              Begin
                   If( Extended( Obj2^.Value^ ) >= Extended( Obj1^.Value^ ) ) Then
                    PBoolean( Obj3^.Value )^ := True
                   Else
                    PBoolean( Obj3^.Value )^ := False;
              End
             Else
              If( ( Obj1^.Kind = aString ) And ( Obj2^.Kind = aString ) ) Then
               Begin
                    If( Obj1^.Value <> Nil ) Then
                     S1 := PString( Obj1^.Value )^
                    Else
                     S1 := An_Empty_String;

                    If( Obj2^.Value <> Nil ) Then
                     S2 := PString( Obj2^.Value )^
                    Else
                     S2 := An_Empty_String;

                    If( S2 >= S1 ) Then
                     PBoolean( Obj3^.Value )^ := True
                    Else
                     PBoolean( Obj3^.Value )^ := False;
               End
              Else
               E := -39;
        End; { If Then }

       PushObj( Obj3, P, E );
       RemoveObj( Obj1, P );
       RemoveObj( Obj2, P );

       OpBTSame := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpSTSame( P : PStackParser ) : Integer;  Far;
 Var
    Obj1, Obj2, Obj3 : PStackElement;
    E          : Integer;
    S1, S2     : String;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( NewObj( Obj3, P, aBoole, E ) ) Then
        Begin
             If( ( Obj1^.Kind = aNumber ) And ( Obj2^.Kind = aNumber ) ) Then
              Begin
                   If( Extended( Obj2^.Value^ ) <= Extended( Obj1^.Value^ ) ) Then
                    PBoolean( Obj3^.Value )^ := True
                   Else
                    PBoolean( Obj3^.Value )^ := False;
              End
             Else
              If( ( Obj1^.Kind = aString ) And ( Obj2^.Kind = aString ) ) Then
               Begin
                    If( Obj1^.Value <> Nil ) Then
                     S1 := PString( Obj1^.Value )^
                    Else
                     S1 := An_Empty_String;

                    If( Obj2^.Value <> Nil ) Then
                     S2 := PString( Obj2^.Value )^
                    Else
                     S2 := An_Empty_String;

                    If( S2 <= S1 ) Then
                     PBoolean( Obj3^.Value )^ := True
                    Else
                     PBoolean( Obj3^.Value )^ := False;
               End
              Else
               E := -39;
        End; { If Then }

       PushObj( Obj3, P, E );
       RemoveObj( Obj1, P );
       RemoveObj( Obj2, P );

       OpSTSame := E;
  End; { Function }

{ ----------------------------------------------------------------------- }

 Function OpNotSame( P : PStackParser ) : Integer;  Far;
 Var
    Obj1, Obj2, Obj3 : PStackElement;
    E          : Integer;
    R          : Boolean;
    S1, S2     : String;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( NewObj( Obj3, P, aBoole, E ) ) Then
        Begin
             If( ( Obj1^.Kind = aNumber ) And ( Obj2^.Kind = aNumber ) ) Then
              Begin
                   If( Extended( Obj2^.Value^ ) <> Extended( Obj1^.Value^ ) ) Then
                    PBoolean( Obj3^.Value )^ := True
                   Else
                    PBoolean( Obj3^.Value )^ := False;
              End
             Else
              If( ( Obj1^.Kind = aString ) And ( Obj2^.Kind = aString ) ) Then
               Begin
                    If( Obj1^.Value <> Nil ) Then
                     S1 := PString( Obj1^.Value )^
                    Else
                     S1 := An_Empty_String;

                    If( Obj2^.Value <> Nil ) Then
                     S2 := PString( Obj2^.Value )^
                    Else
                     S2 := An_Empty_String;

                    If( S2 <> S1 ) Then
                     PBoolean( Obj3^.Value )^ := True
                    Else
                     PBoolean( Obj3^.Value )^ := False;
               End
              Else
               If( ( Obj1^.Kind = aBoole ) And ( Obj2^.Kind = aBoole ) ) Then
                Begin
                     If( Boolean( Obj2^.Value^ ) <> Boolean( Obj1^.Value^ ) ) Then
                      PBoolean( Obj3^.Value )^ := True
                     Else
                      PBoolean( Obj3^.Value )^ := False;
                End
               Else
                E := -39;
        End; { If Then }

       PushObj( Obj3, P, E );
       RemoveObj( Obj1, P );
       RemoveObj( Obj2, P );

       OpNotSame := E;
  End; { Function }

{ -------------------------------------------------------------------- }

 Function OpIFTE( P : PStackParser ) : Integer;  Far;
 Var
    Obj1, Obj2, Obj3 : PStackElement;
    E          : Integer;
    R          : Boolean;
  Begin
       E := No_Error;

       If( PopObj( Obj1, P, E ) ) Then
        E := No_Error;
       If( PopObj( Obj2, P, E ) ) Then
        E := No_Error;
       PopObj( Obj3, P, E );

       If( E = No_Error ) Then
        Begin
             If( Obj3^.Kind = aBoole ) Then
              Begin
                   If( Boolean( Obj3^.Value^ ) ) Then
                    R := True
                   Else
                    R := False;
              End
             Else
              E := -41;
        End; { If Then }

       If( R ) Then
        Begin
             PushObj( Obj2, P, E );
             RemoveObj( Obj1, P );
        End
       Else
        Begin
             PushObj( Obj1, P, E );
             RemoveObj( Obj2, P );
        End;

       RemoveObj( Obj3, P );

       OpIFTE := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpOr( P : PStackParser ) : Integer; Far;       { Or }
   Var
    Obj1, Obj2 : PStackElement;
    E          : Integer;
    V1, V2     : Boolean;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( E = No_Error ) Then
        Begin
             If( ( Obj1^.Kind = aBoole ) And ( Obj2^.Kind = aBoole ) ) Then
              Begin
                   V1 := PBoolean( Obj1^.Value )^;
                   V2 := PBoolean( Obj2^.Value )^;
                   PBoolean( Obj1^.Value )^ := V1 Or V2;
              End  { If Then }
             Else
              E := -45;
        End; { If Then }

       PushObj( Obj1, P, E );
       RemoveObj( Obj2, P );

       OpOr := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpXOr( P : PStackParser ) : Integer; Far;       { XOr }
   Var
    Obj1, Obj2 : PStackElement;
    E          : Integer;
    V1, V2     : Boolean;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( E = No_Error ) Then
        Begin
             If( ( Obj1^.Kind = aBoole ) And ( Obj2^.Kind = aBoole ) ) Then
              Begin
                   V1 := PBoolean( Obj1^.Value )^;
                   V2 := PBoolean( Obj2^.Value )^;
                   PBoolean( Obj1^.Value )^ := V1 XOr V2;
              End  { If Then }
             Else
              E := -46;
        End; { If Then }

       PushObj( Obj1, P, E );
       RemoveObj( Obj2, P );

       OpXOr := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpAnd( P : PStackParser ) : Integer; Far;       { And }
   Var
    Obj1, Obj2 : PStackElement;
    E          : Integer;
    V1, V2     : Boolean;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( E = No_Error ) Then
        Begin
             If( ( Obj1^.Kind = aBoole ) And ( Obj2^.Kind = aBoole ) ) Then
              Begin
                   V1 := PBoolean( Obj1^.Value )^;
                   V2 := PBoolean( Obj2^.Value )^;
                   PBoolean( Obj1^.Value )^ := V1 And V2;
              End  { If Then }
             Else
              E := -47;
        End; { If Then }

       PushObj( Obj1, P, E );
       RemoveObj( Obj2, P );

       OpAnd := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpNot( P : PStackParser ) : Integer; Far;       { Not }
   Var
    Obj1       : PStackElement;
    E          : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        Begin
             If( Obj1^.Kind = aBoole ) Then
              Begin
                   PBoolean( Obj1^.Value )^ := Not PBoolean( Obj1^.Value )^;
              End  { If Then }
             Else
              E := -48;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpNot := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpStrUpr( P : PStackParser ) : Integer; Far;
 Var
    Obj1 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        Begin
             If( Obj1^.Kind = aString ) Then
              Begin
                   If( Obj1^.Value <> Nil ) Then
                    PString( Obj1^.Value )^ := StrUpr( PString( Obj1^.Value )^ )
              End
             Else
              E := -49;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpStrUpr := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpStrLwr( P : PStackParser ) : Integer; Far;
 Var
    Obj1 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        Begin
             If( Obj1^.Kind = aString ) Then
              Begin
                   If( Obj1^.Value <> Nil ) Then
                    PString( Obj1^.Value )^ := StrLwr( PString( Obj1^.Value )^ )
              End
             Else
              E := -50;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpStrLwr := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpIsStr( P : PStackParser ) : Integer; Far;
 Var
    Obj1, Obj2 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( NewObj( Obj2, P, aBoole, E ) ) Then
        Begin
             If( Obj1^.Kind = aString ) Then
              Boolean( Obj2^.Value^ ) := True
             Else
              Boolean( Obj2^.Value^ ) := False;
        End; { If Then }

       PushObj( Obj2, P, E );
       RemoveObj( Obj1, P );

       OpIsStr := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpIsBoole( P : PStackParser ) : Integer; Far;
 Var
    Obj1, Obj2 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( NewObj( Obj2, P, aBoole, E ) ) Then
        Begin
             If( Obj1^.Kind = aBoole ) Then
              Boolean( Obj2^.Value^ ) := True
             Else
              Boolean( Obj2^.Value^ ) := False;
        End; { If Then }

       PushObj( Obj2, P, E );
       RemoveObj( Obj1, P );

       OpIsBoole := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpIsNumber( P : PStackParser ) : Integer; Far;
 Var
    Obj1, Obj2 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( NewObj( Obj2, P, aBoole, E ) ) Then
        Begin
             If( Obj1^.Kind = aNumber ) Then
              Boolean( Obj2^.Value^ ) := True
             Else
              Boolean( Obj2^.Value^ ) := False;
        End; { If Then }

       PushObj( Obj2, P, E );
       RemoveObj( Obj1, P );

       OpIsNumber := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpIsError( P : PStackParser ) : Integer; Far;
 Var
    Obj1, Obj2 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( NewObj( Obj2, P, aBoole, E ) ) Then
        Begin
             If( Obj1^.Kind = anError ) Then
              Begin
                   Boolean( Obj2^.Value^ ) := True;
                   If( ( E <> Memory_Error ) And ( E <> Major_Internal_Error )
                         And ( E <> List_Eval_Error ) ) Then
                    E := No_Error;
              End
             Else
              Boolean( Obj2^.Value^ ) := False;
        End; { If Then }

       PushObj( Obj2, P, E );
       RemoveObj( Obj1, P );

       OpIsError := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpStrLen( P : PStackParser ) : Integer; Far;
 Var
    Obj1, Obj2 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( NewObj( Obj2, P, aNumber, E ) ) Then
        Begin
             If( Obj1^.Kind = aString ) Then
              If( Obj1^.Value = Nil ) Then
               Extended( Obj2^.Value^ ) := 0
              Else
               Extended( Obj2^.Value^ ) := Length( String( Obj1^.Value^ ) )
             Else
              E := -51;
        End; { If Then }

       PushObj( Obj2, P, E );
       RemoveObj( Obj1, P );

       OpStrLen := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpStrPos( P : PStackParser ) : Integer; Far;
 Var
    Obj1, Obj2, Obj3 : PStackElement;
    E    : Integer;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( NewObj( Obj3, P, aNumber, E ) ) Then
        Begin
             If( ( Obj1^.Kind = aString ) And ( Obj2^.Kind = aString ) ) Then
              Begin
                   If( ( Obj2^.Value = Nil ) Or ( Obj1^.Value = Nil ) ) Then
                    Extended( Obj3^.Value^ ) := 0
                   Else
                    Extended( Obj3^.Value^ ) := Pos( String( Obj2^.Value^ ), String( Obj1^.Value^ ) );
              End
             Else
              E := -52;
        End; { If Then }

       PushObj( Obj3, P, E );
       RemoveObj( Obj1, P );
       RemoveObj( Obj2, P );

       OpStrPos := E;
  End; { Function }

{ ---------------------------------------------------------------------- }

 Function OpInsertStr( P : PStackParser ) : Integer; Far;
 Var
    Obj1, Obj2, Obj3 : PStackElement;
    Result           : PStackElement;
    E, Position      : Integer;
    Dest, Source     : String;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );
       PopObj( Obj3, P, E );

       If( NewObj( Result, P, aString, E ) ) Then
        Begin
             If( ( Obj1^.Kind <> aNumber ) Or ( Obj2^.Kind <> aString )
                  Or ( Obj3^.Kind <> aString ) ) Then
              E := -53
             Else
              Begin
                   If( Frac( PExtended( Obj1^.Value )^ ) = 0 ) Then
                    Position := Round( PExtended( Obj1^.Value )^ )
                   Else
                    Position := -1;

                   If( PString( Obj3^.Value ) <> Nil ) Then
                    Dest := PString( Obj3^.Value )^
                   Else
                    Dest := An_Empty_String;

                   If( PString( Obj2^.Value ) <> Nil ) Then
                    Source := PString( Obj2^.Value )^
                   Else
                    Source := An_Empty_String;

                   If( Length( Source ) + Length( Dest ) > 255 ) Then
                    E := -37
                   Else
                    If( ( Position > 0 ) And ( Position <= Length( Dest ) + 1 ) ) Then
                     Begin
                          Insert( Source, Dest, Position );
                          Result^.Value := NewStr( Dest );
                          If( ( Result^.Value = Nil ) And ( Length( Dest ) <> 0 ) ) Then
                           E := Memory_Error;
                     End
                    Else
                     E := -53;
              End; { If Then Else }
        End; { If Then }

       PushObj( Result, P, E );
       RemoveObj( Obj1, P );
       RemoveObj( Obj2, P );
       RemoveObj( Obj3, P );

       OpInsertStr := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpDeleteStr( P : PStackParser ) : Integer; Far;
 Var
    Obj1, Obj2, Obj3 : PStackElement;
    Result           : PStackElement;
    E, Position, Nr  : Integer;
    TempStr          : String;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );
       PopObj( Obj3, P, E );

       If( NewObj( Result, P, aString, E ) ) Then
        Begin
             If( ( Obj1^.Kind <> aNumber ) Or ( Obj2^.Kind <> aNumber )
                  Or ( Obj3^.Kind <> aString ) ) Then
              E := -54
             Else
              Begin
                   If( Frac( PExtended( Obj2^.Value )^ ) = 0 ) Then
                    Position := Round( PExtended( Obj2^.Value )^ )
                   Else
                    Position := -1;

                   If( Frac( PExtended( Obj1^.Value )^ ) = 0 ) Then
                    Nr := Round( PExtended( Obj1^.Value )^ )
                   Else
                    Nr := 300;

                   If( Obj3^.Value <> Nil ) Then
                    TempStr := PString( Obj3^.Value )^
                   Else
                    TempStr := An_Empty_String;

                   If( ( Position > 0 ) And ( Position <= Length( TempStr ) )
                       And ( Nr >= 0 ) And ( Position + Nr <= Length( TempStr ) + 1 ) ) Then
                    Begin
                         Delete( TempStr, Position, Nr );
                         Result^.Value := NewStr( TempStr );
                         If( ( Result^.Value = Nil ) And ( Length( TempStr ) <> 0 ) ) Then
                          E := Memory_Error;
                    End
                   Else
                    E := -54;
              End; { If Then Else }
        End; { If Then }

       PushObj( Result, P, E );
       RemoveObj( Obj1, P );
       RemoveObj( Obj2, P );
       RemoveObj( Obj3, P );

       OpDeleteStr := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpSubStr( P : PStackParser ) : Integer; Far;
 Var
    Obj1, Obj2, Obj3 : PStackElement;
    Result           : PStackElement;
    E, Position, Nr  : Integer;
    TempStr          : String;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );
       PopObj( Obj3, P, E );

       If( NewObj( Result, P, aString, E ) ) Then
        Begin
             If( ( Obj1^.Kind <> aNumber ) Or ( Obj2^.Kind <> aNumber )
                  Or ( Obj3^.Kind <> aString ) ) Then
              E := -55
             Else
              Begin
                   If( Frac( PExtended( Obj2^.Value )^ ) = 0 ) Then
                    Position := Round( PExtended( Obj2^.Value )^ )
                   Else
                    Position := -1;

                   If( Frac( PExtended( Obj1^.Value )^ ) = 0 ) Then
                    Nr := Round( PExtended( Obj1^.Value )^ )
                   Else
                    Nr := 300;

                   If( Obj3^.Value <> Nil ) Then
                    TempStr := PString( Obj3^.Value )^
                   Else
                    TempStr := An_Empty_String;

                   If( ( Position > 0 ) And ( Position <= Length( TempStr ) )
                       And ( Nr >= 0 ) And ( Position + Nr <= Length( TempStr ) + 1 ) ) Then
                    Begin
                         TempStr := Copy( TempStr, Position, Nr );
                         Result^.Value := NewStr( TempStr );
                         If( ( Result^.Value = Nil ) And ( Length( TempStr ) <> 0 ) ) Then
                          E := Memory_Error;
                    End
                   Else
                    E := -55;
              End; { If Then Else }
        End; { If Then }

       PushObj( Result, P, E );
       RemoveObj( Obj1, P );
       RemoveObj( Obj2, P );
       RemoveObj( Obj3, P );

       OpSubStr := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpParseStr( P : PStackParser ) : Integer; Far;
 Var
    Obj1     : PStackElement;
    NewParser : PStackParser;
    E         : Integer;
    TheResult : Pointer;
    Kind      : ElementKind;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( E = No_Error ) Then
        Begin
             If( Obj1^.Kind <> aString ) Then
              E := -56
             Else
              If( Obj1^.Value <> Nil ) Then
               Begin
                    NewParser := New( PStackParser, Init );

                    If( NewParser = Nil ) Then
                     E := Memory_Error
                    Else
                     Begin
                          NewParser^.SetFormula( PString( Obj1^.Value )^ );
                          DisposeStr( Obj1^.Value );
                          Obj1^.Value := Nil;

                          Dispose( NewParser^.Variables, Done );  { pass on the current vars by refs }
                          NewParser^.Variables := P^.Variables;

                          If( NewParser^.GetResult( TheResult, Kind ) ) Then
                           Begin
                                Obj1^.Kind := Kind;
                                Obj1^.Value := TheResult;
                                NewParser^.Result := Nil;
                           End { If Then }
                          Else
                           Begin
                                Obj1^.Value := New( PExtended );
                                If( Obj1^.Value <> Nil ) Then
                                 Begin
                                      Extended( Obj1^.Value^ ) := NewParser^.GetError;
                                      Obj1^.Kind := anError;
                                 End
                                Else
                                 E := Memory_Error;
                           End; { If Then Else }

                          NewParser^.Variables := Nil;
                          Dispose( NewParser, Done );
                     End; { If Then Else }
               End { If Then }
              Else
               E := -56;
        End; { If Then }

       PushObj( Obj1, P, E );

       OpParseStr := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpMakeStr( P : PStackParser ) : Integer; Far;
 Var
    Obj1, Obj2 : PStackElement;
    E, TmpE    : Integer;
    TmpStr     : String;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );

       If( NewObj( Obj2, P, aString, E ) ) Then
        Begin
             Case( Obj1^.Kind ) Of
                 aNumber: Str( Extended( Obj1^.Value^ ), TmpStr );
                 aBoole:  If( PBoolean( Obj1^.Value )^ ) Then
                           TmpStr := 'TRUE'
                          Else
                           TmpStr := 'FALSE';
                 aString: If( Obj1^.Value <> Nil ) Then
                           TmpStr := PString( Obj1^.Value )^
                          Else
                           TmpStr := An_Empty_String;
                 anError: Begin
                               TmpE := P^.Error;
                               P^.Error := E;
                               TmpStr := P^.GetErrorStr;
                               P^.Error := TmpE;
                               If( ( E <> Memory_Error ) And ( E <> List_Eval_Error )
                                   And ( E <> Major_Internal_Error ) )  Then
                                E := No_Error;
                          End;
             Else
                 E := -57;
             End; { Case }

             Obj2^.Value := NewStr( TmpStr );
             If( ( Obj2^.Value = Nil ) And ( Length( TmpStr ) <> 0 ) ) Then
              E := Memory_Error;
        End; { If Then }

       PushObj( Obj2, P, E );
       RemoveObj( Obj1, P );

       OpMakeStr := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpAssign( P : PStackParser ) : Integer; Far;
 Var
    Obj1, Obj2 : PStackElement;
    E          : Integer;
    TheVar     : PVariable;
  Begin
       E := No_Error;

       PopObj( Obj1, P, E );
       PopObj( Obj2, P, E );

       If( E = No_Error ) Then
        Begin
             If( Obj2^.Index < 0 ) Then          { was it derived from a var ? }
              Begin
                   TheVar := P^.Variables^.At( -( Obj2^.Index + 1 ) );

                   Case( Obj1^.Kind ) Of
                      aNumber:  Begin
                                     FreeThisItem( TheVar^.Kind, TheVar^.Value );
                                     TheVar^.Kind := aNumber;
                                     TheVar^.Value := New( PExtended );
                                     If( TheVar^.Value = Nil ) Then
                                      E := Memory_Error
                                     Else
                                      Extended( TheVar^.Value^ ) := Extended( Obj1^.Value^ );
                                End;
                      aString:  Begin
                                     FreeThisItem( TheVar^.Kind, TheVar^.Value );
                                     TheVar^.Kind := aString;
                                     If( Obj1^.Value = Nil ) Then
                                      TheVar^.Value := Nil
                                     Else
                                      TheVar^.Value := NewStr( PString( Obj1^.Value )^ );

                                     If( ( TheVar^.Value = Nil ) And ( Obj1^.Value <> Nil ) ) Then
                                      E := Memory_Error;
                                End;
                      aBoole:   Begin
                                     FreeThisItem( TheVar^.Kind, TheVar^.Value );
                                     TheVar^.Kind := aBoole;
                                     TheVar^.Value := New( PBoolean );
                                     If( TheVar^.Value = Nil ) Then
                                      E := Memory_Error
                                     Else
                                      Boolean( TheVar^.Value^ ) := Boolean( Obj1^.Value^ );
                                End;
                      Else
                       E := Major_Internal_Error;
                   End; { Case }

                   P^.ReEvaluate := True;
              End
             Else
              E := -58;
        End; { If Then }

       PushObj( Obj1, P, E );
       RemoveObj( Obj2, P );

       OpAssign := E;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function OpEnd( P : PStackParser ) : Integer; Far;
 Var
    Obj : PStackElement;
    E   : Integer;
  Begin
       E := No_Error;
       PopObj( Obj, P, E );
       RemoveObj( Obj, P );
       OpEnd := E;
  End; { Function }

{ -------------------- OpTable ---------------------------------------- }
{ The grammar field:

 Bit -> Function

 1   -> Operand in front = 1, operator in front = 0.
 2   -> Followed by operand = 1, followed by operator = 0.
 3   -> Increase priority level.
 4   -> Decrease priority level.
 5   -> Operator is a function, start counting operands until priority level is back.
 6   -> Operand seperator char, add 1 to variable counter.
 8   -> Dont put object in RList.
 9   -> This is an end of statement operator.

REMEMBER: WRONG VALUES IN THIS TABLE PRODUCES WRONG OUTPUT ! THERE MAY BE
MORE ENTRIES OF AN OPERATOR, BUT THE FIRST NONE ERROR ONE WILL BE PICKED !
THE GRAMMAR ROUTINE CHECKS THE CURRENT OPERATOR WITH THE PREVIOUS OPERATOR !!
}
Const                                  { Operator table definition, All ID's in capitals !! }
     ANonOperator           = 0;
     ALeftUnaryOperator     = 2;         { bit 2 = 1 }
     ARightUnaryOperator    = 1;         { bit 1 = 1 }
     ABinaryOperator        = 3;         { bits 1 and 2 = 1 }
     AFunctionOperator      = 16;        { bit 5 = 1 }
     AnEndStatementOperator = 256;       { bits 9 = 1 }
     OpTable : Array [ 1 .. 58 ] Of OpElement = ( ( ID: '(';       Level: 100; Grammar: 134; ParCount: 0;  Func: OpDummy ),
                                                  ( ID: ')';       Level: 0;   Grammar: 137; ParCount: 0;  Func: OpDummy ),
                                                  ( ID: '^';       Level: 70;  Grammar: 3;   ParCount: 0;  Func: OpPower ),
                                                  ( ID: '';       Level: 70;  Grammar: 2;   ParCount: 0;  Func: OpSqrt ),
                                                  ( ID: '*';       Level: 50;  Grammar: 3;   ParCount: 0;  Func: OpMultiply ),
                                                  ( ID: '/';       Level: 50;  Grammar: 3;   ParCount: 0;  Func: OpDivide ),
                                                  ( ID: '+';       Level: 40;  Grammar: 3;   ParCount: 0;  Func: OpAdd ),
                                                  ( ID: '+';       Level: 75;  Grammar: 2;   ParCount: 0;  Func: OpDummy ),
                                                  ( ID: ',';       Level: 10;  Grammar: 163; ParCount: 0;  Func: OpDummy ),
                                                  ( ID: '-';       Level: 40;  Grammar: 3;   ParCount: 0;  Func: OpSubstract ),
                                                  ( ID: '-';       Level: 75;  Grammar: 2;   ParCount: 0;  Func: OpNegate ),
                                                  ( ID: '!';       Level: 70;  Grammar: 1;   ParCount: 0;  Func: OpFac ),
                                                  ( ID: 'EXP(';    Level: 80;  Grammar: 18;  ParCount: 1;  Func: OpExp ),
                                                  ( ID: 'LN(';     Level: 80;  Grammar: 18;  ParCount: 1;  Func: OpLn ),
                                                  ( ID: 'LOG(';    Level: 80;  Grammar: 18;  ParCount: 1;  Func: OpLog ),
                                                  ( ID: 'RND';     Level: 90;  Grammar: 0;   ParCount: 0;  Func: OpRnd ),
                                                  ( ID: 'DIV';     Level: 60;  Grammar: 3;   ParCount: 0;  Func: OpDiv ),
                                                  ( ID: 'MOD';     Level: 60;  Grammar: 3;   ParCount: 0;  Func: OpMod ),
                                                  ( ID: 'INT(';    Level: 80;  Grammar: 18;  ParCount: 1;  Func: OpInt ),
                                                  ( ID: 'COS(';    Level: 80;  Grammar: 18;  ParCount: 1;  Func: OpCos ),
                                                  ( ID: 'SIN(';    Level: 80;  Grammar: 18;  ParCount: 1;  Func: OpSin ),
                                                  ( ID: 'TAN(';    Level: 80;  Grammar: 18;  ParCount: 1;  Func: OpTan ),
                                                  ( ID: 'ABS(';    Level: 80;  Grammar: 18;  ParCount: 1;  Func: OpAbs ),
                                                  ( ID: 'SUM(';    Level: 80;  Grammar: 18;  ParCount: -1; Func: OpSum ),
                                                  ( ID: 'SQRT(';   Level: 80;  Grammar: 18;  ParCount: 1;  Func: OpSqrt ),
                                                  ( ID: 'ATAN(';   Level: 80;  Grammar: 18;  ParCount: 1;  Func: OpATan ),
                                                  ( ID: 'ASIN(';   Level: 80;  Grammar: 18;  ParCount: 1;  Func: OpASin ),
                                                  ( ID: 'ACOS(';   Level: 80;  Grammar: 18;  ParCount: 1;  Func: OpACos ),
                                                  ( ID: 'XROOT(';  Level: 80;  Grammar: 18;  ParCount: 2;  Func: OpXRoot ),
                                                  ( ID: 'SQRSUM('; Level: 80;  Grammar: 18;  ParCount: -1; Func: OpSqrSum ),
                                                  ( ID: 'MEAN(';   Level: 80;  Grammar: 18;  ParCount: -1; Func: OpMean ),
                                                  ( ID: 'NELEM(';  Level: 80;  Grammar: 18;  ParCount: -1; Func: OpNelem ),
                                                  ( ID: '>';       Level: 30;  Grammar: 3;   ParCount: 0;  Func: OpBT ),
                                                  ( ID: '<';       Level: 30;  Grammar: 3;   ParCount: 0;  Func: OpST ),
                                                  ( ID: '<>';      Level: 30;  Grammar: 3;   ParCount: 0;  Func: OpNotSame ),
                                                  ( ID: '=';       Level: 30;  Grammar: 3;   ParCount: 0;  Func: OpSame ),
                                                  ( ID: '>=';      Level: 30;  Grammar: 3;   ParCount: 0;  Func: OpBTSame ),
                                                  ( ID: '<=';      Level: 30;  Grammar: 3;   ParCount: 0;  Func: OpSTSame ),
                                                  ( ID: 'OR';      Level: 20;  Grammar: 3;   ParCount: 0;  Func: OpOr ),
                                                  ( ID: 'XOR';     Level: 20;  Grammar: 3;   ParCount: 0;  Func: OpXOr ),
                                                  ( ID: 'AND';     Level: 20;  Grammar: 3;   ParCount: 0;  Func: OpAnd ),
                                                  ( ID: 'NOT';     Level: 20;  Grammar: 2;   ParCount: 0;  Func: OpNot ),
                                                  ( ID: 'IFTE(';   Level: 80;  Grammar: 18;  ParCount: 3;  Func: OpIFTE ),
                                                  ( ID: 'STRUPR('; Level: 80;  Grammar: 18;  ParCount: 1;  Func: OpStrUpr ),
                                                  ( ID: 'STRLWR('; Level: 80;  Grammar: 18;  ParCount: 1;  Func: OpStrLwr ),
                                                  ( ID: 'ISSTR('; Level: 80;  Grammar: 18;  ParCount: 1;   Func: OpIsStr ),
                                                  ( ID: 'ISBOOLE('; Level: 80;  Grammar: 18;  ParCount: 1;  Func: OpIsBoole ),
                                                  ( ID: 'ISNUMBER('; Level: 80;  Grammar: 18;  ParCount: 1; Func: OpIsNumber ),
                                                  ( ID: 'ISERROR('; Level: 80;  Grammar: 18;  ParCount: 1;  Func: OpIsError ),
                                                  ( ID: 'STRLEN('; Level: 80;  Grammar: 18;  ParCount: 1;  Func: OpStrLen ),
                                                  ( ID: 'STRPOS('; Level: 80;  Grammar: 18;  ParCount: 2;  Func: OpStrPos ),
                                                  ( ID: 'INSERTSTR('; Level: 80; Grammar: 18; ParCount: 3; Func: OpInsertStr ),
                                                  ( ID: 'SUBSTR('; Level: 80;  Grammar: 18;  ParCount: 3;  Func: OpSubStr ),
                                                  ( ID: 'DELETESTR('; Level: 80; Grammar: 18; ParCount: 3; Func: OpDeleteStr ),
                                                  ( ID: 'PARSE('; Level: 80; Grammar: 18; ParCount: 1; Func: OpParseStr ),
                                                  ( ID: 'STR('; Level: 80; Grammar: 18; ParCount: 1; Func: OpMakeStr ),
                                                  ( ID: ':=';  Level: 30;  Grammar: 3;  ParCount: 0;  Func: OpAssign ),
                                                  ( ID: ';';   Level: 0;  Grammar: 259;  ParCount: 0;  Func: OpEnd )
                                               );
{ ------------------- CnstTable --------------------------------------- }
{ All ID's in capitals !! }
                                 { Constant table definition, only kinds of numeric, booleans or strings allowed }
     CnstTable : Array[ 1 .. 7 ] Of CnstElement = ( ( ID: 'PI';     Kind: aNumber;  NumValue: 3.14159265358979E0000 ),
                                                    ( ID: '';      Kind: aNumber;  NumValue: 3.14159265358979E0000 ),
                                                    ( ID: 'E';      Kind: aNumber;  NumValue: 2.71828182845905E0000 ),
                                                    ( ID: 'ERROR';  Kind: anError;   NumValue: -44 ),
                                                    ( ID: 'TRUE';   Kind: aBoole;    NumValue: 1; BlnValue: True ),
                                                    ( ID: 'FALSE';  Kind: aBoole;    NumValue: 0; BlnValue: False ),
                                                    ( ID: 'RULES$'; Kind: aString;   NumValue: 0; BlnValue: False;
                                                       StrValue: 'OOP PARSER 6.00 By: MFSomers' )
                                                  );

{ ------------------ ErrorStrTable ------------------------------------ }

     ErrorStrTable : Array[ 0 .. 58 ] Of String[ 45 ] = ( 'No error.',
                                                          'Memory error.',
                                                          'Variable not found.',
                                                          'Bad operand.',
                                                          'Not known object in string.',
                                                          'Operand expected.',
                                                          'Operator expected.',
                                                          'Bad formula.',
                                                          'Priority level not ending zero.',
                                                          'Parameter seperator character error.',
                                                          'Not enough operands passed to function.',
                                                          'Wrong number of operands passed to function.',
                                                          'List error during evaluation.',
                                                          'Division by zero.',
                                                          'Wrong power operand.',
                                                          'Wrong logarithm operand.',
                                                          'Wrong root operand.',
                                                          'Wrong faculty operand.',
                                                          'Overflow.',
                                                          'Wrong arcsin operand.',
                                                          'Wrong arccos operand.',
                                                          'Wrong add operand.',
                                                          'Wrong substract operand.',
                                                          'Wrong multiply operand.',
                                                          'Wrong divide operand.',
                                                          'Priority level below zero.',
                                                          'Wrong modulo operand.',
                                                          'Wrong negate operand.',
                                                          'Wrong int operand.',
                                                          'Wrong absolute operand.',
                                                          'Wrong sin operand.',
                                                          'Wrong cos operand.',
                                                          'Wrong tan operand.',
                                                          'Wrong sum operand.',
                                                          'Wrong sqrsum operand.',
                                                          'Wrong mean operand.',
                                                          'Wrong xroot operand.',
                                                          'String exceeds 255 characters.',
                                                          'Not correct variable ID.',
                                                          'Wrong compare operand.',
                                                          'Wrong variable data type.',
                                                          'Wrong IFTE operand.',
                                                          'No list to evaluate.',
                                                          'Major internal error.',
                                                          'User generated error.',
                                                          'Wrong or operand.',
                                                          'Wrong xor operand.',
                                                          'Wrong and operand.',
                                                          'Wrong not operand.',
                                                          'Wrong strupr operand.',
                                                          'Wrong strlwr operand.',
                                                          'Wrong strlen operand.',
                                                          'Wrong strpos operand.',
                                                          'Wrong insertstr operand.',
                                                          'Wrong deletestr operand.',
                                                          'Wrong substr operand.',
                                                          'Wrong parse operand.',
                                                          'Wrong str operand.',
                                                          'Wrong assingment operand.'
                                                        );
     NrOfPossibleErrors : Integer = SizeOf( ErrorStrTable ) Div SizeOf( ErrorStrTable[ 0 ] );

{ ------------------ NumericChars ------------------------------------ }

     NumericChars : Set Of Char = [ '0' .. '9', '.' ];
     StringChar   : Char = '"';

{ -------------- ALL NONE OBJECT PROCEDURES WICH ARE NEEDED ------------ }

 Function Space( C : Char ) : Boolean;
  Begin
       Space := C in [ #32, #9, #10, #13 ];
  End; { Function }

{ --------------------------------------------------------------------- }

 Procedure ErrorHandler; Far;
  Begin                         { if overflow or underflow or divide by zero or illegal floatingpoint operation ? }
       If( ( ( ExitCode >= 205 ) And ( ExitCode <= 207 ) ) Or ( ExitCode = 200 ) ) Then
        Begin
             OverFlow := True;  { set overflow flag }
             __ES := ErrorCaller^._ES;
             __SS := ErrorCaller^._SS;
             __BP := ErrorCaller^._BP;
             __SP := ErrorCaller^._SP;
             __DS := ErrorCaller^._DS;
             __FLAG := ErrorCaller^._FLAG;
             __JMPSEG := ErrorCaller^._JMPSEG;
             __JMPOFS := ErrorCaller^._JMPOFS;
             ASM                { restore regs etc. }
                mov ax, [__FLAG]         { restore flags }
                push ax
                popf

                mov ds, [__DS]           { restore regs }
                mov es, [__ES]
                mov ss, [__SS]
                mov sp, [__SP]
                mov bp, [__BP]

                mov ax, [__JMPSEG]
                push ax
                mov ax, [__JMPOFS]
                push ax

                retf                    { jump back to the defined overflow label }
             End; { Asm }
        End; { If Then }

       ExitProc := ErrorCaller^.OldExit;            { else continue with the exitproc list }
  End; { Procedure }

{ ----------------------------------------------------------------------- }

 Procedure NewExitProc; Far;
  Begin
       If( _Parser_Author_Message ) Then
        Writeln( 'TURBO PASCAL OOP PARSER UNIT 6.0 By: MFSomers, 2000.' );

       ExitProc := OldExitProc;
  End; { Procedure }

{ ---------------------- TStackElement -------------------------------- }

 Constructor TStackElement.Init( K : ElementKind; V : Pointer; I, S, L : Integer );
  Begin
       TObject.Init;

       Kind := K;
       Index := I;
       Value := V;
       FormPos := S;
       ObjStrLen := L;
  End; { Constructor }

{ -------------------------------------------------------------------- }

 Destructor TStackElement.Done;
  Begin
       FreeThisItem( Kind, Value );
       TObject.Done;
  End; { Constructor }

{ -------------------------------------------------------------------- }

 Procedure TStackElement.Store( Var S : TStream );
 Var
    B : Boolean;
  Begin
       S.Write( Kind, SizeOf( Kind ) );
       S.Write( Index, SizeOf( Index ) );
       S.Write( FormPos, SizeOf( FormPos ) );
       S.Write( ObjStrLen, SizeOf( ObjStrLen ) );

       B := False;

       Case Kind Of
            aCounter,
            anOperator,
            anError,
            aNumber : S.Write( Value^, SizeOf( Extended ) );
            aBoole :   S.Write( Value^, SizeOf( Boolean ) );
            aString:  Begin
                           If( Value <> Nil ) Then
                            Begin
                                 B := True;
                                 S.Write( B, SizeOf( B ) );
                                 S.WriteStr( PString( Value ) );
                            End
                           Else
                             S.Write( B, SizeOf( B ) );
                      End; { aString }
       End; { Case }

  End; { Procedure }

{ --------------------------------------------------------------------- }

 Constructor TStackElement.Load( Var S : TStream );
 Var
    OkString : Boolean;
  Begin
       S.Read( Kind, SizeOf( Kind ) );
       S.Read( Index, SizeOf( Index ) );
       S.Read( FormPos, SizeOf( FormPos ) );
       S.Read( ObjStrLen, SizeOf( ObjStrLen ) );

       Case Kind Of
            aCounter,
            anOperator,
            anError,
            aNumber : Begin
                           Value := New( PExtended );
                           If( Value <> Nil ) Then
                            S.Read( Value^, SizeOf( Extended ) );
                      End; { anOperator }
            aBoole:    Begin
                           Value := New( PBoolean );
                           If( Value <> Nil ) Then
                            S.Read( Value^, SizeOf( Boolean ) );
                      End; { aBoole }
            aString:  Begin
                           S.Read( OkString, SizeOf( OkString ) );
                           If( Not OkString ) Then
                            Value := Nil
                           Else
                            Value := S.ReadStr;
                      End;
            aVariable: Value := Nil;         { set Value = Nil !!! }
       End;

  End; { Constructor }

{ ------------------ TVariable -------------------------------------- }

 Constructor TVariable.Init( Var S : String; K : ElementKind; V : Pointer );
  Begin
       TObject.Init;

       ID := NewStr( S );
       Kind := K;
       Value := V;
  End; { Constructor }

{ ------------------------------------------------------------------- }

 Destructor TVariable.Done;
  Begin
       DisposeStr( ID );
       FreeThisItem( Kind, Value );
       TObject.Done;
  End; { Destructor }

{ ------------------------------------------------------------------ }

 Procedure TVariable.Store( Var S : TStream );
 Var
    B : Boolean;
  Begin
       S.WriteStr( ID );
       S.Write( Kind, SizeOf( Kind ) );

       B := False;

       Case Kind Of
            aCounter,
            aNumber,
            anError,
            anOperator :  S.Write( Value^, SizeOf( Extended ) );
            aBoole :   S.Write( Value^, SizeOf( Boolean ) );
            aString:  If( Value <> Nil ) Then
                       Begin
                            B := True;
                            S.Write( B, SizeOf( B ) );
                            S.WriteStr( PString( Value ) );
                       End
                      Else
                       S.Write( B, SizeOf( B ) );
       End; { Case }
  End; { Procedure }

{ ----------------------------------------------------------------- }

 Constructor TVariable.Load( Var S : TStream );
 Var
    OkString : Boolean;
  Begin
       ID := S.ReadStr;
       S.Read( Kind, SizeOf( Kind ) );

       Case Kind Of
            aCounter,
            anOperator,
            anError,
            aNumber : Begin
                           Value := New( PExtended );
                           If( Value <> Nil ) Then
                            S.Read( Value^, SizeOf( Extended ) );
                      End; { anOperator }
            aBoole   : Begin
                           Value := New( PBoolean );
                           If( Value <> Nil ) Then
                            S.Read( Value^, SizeOf( Boolean ) );
                      End;
            aString : Begin
                           S.Read( OkString, SizeOf( OkString ) );
                           If( OkString ) Then
                            Value := S.ReadStr
                           Else
                            Value := Nil;
                      End;
            aVariable: Value := Nil;         { set Value = Nil !!! }
       End;

  End; { Constructor }

{ ---------------------- TStackParser --------------------------------- }

 Constructor TStackParser.Init;           { To init the parser }
  Begin
       TObject.Init;

       Variables    := New( PCollection, Init( Vars_Size, Vars_Size Div 2 ) );
       RList        := New( PCollection, Init( Stck_Size, Stck_Size Div 2 ) );
       EvalStack    := New( PCollection, Init( Stck_Size, Stck_Size Div 2 ) );
       HeapSaver    := New( PCollection, Init( Stck_Size, Stck_Size Div 2 ) );
       Formula      := NewStr( '0' );
       KindResult   := aNumber;
       Result       := New( PExtended );
       ReBuildList  := True;
       ReEvaluate   := True;

       If( ( Variables = Nil ) Or ( RList = Nil ) Or ( Formula = Nil ) Or
           ( Result = Nil ) Or ( EvalStack = Nil ) Or ( HeapSaver = Nil ) ) Then
        Error := Memory_Error                      { Memory error }
       Else
        Begin
             Extended( Result^ ) := 0.0;
             Error   := No_Error;
        End; { If Then Else }

  End; { Constructor }

{ -------------------------------------------------------------------- }

 Destructor TStackParser.Done;              { Called to dispose the stuff }
  Begin
       If( Variables <> Nil ) Then
        Dispose( Variables, Done );

       If( RList <> Nil ) Then
        Dispose( RList, Done );

       If( EvalStack <> Nil ) Then
        Dispose( EvalStack, Done );

       If( HeapSaver <> Nil ) Then
        Dispose( HeapSaver, Done );

       If( Formula <> Nil ) Then
        DisposeStr( Formula );

       FreeThisItem( KindResult, Result );
       TObject.Done;
  End; { Destructor }

{ --------------------------------------------------------------------- }

 Procedure TStackParser.Reset;                { Must be called after an error }
  Begin                                       { otherwise every action wil be ignored }
       If( Error = Memory_Error ) Then                  { Cant reset a memory error }
        Exit;

       FreeThisItem( KindResult, Result );

       Result := New( PExtended );

       If( Result = Nil ) Then
        Begin
             Error := Memory_Error;
             Exit;
        End; { If Then }

       Extended( Result^ ) := 0.0;
       KindResult   := aNumber;
       ReBuildList  := True;
       ReEvaluate   := True;
       Error        := No_Error;

       HeapSaver^.FreeAll;   { free the memory used so far if any }
       EvalStack^.FreeAll;
       RList^.FreeAll;
  End; { Procedure }

{ --------------------------------------------------------------------- }

 Procedure TStackParser.SetFormula( F : String );
  Begin
       If( Error <> No_Error ) Then
        Exit;

       If( Length( F ) = 0 ) Then
        Begin
             Error := -7;               { Bad formula }
             Exit;
        End; { If Then }

       If( Formula <> Nil ) Then
        DisposeStr( Formula );

       Formula := NewStr( F );

       If( Formula = Nil ) Then
        Error := Memory_Error                      { Memory error }
       Else
        Error := No_Error;

       TStackParser.Reset;              { reset status now }
  End; { Procedure }

{ --------------------------------------------------------------------- }

 Function TStackParser.GetError : Integer;
  Begin
       GetError := Error;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function TStackParser.GetErrorStr : String;
  Begin
       If( Error >= -NrOfPossibleErrors ) Then
        GetErrorStr := ErrorStrTable[ -Error ]
       Else
        GetErrorStr := 'Internal unresolved error !';
  End; { Function }

{ --------------------------------------------------------------------- }

 Function TStackParser.GetResult( Var R: Pointer; Var K : ElementKind ): Boolean;
  Begin
       GetResult := False;
       R := Result;
       K := KindResult;

       If( Error <> No_Error ) Then     { Start if no errors yet }
        Exit;

       If( RebuildList ) Then
        BuildUpList;

       If( Error <> No_Error ) Then
        Exit;

       If( ReEvaluate ) Then
        EvaluateList;

       If( Error <> No_Error ) Then
        Exit;

       GetResult := True;
       R := Result;
       K := KindResult;
  End; { Function }

{ --------------------------------------------------------------------- }

 Function TStackParser.GetFormula : String;
  Begin
       GetFormula := String( Formula^ );
  End; { Function }

{ --------------------------------------------------------------------- }

 Procedure TStackParser.SetVar( V : String; K : ElementKind; Value : Pointer );
 Var
    P : PVariable;

  Function FindIT( P : PVariable ) : Boolean; Far;
   Begin
        If( String( P^.ID^ ) =  V ) Then
         FindIT := True
        Else
         FindIT := False;
   End; { Function }

  Begin
       If( Error <> No_Error ) Then
        Exit;

       If( Not ( K In [ aNumber, aString, aBoole ] ) ) Then
        Begin
             Error := -40;
             Exit;
        End; { If Then }

       If( Length( V ) = 0 ) Then
        Begin
             Error := -38;
             Exit;
        End; { If Then }

       P := Variables^.FirstThat( @FindIT );

       If( P <> Nil ) Then        { Found one }
        Begin
             FreeThisItem( P^.Kind, P^.Value );
             P^.Value := Value;
             P^.Kind := K;
             ReEvaluate := True;
             Exit;
        End; { If Then }
                                                 { None found, create one }
       P := New( PVariable, Init( V, K, Value ) );
       If( P = Nil ) Then
        Begin
             Error := Memory_Error;
             Exit;
        End; { If Then }

       Variables^.Insert( P );  { Into the list }
       ReEvaluate := True;
       ReBuildList := True;
  End; { Procedure }

{ --------------------------------------------------------------------- }

 Procedure TStackParser.RemoveVar( V : String );
 Var
    P : PVariable;

  Function FindIT( P : PVariable ) : Boolean; Far;
   Begin
        If( String( P^.ID^ ) =  V ) Then
         FindIT := True
        Else
         FindIT := False;
   End; { Function }

  Begin
       If( Error <> No_Error ) Then
        Exit;

       If( Length( V ) = 0 ) Then
        Begin
             Error := -38;
             Exit;
        End; { If Then }

       V := StrUpr( V );

       P := Variables^.FirstThat( @FindIT );

       If( P <> Nil ) Then        { Found one }
        Begin
             Variables^.Free( P );      { Whipe it away }
             ReEvaluate := True;
             ReBuildList := True;
             Exit;
        End; { If Then }

       Error := -2;     { Variable not found }
  End; { Procedure }

{ -------------------------------------------------------------------- }

 Procedure TStackParser.RemoveAllVars;
  Begin
       If( Error <> No_Error ) Then
        Exit;

       If( Variables^.Count <> 0 ) Then
        Begin
             Variables^.FreeAll;
             ReBuildList := True;
             ReEvaluate := True;
        End; { If Then }
  End; { Procedure }

{ --------------------------------------------------------------------- }

 Procedure TStackParser.Store( Var S : TStream );
 Var
    B : Boolean;
  Begin
       S.Put( Variables );
       S.Put( RList );
       S.WriteStr( Formula );
       S.Write( Error, SizeOf( Error ) );
       S.Write( ReEvaluate, SizeOf( ReEvaluate ) );
       S.Write( ReBuildList, SizeOf( ReBuildList ) );
       S.Write( KindResult, SizeOf( KindResult ) );

       B := False;

       Case KindResult Of
            aCounter,
            aNumber,
            anError,
            anOperator : S.Write( Result^, SizeOf( Extended ) );
            aBoole :    S.Write( Result^, SizeOf( Boolean ) );
            aString:   If( Result <> Nil ) Then
                        Begin
                             B := True;
                             S.Write( B, SizeOf( B ) );
                             S.WriteStr( PString( Result ) );
                        End
                       Else
                        S.Write( B, SizeOf( B ) );
       End; { Case Of }

  End; { Procedure }

{ --------------------------------------------------------------------- }

 Constructor TStackParser.Load( Var S : TStream );
 Var
    OkString : Boolean;
  Begin
       Variables := PCollection( S.Get );
       RList := PCollection( S.Get );
       Formula := PString( S.ReadStr );
       S.Read( Error, SizeOf( Error ) );
       S.Read( ReEvaluate, SizeOf( ReEvaluate ) );
       S.Read( ReBuildList, SizeOf( ReBuildList ) );
       S.Read( KindResult, SizeOf( KindResult ) );

       OkString := True;

       Case KindResult Of
            aCounter,
            anOperator,
            anError,
            aNumber : Begin
                           Result := New( PExtended );
                           If( Result <> Nil ) Then
                            S.Read( Result^, SizeOf( Extended ) );
                      End; { anOperator }
            aBoole   : Begin
                           Result := New( PBoolean );
                           If( Result <> Nil ) Then
                            S.Read( Result^, SizeOf( Boolean ) );
                      End;
            aString : Begin
                           S.Read( OkString, SizeOf( OkString ) );
                           If( OkString ) Then
                            Result := S.ReadStr
                           Else
                            Result := Nil;
                      End;
            aVariable : Result := Nil;
       End;

       EvalStack := New( PCollection, Init( Stck_Size, Stck_Size Div 2 ) );
       HeapSaver := New( PCollection, Init( Stck_Size, Stck_Size Div 2 ) );

       If( ( Variables = Nil ) Or ( RList = Nil ) Or ( Formula = Nil ) Or
           ( ( Result = Nil ) And OkString ) Or ( EvalStack = Nil ) Or
           ( HeapSaver = Nil ) ) Then
        Error := Memory_Error;                      { Memory error }

  End; { Constructor }

{ --------------------------------------------------------------------- }

 Procedure TStackParser.BuildUpList;
 Var
    OpStack       : PCollection;              { The operator stack }
    CntStack      : PCollection;              { Counter stack }
    Priority      : Integer;                  { Current priority level }
    FPos          : Integer;                  { Pos in formula }
    FormulaLength : Integer;                  { Length of formula }
    CurrObj       : PStackElement;            { The current object handled }
    PrevObj       : PStackElement;            { The prev object }

 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }
 {$IFDEF INCDEBUG}

 Procedure DebugIt( Message : String );
 Var   X : Integer;
       S : String;
       C : Char;
  Begin
       TextBackGround( BLUE );
       TextColor( YELLOW );
       ClrScr;
       GotoXY( 22, 1 );
       Writeln( '-* PARSER DEBUGGING MODE (TRANSLATING) *-' );
       TextColor( LIGHTGRAY );
       For X := 1 To 80 Do
        Write( '-' );
       Window( 1, 3, 80, 4 );
       TextColor( YELLOW );
       Write( 'FORMULA: ' );
       TextColor( WHITE );
       TextBackGround( RED );
       If( FPos <= FormulaLength ) Then
        Begin
             For X := 1 To FPos - 1 Do
              Write( Formula^[ X ] );
             TextColor( GREEN );
             TextBackGround( BLUE );
             For X := FPos To FormulaLength Do
              Write( Formula^[ X ] );
        End { If Then }
       Else
        Write( Formula^ );
      TextColor( LIGHTGRAY );
      TextBackGround( BLUE );
      Window( 1, 1, 80, 25 );
      GotoXY( 1, 5 );
      For X := 1 To 80 Do
        Write( '-' );
      Window( 1, 6, 80, 14 );
      TextColor( YELLOW );
      Write( 'REVERSE LIST: ' );
      TextColor( WHITE );
      TextBackGround( RED );
      For X := 0 To RList^.Count - 1 Do
       Begin
        Case( PStackElement( RList^.At( X ) )^.Kind ) Of
           aVariable: Write( 'VAR(', PString( PVariable( Variables^.At( PStackElement( RList^.At( X ) )^.Index ) )^.ID )^,
                             ')' );
           aNumber:  Write( PExtended( PStackElement( RList^.At( X ) )^.Value )^ );
           aBoole:    Write( PBoolean( PStackElement( RList^.At( X ) )^.Value )^ );
           aString:   If( PStackElement( RList^.At( X ) )^.Value <> Nil ) Then
                       Write( '"', PString( PStackElement( RList^.At( X ) )^.Value )^, '"' )
                      Else
                       Write( '""' );
           anOperator:  Write( OpTable[ PStackElement( RList^.At( X ) )^.Index ].ID );
           aCounter:  Write( 'COUNT[', PStackElement( RList^.At( X ) )^.Index, ']' );
           anError:   Write( 'ERROR[', Int( PExtended( PStackElement( RList^.At( X ) )^.Value )^ ):2:0, ']' );
        End; { Case }
        Write( ' ' );
       End; { For Do }
      TextColor( LIGHTGRAY );
      TextBackGround( BLUE );
      Window( 1, 1, 80, 25 );
      GotoXY( 1, 15 );
      For X := 1 To 80 Do
        Write( '-' );
      Window( 1, 16, 80, 17 );
      TextColor( YELLOW );
      Write( 'OPERATOR STACK: ' );
      TextColor( WHITE );
      TextBackGround( RED );
      For X := 0 To OpStack^.Count - 1 Do
       Write( OpTable[ PStackElement( OpStack^.At( X ) )^.Index ].ID, ' ' );
      TextColor( LIGHTGRAY );
      TextBackGround( BLUE );
      Window( 1, 1, 80, 25 );
      GotoXY( 1, 18 );
      For X := 1 To 80 Do
       Write( '-' );
      Window( 1, 19, 80, 19 );
      TextColor( YELLOW );
      Write( 'COUNTER STACK: ' );
      TextColor( WHITE );
      TextBackGround( RED );
      For X := 0 To CntStack^.Count - 1 Do
       Write( PStackElement( CntStack^.At( X ) )^.Index, ' ' );
      TextColor( LIGHTGRAY );
      TextBackGround( BLUE );
      Window( 1, 1, 80, 25 );
      GotoXY( 1, 20 );
      For X := 1 To 80 Do
       Write( '-' );
      GotoXY( 1, 21 );
      TextColor( YELLOW );
      Write( 'GLOBAL PRIORITY: ' );
      TextBackGround( RED );
      TextColor( WHITE );
      Writeln( Priority );
      TextColor( YELLOW );
      TextBackGround( BLUE );
      Write( 'FORMPOS:         ' );
      TextBackGround( RED );
      TextColor( WHITE );
      Writeln( FPos );
      TextColor( YELLOW );
      TextBackGround( BLUE );
      Write( 'FOMULA LENGTH:   ' );
      TextBackGround( RED );
      TextColor( WHITE );
      Writeln( FormulaLength );
      GotoXY( 1, 24 );
      TextColor( LIGHTGRAY );
      TextBackGround( BLUE );
      For X := 1 To 80 Do
        Write( '-' );
      GotoXY( 1, 25 );
      TextColor( YELLOW );
      Write( 'MESSAGE: ' );
      TextColor( WHITE );
      Write( Message );
      C := ReadKey;
  End; { Procedure }

 {$ENDIF}
 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }

  Procedure CleanUp;
   Begin
         If( ( CurrObj <> Nil ) And ( RList^.IndexOf( CurrObj ) < 0 )
            And ( OpStack^.IndexOf( CurrObj ) < 0 ) And
                ( HeapSaver^.IndexOf( CurrObj ) < 0 ) ) Then
          Dispose( CurrObj, Done );

         If( OpStack <> Nil ) Then
          Dispose( OpStack, Done );

         If( CntStack <> Nil ) Then
          Dispose( CntStack, Done );

         HeapSaver^.FreeAll;
   End; { Procedure }

 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }

  Procedure CheckGrammar;          { check the parameter grammar }
  Var
     CurrGrammar : Word;
     PrevGrammar : Word;
     CurrOprtr   : Boolean;
     PrevOprtr   : Boolean;
     CurrOprnd   : Boolean;
     PrevOprnd   : Boolean;
   Begin
        CurrOprtr := False;
        PrevOprtr := False;
        CurrOprnd := False;
        PrevOprnd := False;

        If( ( CurrObj <> Nil ) And ( CurrObj^.Kind = anOperator ) ) Then
         Begin
              CurrGrammar := OpTable[ CurrObj^.Index ].Grammar;      { Define CurrGrammar }
              CurrOprtr := True;
         End; { If Then }

        If( ( CurrObj <> Nil ) And ( Not CurrOprtr ) ) Then
         CurrOprnd := True;

        If( ( PrevObj <> Nil ) And ( PrevObj^.Kind = anOperator ) ) Then
         Begin
              PrevGrammar := OpTable[ PrevObj^.Index ].Grammar;      { Define PrevGrammar }
              PrevOprtr := True;
         End; { If Then }

        If( ( PrevObj <> Nil ) And ( Not PrevOprtr ) ) Then
         PrevOprnd := True;

        If( CurrOprtr ) Then        { if curr object is operator }
         If( CurrGrammar And 1 <> 0 ) Then         { Bit 1 set, operand in front }
          Begin                                    { and no prevobj }
               If( PrevObj = Nil ) Then
                Error := -5;
                                                { if prevop followed by operand and not( prev is funct
                                                  and cur is decr priority and parcount = 0 ) }
               If( ( PrevOprtr ) And ( PrevGrammar And 2 <> 0 ) And Not
                 ( ( ( PrevGrammar And 16 <> 0 ) And ( CurrGrammar And 8 <> 0 ) ) And
                 ( OpTable[ PrevObj^.Index ].ParCount = 0 ) ) ) Then
                Error := -5;
          End { If Then }
         Else   { no operand infront of curr }
          Begin                                { if prevoprtr and not prevobj followed by operand }
               If( ( PrevOprtr ) And Not ( PrevGrammar And 2 <> 0 ) ) Then
                Error := -6;

               If( PrevOprnd ) Then
                Error := -6;
          End { If Then Else currgrammar and 1 }
        Else   { Not Curroprtr }
         Begin                        { if prev needs to be followed by operand }
              If( ( PrevOprtr ) And ( PrevGrammar And 2 <> 0 ) And ( Not CurrOprnd ) ) Then
               Error := -5;

              If( ( PrevOprtr ) And ( PrevGrammar And 2 = 0 ) And ( CurrOprnd ) ) Then
               Error := -6;

              If( CurrOprnd And PrevOprnd ) Then
               Error := -6;
         End; { If Then else CurrOprtr }

   End; { Procedure }

 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }

  Procedure GetObject;
  Label MoreString;
  Var
     TmpVal    : Extended;
     TmpInt    : Integer;
     Count     : Integer;
     ObjLngth  : Integer;
     ObjStr    : String;
     NextEntry : Boolean;
     P         : Pointer;
     K         : ElementKind;
   Begin
        TmpInt := 0;
        NextEntry := False;
        P := Nil;

        While( Space( Formula^[ FPos ] ) And ( FPos <= FormulaLength ) ) Do  { Remove all spaces }
         Inc( FPos );

        If( ( Formula^[ FPos ] = StringChar ) And ( FPos <= FormulaLength ) ) Then
         Begin
              ObjLngth := 1;
              TmpInt := FPos;
              Inc( FPos );

              MoreString:

              While( ( Formula^[ FPos ] <> StringChar ) And ( FPos <= FormulaLength ) And ( ObjLngth < 257 ) ) Do
               Begin
                    ObjStr[ ObjLngth ] := Formula^[ FPos ];
                    Inc( FPos );
                    Inc( ObjLngth );
               End; { While Do }

              If( FPos > FormulaLength ) Then
               Error := -3;

              If( ObjLngth > 256 ) Then
               Error := -37;

              If( Error <> No_Error ) Then
               Exit;

              If( ( FPos < FormulaLength ) And ( Formula^[ FPos + 1 ] = StringChar ) And ( ObjLngth < 255 ) ) Then
               Begin
                    ObjStr[ ObjLngth ] := StringChar;
                    Inc( ObjLngth );
                    Inc( FPos, 2 );
                    Goto MoreString
               End; { If Then }

              ObjStr[ 0 ] := Char( ObjLngth - 1 );

              P := NewStr( ObjStr );

              If( ( ObjLngth > 1 ) And ( P = Nil ) ) Then
               Begin
                    Error := Memory_Error;
                    Exit;
               End; { If Then }

              CurrObj := New( PStackElement, Init( aString, P, 0, TmpInt, FPos - TmpInt ) );

              If( CurrObj = Nil ) Then
               Begin
                    Error := Memory_Error;
                    DisposeStr( PString( P ) );
               End; { If Then }

              Inc( FPos );
              CheckGrammar;
              Exit;
         End; { If Then }
                                                { Check for all known variables }
        For Count := 1 To Variables^.Count Do
         Begin
              ObjStr   := TVariable( Variables^.At( Count - 1 )^ ).ID^;
              ObjLngth := Length( ObjStr );

              If( ( Copy( Formula^, FPos, ObjLngth ) = ObjStr ) And ( ObjLngth > TmpInt )  ) Then
               Begin
                    If( TmpInt <> 0 ) Then
                     Dispose( CurrObj, Done );

                    CurrObj := New( PStackElement, Init( aVariable, Nil, ( Count - 1 ), FPos, ObjLngth ) );

                    If( CurrObj = Nil ) Then
                     Begin
                          Error := Memory_Error;
                          Exit;
                     End; { If Then }

                    TmpInt := ObjLngth;
               End; { If Then }
         End; { For Do }
                                                { Check for all constants }
        For Count := 1 To ( SizeOf( CnstTable ) Div SizeOf( CnstElement ) ) Do
         Begin
              ObjStr   := CnstTable[ Count ].ID;
              ObjLngth := Length( ObjStr );

              If( ( StrUpr( Copy( Formula^, FPos, ObjLngth ) ) = ObjStr ) And ( ObjLngth > TmpInt ) ) Then
               Begin
                    If( TmpInt <> 0 ) Then
                     Dispose( CurrObj, Done );

                    Case CnstTable[ Count ].Kind Of
                         anError,
                         aNumber   : Begin
                                          P := New( PExtended );
                                          If( P <> Nil ) Then
                                           Extended( P^ ) := CnstTable[ Count ].NumValue;
                                     End; { Operand }
                         aBoole   : Begin
                                        P := New( PBoolean );
                                        If( P <> Nil ) Then
                                         Boolean( P^ ) := CnstTable[ Count ].BlnValue;
                                   End; { aBoole }
                         aString : P := NewStr( CnstTable[ Count ].StrValue );
                         aCounter,
                         anOperator,
                         aVariable : Begin
                                          Error := Major_Internal_Error;
                                          Exit;
                                     End; { all error types }
                    End; { Case }

                    If( ( P = Nil ) And Not ( ( Length( CnstTable[ Count ].StrValue ) = 0 ) And (
                                              CnstTable[ Count ].Kind = aString ) ) ) Then
                     Begin
                          Error := Memory_Error;
                          Exit;
                     End; { If Then }

                    CurrObj := New( PStackElement, Init( CnstTable[ Count ].Kind, P, Count, FPos, ObjLngth ) );

                    If( CurrObj = Nil ) Then
                     Begin
                          Error := Memory_Error;
                          Exit;
                     End; { If Then }

                    TmpInt := ObjLngth;
               End; { If Then }
         End; { For Do }
                                                { Check for all operators }
        For Count := 1 To ( SizeOf( OpTable ) Div SizeOf( OpElement ) ) Do
         Begin
              ObjStr   := OpTable[ Count ].ID;
              ObjLngth := Length( ObjStr );

              If( ( ( StrUpr( Copy( Formula^, FPos, ObjLngth ) ) = ObjStr ) And ( ObjLngth > TmpInt ) ) Or
               ( ( StrUpr( Copy( Formula^, FPos, ObjLngth ) ) = ObjStr ) And ( ObjLngth >= TmpInt ) And ( NextEntry ) ) ) Then
               Begin
                    If( TmpInt <> 0 ) Then
                     Dispose( CurrObj, Done );

                    P := New( PExtended );

                    If( P = Nil ) Then
                     Begin
                          Error := Memory_Error;
                          Exit;
                     End; { If Then }

                    Extended( P^ ) := OpTable[ Count ].Level;

                    CurrObj := New( PStackElement,
                                    Init( anOperator, P, Count, FPos, ObjLngth ) );

                    If( CurrObj = Nil ) Then
                     Begin
                          Error := Memory_Error;
                          Dispose( PExtended( P ) ) ;
                          Exit;
                     End; { If Then }

                    CheckGrammar;

                    If( Error <> No_Error ) Then       { if error search for next entry }
                     Begin
                          Error := No_Error;
                          NextEntry := True;
                     End { If Then }
                    Else
                      NextEntry := False;

                    TmpInt := ObjLngth;
               End; { If Then }

         End; { For Do }


        If( TmpInt <> 0 ) Then         { Something found ? }
         Begin
              Inc( FPos, TmpInt );      { Adjust FPos and exit }
              CheckGrammar;
              Exit;
         End; { If Then }
                                        { No, so it must be a numeric }
        Count := FPos;                        { Check it }

        While( ( Formula^[ Count ] In NumericChars ) And ( Count <= FormulaLength ) ) Do
         Inc( Count );

        If( ( Formula^[ Count ] = 'E' ) And ( Count <= FormulaLength ) ) Then
         Begin
              Inc( Count );

              If( ( ( Formula^[ Count ] = '+' ) Or ( Formula^[ Count ] = '-' ) ) And ( Count <= FormulaLength ) ) Then
               Inc( Count );

              While( ( Count <= FormulaLength ) And ( Formula^[ Count ] In NumericChars ) ) Do
               Inc( Count );
         End; { If Then }

        If( Count > FPos ) Then         { Operand found }
         Begin                          { Get value }
              Val( ( Copy( Formula^, FPos, Count - FPos ) ), TmpVal, TmpInt );

              If( TmpInt <> 0 ) Then                { Error, bad operand }
               Begin
                    Error := -3;
                    Exit;
               End; { If Then }

              P := New( PExtended );

              If( P = Nil ) Then
               Begin
                    Error := Memory_Error;
                    Exit;
               End; { If Then }

              Extended( P^ ) := TmpVal;

              CurrObj := New( PStackElement, Init( aNumber, P, 0, FPos, Count - FPos ) );

              If( CurrObj = Nil ) Then       { Error, memory full ? }
               Begin
                    Error := Memory_Error;
                    Dispose( PExtended( P ) );
                    Exit;
               End; { If Then }

              FPos := Count;    { All okay }
              CheckGrammar;       { check the grammar }
              Exit;               { Back to caller }
         End; { If Then }

         Error := -4;                   { Not known object }

   End; { Procedure }

 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }

  Procedure HandleObject;
  Var
     CntObj  : PStackElement;
     TmpObj  : PStackElement;
     P       : Pointer;
     V1, V2  : Extended;
     MustPop : Boolean;
   Begin

        If( CurrObj^.Kind <> anOperator ) Then
         RList^.Insert( CurrObj )                     { put any other obj but operator at the end of the list }
        Else
         Begin
              If( OpStack^.Count > 0 ) Then
               V1 := Extended( PStackElement( OpStack^.At( 0 ) )^.Value^ )
              Else
               V1 := 0;

              V2 := Extended( CurrObj^.Value^ ) + Priority;

              Repeat
                    MustPop := ( OpStack^.Count > 0 ) And ( V1 >= V2 );

                    If( ( PrevObj <> Nil ) And ( PrevObj^.Kind = anOperator ) ) Then
                     If( ( OpTable[ PrevObj^.Index ].Grammar And ALeftUnaryOperator <> 0 ) And
                         ( OpTable[ CurrObj^.Index ].Grammar And 3 = ALeftUnaryOperator ) ) Then
                      MustPop := False;

                    If( MustPop ) Then
                     Begin
                          RList^.Insert( OpStack^.At( 0 ) );
                          OpStack^.AtDelete( 0 );

                          If( OpStack^.Count > 0 ) Then
                           V1 := Extended( PStackElement( OpStack^.At( 0 ) )^.Value^ );
                     End; { If Then }
              Until( Not MustPop );

              Extended( CurrObj^.Value^ ) := Extended( CurrObj^.Value^ ) + Priority;      { update priority }

              If( OpTable[ CurrObj^.Index ].Grammar And 128 = 0 ) Then
               Begin
                    If( OpTable[ CurrObj^.Index ].Grammar And AnEndStatementOperator = 0 ) Then
                     OpStack^.AtInsert( 0, CurrObj )                { insert current operator in opstack }
                    Else
                     RList^.Insert( CurrObj );
               End
              Else
               HeapSaver^.Insert( CurrObj );

                                        { decrease priority and check parameter count stack }
              If( OpTable[ CurrObj^.Index ].Grammar And 8 <> 0 ) Then
               Begin
                    Dec( Priority, Op_Priority_Delta );

                    If( Priority < 0 ) Then
                     Error := -25;

                    If( CntStack^.Count > 0 ) Then
                     CntObj := PStackElement( CntStack^.At( 0 ) )       { get object }
                    Else
                     CntObj := Nil;
                                                                        { a closing bracket of a function }
                    If( ( CntObj <> Nil ) And ( Extended( CntObj^.Value^ ) = Priority ) ) Then
                     Begin
                          If( ( CntObj^.FormPos < 0 ) And ( -CntObj^.Index > CntObj^.FormPos  ) ) Then
                           Error := -10;                { not enough at least parameters }
                                                          { not right number of parameters }
                          If( ( CntObj^.FormPos > 0 ) And ( CntObj^.FormPos <> CntObj^.Index ) ) Then
                           Error := -11;

                          If( Error = No_Error ) Then
                           If( CntObj^.FormPos < 0 ) Then
                            Begin
                                 Extended( CntObj^.Value^ ) := CntObj^.Index;    { set value to index, nr parameters }
                                 RList^.Insert( CntObj );                        { add to the list }
                                 CntStack^.AtDelete( 0 );                        { remove item from cnt stack }
                                 RList^.Insert( OpStack^.At( 0 ) );        { move the var param. operator to the list }
                                 OpStack^.AtDelete( 0 );
                            End { If Then }
                           Else
                            Begin
                                 CntStack^.AtFree( 0 );                         { remove and dispose item from cnt stack }
                            End; { If Then Else }

                     End; { If Then }
               End; { If Then }

                                        { increase priority }
              If( OpTable[ CurrObj^.Index ].Grammar And 4 <> 0 ) Then
               Inc( Priority, Op_Priority_Delta );

                                        { function, start counting parameters }
              If( OpTable[ CurrObj^.Index ].Grammar And 16 <> 0 ) Then
               Begin
                    P := New( PExtended );

                    If( P = Nil ) Then
                     Begin
                          Error := Memory_Error;
                          Exit;
                     End; { If Then }

                    Extended( P^ ) := Priority;

                    CntObj := New( PStackElement, Init( aCounter, P, 1, OpTable[ CurrObj^.Index ].ParCount, 0 ) );

                    If( CntObj = Nil ) Then
                     Begin
                          Error := Memory_Error;
                          Exit;
                     End; { If Then }

                    CntStack^.AtInsert( 0, CntObj );
                    Inc( Priority, Op_Priority_Delta );
               End; { If Then }

                                        { parameter seperator char }
              If( OpTable[ CurrObj^.Index ].Grammar And 32 <> 0  ) Then
               If( CntStack^.Count <> 0 ) Then
                Inc( PStackElement( CntStack^.At( 0 ) )^.Index )
               Else
                Error := -9;

         End; { If Then Else }

   End; { Procedure }

 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }

  Begin
       If( Error <> No_Error ) Then             { If error, than quit }
        Exit;

       If( Not ReBuildList ) Then      { Don't have to rebuild it }
        Exit;

       FPos       := 1;
       Priority   := 0;
       CurrObj    := Nil;
       PrevObj    := Nil;
       FormulaLength := Length( Formula^ );

       RList^.FreeAll;                  { Remove all items in the list }

       OpStack  := New( PCollection, Init( 6, 2 ) );
       CntStack := New( PCollection, Init( 4, 2 ) );

       If( ( OpStack = Nil ) Or ( CntStack = Nil ) ) Then       { Error ? }
        Begin
             CleanUp;
             Error := Memory_Error;
             Exit;
        End; { If Then }

       While( FPos <= FormulaLength ) Do    { While not done }
        Begin
             PrevObj := CurrObj;

             GetObject;                          { Get the object, into CurrObj }

             {$IFDEF INCDEBUG}
             DebugIt( 'An object has been found...' );
             {$ENDIF}

             If( Error <> No_Error ) Then                { Something wrong ? }
              Begin
                   CleanUp;                      { CleanUp and go }
                   Exit;
              End; { If Then }

             HandleObject;                       { Do what you must with it }

             {$IFDEF INCDEBUG}
             DebugIt( 'The object has been handled...' );
             {$ENDIF}

             If( Error <> No_Error ) Then                { Anything happened ? }
              Begin
                   CleanUp;                      { CleanUp and go }
                   Exit;
              End; { If Then }
        End; { While Do }

       PrevObj := CurrObj;
       CurrObj := Nil;
       CheckGrammar;

       If( Error <> No_Error ) Then                { Anything happened ? }
        Begin
             CleanUp;                      { CleanUp and go }
             Exit;
        End; { If Then }

       While( OpStack^.Count > 0 ) Do           { dump rest of opstack into the list }
        Begin
             RList^.Insert( OpStack^.At( 0 ) );
             OpStack^.AtDelete( 0 );
        End; { While Do }

       {$IFDEF INCDEBUG}
       DebugIt( 'The operator stack has been dumped in the list and we''re done...' );
       {$ENDIF}

       If( Priority <> 0 ) Then          { not ending zero ? }
        Error := -8;

       Dispose( CntStack, Done );       { Free the stacks }
       Dispose( OpStack, Done );
       HeapSaver^.FreeAll;             { free the non disposed stuff too }

       If( Error = No_Error ) Then            { No Error => Ok result }
        ReBuildList := False;

  End; { Procedure }

{ --------------------------------------------------------------------- }

 Procedure TStackParser.EvaluateList;
 Label OverFlowLabel;                  { the error handler jump label }
 Var
    Count     : Integer;
    Obj       : PStackElement;
    TmpObj    : PStackElement;
    TmpVar    : PVariable;
    Func      : OpFunc;
    P         : Pointer;
    OkNil     : Boolean;

 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }
 {$IFDEF INCDEBUG}

 Procedure DebugIt( Message : String );
 Var X : Integer;
     C : Char;
  Begin
       TextBackGround( BLUE );
       TextColor( YELLOW );
       ClrScr;
       GotoXY( 22, 1 );
       Writeln( '-* PARSER DEBUGGING MODE (EVALUATION) *-' );
       TextColor( LIGHTGRAY );
       For X := 1 To 80 Do
        Write( '-' );
       Window( 1, 3, 80, 4 );
       TextColor( YELLOW );
       Write( 'FORMULA: ' );
       TextColor( WHITE );
       TextBackGround( BLUE );
       Write( Formula^ );
       TextColor( LIGHTGRAY );
       TextBackGround( BLUE );
       Window( 1, 1, 80, 25 );
       GotoXY( 1, 5 );
       For X := 1 To 80 Do
         Write( '-' );
       Window( 1, 6, 80, 14 );
       TextColor( YELLOW );
       Write( 'REVERSE LIST: ' );
       TextColor( WHITE );
       TextBackGround( RED );
       For X := 0 To RList^.Count - 1 Do
        Begin
         If( X > Count ) Then
          Begin
                 TextColor( GREEN );
                 TextBackGround( BLUE );
          End; { If Then }

         Case( PStackElement( RList^.At( X ) )^.Kind ) Of
            aVariable: Write( 'VAR(', PString( PVariable( Variables^.At( PStackElement( RList^.At( X ) )^.Index ) )^.ID )^,
                              ')' );
            aNumber:  Write( PExtended( PStackElement( RList^.At( X ) )^.Value )^ );
            aBoole:    Write( PBoolean( PStackElement( RList^.At( X ) )^.Value )^ );
            aString:   If( PStackElement( RList^.At( X ) )^.Value <> Nil ) Then
                        Write( '"', PString( PStackElement( RList^.At( X ) )^.Value )^, '"' )
                       Else
                        Write( '""' );
            anOperator:  Write( OpTable[ PStackElement( RList^.At( X ) )^.Index ].ID );
            aCounter:  Write( 'COUNT[', PStackElement( RList^.At( X ) )^.Index, ']' );
            anError:   Write( 'ERROR[', Int( PExtended( PStackElement( RList^.At( X ) )^.Value )^ ):2:0, ']' );
         End; { Case }
         Write( ' ' );
        End; { For Do }
       TextColor( LIGHTGRAY );
       TextBackGround( BLUE );
       Window( 1, 1, 80, 25 );
       GotoXY( 1, 15 );
       For X := 1 To 80 Do
        Write( '-' );
       TextColor( YELLOW );
       Window( 1, 16, 80, 25 );
       Write( 'EVAL STACK: ' );
       TextColor( WHITE );
       TextBackGround( RED );
       For X := 0 To EvalStack^.Count - 1 Do
        Begin
         Case( PStackElement( EvalStack^.At( X ) )^.Kind ) Of
            aVariable: Write( 'VAR(', PString( PVariable( Variables^.At( PStackElement( EvalStack^.At( X ) )^.Index ) )^.ID )^,
                              ')' );
            aNumber:  Write( PExtended( PStackElement( EvalStack^.At( X ) )^.Value )^ );
            aBoole:    Write( PBoolean( PStackElement( EvalStack^.At( X ) )^.Value )^ );
            aString:   If( PStackElement( EvalStack^.At( X ) )^.Value <> Nil ) Then
                        Write( '"', PString( PStackElement( EvalStack^.At( X ) )^.Value )^, '"' )
                       Else
                        Write( '""' );
            anOperator:  Write( OpTable[ PStackElement( EvalStack^.At( X ) )^.Index ].ID );
            aCounter:  Write( 'COUNT[', PStackElement( EvalStack^.At( X ) )^.Index, ']' );
            anError:   Write( 'ERROR[', Int( PExtended( PStackElement( EvalStack^.At( X ) )^.Value )^ ):2:0, ']' );
         End; { Case }
         Write( ' ' );
        End; { For Do }
       TextColor( LIGHTGRAY );
       TextBackGround( BLUE );
       Window( 1, 1, 80, 25 );
       GotoXY( 1, 24 );
       For X := 1 To 80 Do
         Write( '-' );
       GotoXY( 1, 25 );
       TextColor( YELLOW );
       Write( 'MESSAGE: ' );
       TextColor( WHITE );
       Write( Message );
       C := ReadKey;
  End; { Procedure }

 {$ENDIF}
 { - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }

  Begin
       If( Error <> No_Error ) Then      { Error, so quit }
        Exit;

       If( Not ReEvaluate ) Then           { Already up to date }
        Exit;
                                           { check for a present list }
       If( ( RebuildList ) Or ( RList^.Count < 1 ) ) Then
        Begin
             Error := -42;
             Exit;
        End;

       HeapSaver^.FreeAll;   { remove old shit from memory }
       EvalStack^.FreeAll;

       OldExit := ExitProc;                { set new exit proc as error handler }
       ExitProc := Addr( ErrorHandler );

       OverFlow := False;

       ASM
           mov ax, OFFSET( OverFlowLabel )  { save the regs in global vars }
           mov bx, SEG( OverFlowLabel )
           mov [__JMPOFS], ax
           mov [__JMPSEG], bx

           pushf
           pop ax
           mov [__FLAG], ax         { save flags }
           mov [__DS], ds           { save regs }
           mov [__ES], es
           mov [__SS], ss
           mov [__SP], sp
           mov [__BP], bp
       End; { ASM }

       _ES     := __ES;            { transfer saved regs into current parser object }
       _DS     := __DS;
       _SS     := __SS;
       _SP     := __SP;
       _BP     := __BP;
       _FLAG   := __FLAG;
       _JMPOFS := __JMPOFS;
       _JMPSEG := __JMPSEG;

OverFlowLabel:      { to this adress will be jumped if the errorhandler exit proc }
                    { encounters an overflow etc. the regs of importance are restored then. }
                    { the stack frame is still intact upon entry by errorhandler ! }

       If( OverFLow ) Then              { Check on overflow entry by errorhandler }
        Begin
             HeapSaver^.FreeAll;       { release all stuff from the heapsaver }
             EvalStack^.FreeAll;
             ExitProc := OldExit;      { remove error handler from the exitproc list }
             Error := -18;             { Overflow error }
             Exit;                     { don't continue }
        End; { If Then }

       ReEvaluate := False;

       For Count := 0 To RList^.Count - 1 Do
        Begin
             Obj := PStackElement( RList^.At( Count ) );

             {$IFDEF INCDEBUG}
             DebugIt( 'Stack element has been obtained...' );
             {$ENDIF}

             OkNil := False;

             If( Obj^.Kind = anOperator ) Then
              Begin
                  Func := OpTable[ Obj^.Index ].Func;           { call the right proc }
                  ErrorCaller := PStackParser( @Self );
                  Error := Func( PStackParser( @Self ) );       { whenever there occurs a runtime error like an overflow }
                                                                { the exitproc just installed jumps to the label and there }
                                                                { will be made a check to the flag set by errorhandler }
                  If( Error < Memory_Error ) Then      { if error returned by func, push it into stack }
                   Begin                               { and continue evaluation because formula can still }
                        P := New( PExtended );         { be a valid non error formula !! }
                         If( P = Nil ) Then
                         Error := Memory_Error
                        Else
                         Begin
                              Extended( P^ ) := Error;
                              TmpObj := New( PStackElement, Init( anError, P, Obj^.Index, Obj^.FormPos, Obj^.ObjStrLen ) );
                               If( TmpObj = Nil ) Then
                               Error := Memory_Error
                              Else
                               Begin
                                    Error := No_Error;
                                    EvalStack^.AtInsert( 0, TmpObj );
                               End; { If Then Else }
                         End; { If Then Else }
                   End; { If Then }
                   {$IFDEF INCDEBUG}
                   DebugIt( 'Operator runtime routine has been called...' );
                   {$ENDIF}
              End
             Else
              Begin                   { every new created obj will be in a list, the eval stack or the heapsaver }
                   Case Obj^.Kind Of
                        aCounter,
                        anError,
                        aNumber  : Begin
                                        P := New( PExtended );
                                        If( P <> Nil ) Then
                                         Extended( P^ ) := Extended( Obj^.Value^ );
                                   End; { Case aCounter }
                        aBoole   : Begin
                                        P := New( PBoolean );
                                        If( P <> Nil ) Then
                                         Boolean( P^ ) := Boolean( Obj^.Value^ );
                                   End; { Case aBoole }
                        aString  : Begin
                                        If( Obj^.Value <> Nil ) Then
                                         P := NewStr( PString( Obj^.Value )^ )
                                        Else
                                         Begin
                                              P := Nil;
                                              OkNil := True;
                                         End;
                                   End; { aString }
                        aVariable: Begin
                                        TmpVar := PVariable( Variables^.At( Obj^.Index ) );   { reload with current value }
                                        Case ( TmpVar^.Kind ) Of
                                             aNumber:  Begin
                                                             P := New( PExtended );
                                                             If( P <> Nil ) Then
                                                              Extended( P^ ) := Extended( TmpVar^.Value^ );
                                                        End;
                                             aBoole  :  Begin
                                                             P := New( PBoolean );
                                                              If( P <> Nil ) Then
                                                               Boolean( P^ ) := Boolean( TmpVar^.Value^ );
                                                        End;
                                             aString :  Begin
                                                             If( TmpVar^.Value <> Nil ) Then
                                                              P := NewStr( PString( TmpVar^.Value )^ )
                                                             Else
                                                              Begin
                                                                   P := Nil;
                                                                   OkNil := True;
                                                              End;
                                                        End; { aString }
                                        Else
                                            Begin
                                                 Error := Major_Internal_Error;
                                                 P := Nil;
                                                 OkNil := True;
                                            End; { all error types }
                                        End; { Case }
                                   End; { Case of variable }
                   End; { Case }

                   If( ( P = Nil ) And ( Not OkNil ) ) Then
                    Error := Memory_Error;

                   If( Error = No_Error ) Then
                    Begin
                         TmpObj := New( PStackElement, Init( Obj^.Kind, P, Obj^.Index, Obj^.FormPos, Obj^.ObjStrLen ) );

                         If( TmpObj = Nil ) Then
                          Error := Memory_Error
                         Else
                          If( Obj^.Kind = aVariable ) Then
                           Begin
                                TmpObj^.Index := -( Obj^.Index + 1 );        { signal that it was derived from a var }
                                TmpObj^.Kind := PVariable( TmpVar )^.Kind;
                           End; { If Then }
                    End; { If Then Else }

                   If( Error = No_Error ) Then
                    EvalStack^.AtInsert( 0, TmpObj );

                   {$IFDEF INCDEBUG}
                   DebugIt( 'Stack element has been handled...' );
                   {$ENDIF}
              End; { If Then Else }

             If( Error <> No_Error ) Then
              Begin
                   ExitProc := OldExit;         { remove errorhandler }
                   HeapSaver^.FreeAll;          { remove all obj from heap saver }
                   EvalStack^.FreeAll;
                   Exit;
              End;  { If Then }
        End; { For Do }

       If( EvalStack^.Count = 1 ) Then                       { Get result }
        Begin
            TmpObj := PStackElement( EvalStack^.At( 0 ) );

            If( Not( TmpObj^.Kind In [ aString, aNumber, aBoole ] ) ) Then
             Begin
                  If( TmpObj^.Kind = anError ) Then
                   Error := Round( Extended( TmpObj^.Value^ ) )
                  Else
                   Error := Major_Internal_Error;            { generate programmers error !! }
                  EvalStack^.FreeAll;
                  HeapSaver^.FreeAll;     { remove all none disposed obj from heapsaver }
                  ExitProc := OldExit;    { reset error handler and quit }
                  Exit;
             End; { If Then }

             FreeThisItem( KindResult, Result );
             KindResult := TmpObj^.Kind;

             OkNil := False;

             Case KindResult Of
                  anError,
                  aNumber   : Begin
                                   Result := New( PExtended );
                                   If( Result <> Nil ) Then
                                    Extended( Result^ ) := Extended( TmpObj^.Value^ );
                              End; { Case aCounter }
                  aBoole     : Begin
                                   Result := New( PBoolean );
                                   If( Result <> Nil ) Then
                                    Boolean( Result^ ) := Boolean( TmpObj^.Value^ );
                              End; { Case aCounter }
                  aString    : Begin
                                    P := TmpObj^.Value;
                                    If( P <> Nil ) Then
                                     Result := NewStr( PString( P )^ )
                                    Else
                                     Begin
                                          OkNil := True;
                                          Result := Nil;
                                     End;
                               End; { Case aString }
             End; { Case }

            If( ( Result = Nil ) And ( Not OkNil ) ) Then
             Error := Memory_Error;
        End { If Then }
       Else
        If( EvalStack^.Count > 1 ) Then       { then an error must be present at the top }
         Begin
              TmpObj := PStackElement( EvalStack^.At( 0 ) );   { get the object }

              If( TmpObj^.Kind <> anError ) Then
               Error := List_Eval_Error
              Else
               Error := Round( PExtended( TmpObj^.Value )^ );
         End
        Else
         Error := List_Eval_Error;       { if nothing left bad evaluation occured !!! }

       EvalStack^.FreeAll;
       HeapSaver^.FreeAll;    { remove all none disposed obj from heapsaver }

       {$IFDEF INCDEBUG}
       DebugIt( 'Result has been obtained and we''re done...' );
       {$ENDIF}

       ExitProc := OldExit;
  End; { Procedure }

{ --------------------------------------------------------------------- }

Procedure RegisterParser;
 Begin
      RegisterType( RVariable );
      RegisterType( RStackElement );
      RegisterType( RStackParser );
 End; { Procedure }

{ --------------------------------------------------------------------- }

 Begin
      Randomize;
      _Parser_Author_Message := True;
      OldExitProc := ExitProc;
      ExitProc := Addr( NewExitProc );
 End. { Unit }