UNIT TrapInt;

(**********************************************************************

 Intercept Interrupts, provide huge stack for the interrupt procedure.

 (c) Copyright 1996 by Klaus Hartnegg,
     H.v.Kleist-street 7, D-79331 Teningen, Germany
     hartnegg@iname.com
     http://home.pages.de/~hartnegg/

 Explanation
   Borland's Turbo Pascal and Borland Pascal 7 allow to declare
   Interrupt procedures. However these procedures do have a big
   drawback: they must work with the stack that happens to be
   active in the moment the procedure is called. This stack may
   be as tiny as 16 bytes. But for example passing a string to
   another pascal procedures requires 256 bytes. Passing 3 real
   numbers requires 15 bytes.

 Limitation
   Only one interrupt may be hooked at a time.

**********************************************************************)


INTERFACE

uses
 dos;

type
  TrapProcedure = procedure (var reg:registers; var abort:boolean);

Procedure TrapInterrupt (_IntNr:byte; _IntProc:TrapProcedure);
Procedure ReleaseInt;


IMPLEMENTATION

var
  IntNr   : byte;
  IntProc : TrapProcedure;

var
  OldInt  : pointer;
  OldExit : pointer;
  SSSave, SPSave,
  CallIP, CallCS  : word;
  active : byte;
  LostIntCounter : word;
  abort  : boolean;
  PReg   : registers;

{ stack for interrupt handler }
const IntrStackSize = 16384;
var   IntrStack : array [1..IntrStackSize] of byte;
const IntrSS : word = Seg(IntrStack[IntrStackSize]);
      IntrSP : word = Ofs(IntrStack[IntrStackSize]);


Procedure IntHandler; FAR; Assembler;
asm
  cmp   byte ptr @firstcall, 0 { check if initialization call }
  je    @install

  push  ds
  push  ax

  mov   ax, seg @data          { set data segment of turbo pascal }
  mov   ds,ax

  cmp   byte ptr active,0      { test activity flag }
  jne   @notreentrant

  inc   byte ptr active        { set activity flag        }

  pop   [PReg.&ax]             { get AX and DS from stack }
  pop   [PReg.&ds]             { and store them           }

  pop   [CallIP]               { store stack              }
  pop   [CallCS]
  pop   [PReg.Flags]

  mov   [PReg.&bx], bx         { store register contents  }
  mov   [PReg.&cx], cx
  mov   [PReg.&dx], dx
  mov   [PReg.&bp], bp
  mov   [PReg.&si], si
  mov   [PReg.&di], di
  mov   [PReg.&es], es

  cli
  mov   [SSSave], ss           { store old stack pointer    }
  mov   [SPSave], sp
  mov   ss,[IntrSS]            { set stack for turbo pascal }
  mov   sp,[IntrSP]
  sti

  mov   byte ptr abort, 0

  mov   di,offset preg
  push  ds
  push  di
  mov   di,offset abort
  push  ds
  push  di
  call  dword ptr IntProc      { call pascal procedure }

  cli
  mov   ss,[SSSave]            { old stack back  }
  mov   sp,[SPSave]
  sti

  mov   es,[PReg.&es]          { old registers back }
  mov   bp,[PReg.&bp]
  mov   di,[PReg.&di]
  mov   si,[PReg.&si]
  mov   dx,[PReg.&dx]
  mov   cx,[PReg.&cx]
  mov   bx,[PReg.&bx]

  push  [PReg.Flags]
  push  [CallCS]
  push  [CallIP]
  mov   ax,[PReg.&ax]

  dec   byte ptr active        { clear activity flag }

  cmp   byte ptr abort,0       { test abort flag     }
  jne   @abort


@original:
  push  [PReg.&ds]             { restore DS }
  pop   ds
  jmp   dword ptr cs:[@IntAddr]


@abort:
  push  [PReg.&ds]             { restore DS }
  pop   ds
  iret


@notreentrant:
  inc   word ptr LostIntCounter
  pop   ax
  pop   ds
  jmp   dword ptr cs:[@IntAddr]


@install:   { write address of original handler into code segment }
  inc   byte ptr @firstcall

  mov   ax, word ptr OldInt
  mov   word ptr @intaddr, ax
  mov   ax, word ptr OldInt+2
  mov   word ptr @intaddr+2, ax

  retf

@firstcall: db 0

@intaddr:   dw 0
            dw 0

end;



Procedure TrapInterrupt (_IntNr:byte; _IntProc:TrapProcedure);
begin
  IntNr   := _IntNr;
  IntProc := _IntProc;

  LostIntCounter := 0;

  GetIntVec (IntNr, OldInt);
  IntHandler; { first call does initialization }
  writeln ('init ok');

  OldExit := ExitProc;
  ExitProc := @ReleaseInt;
  SetIntVec (IntNr, @IntHandler);
end;


Procedure ReleaseInt;
begin
  ExitProc := OldExit;
  SetIntVec (IntNr, OldInt);
end;


END.

