Unit Modey;

Interface

{$G+}

Var
 Page:word;

Const
 MaxX=319;
 MaxY=199;

 NextLine=80;

Procedure GraphMode;
Procedure TextMode;

Procedure Clear;
Procedure Swap;

Procedure SetRGB(c,r,g,b:word);
Procedure SetPalette(Palette:pointer);

Procedure InitZBuffers;
{Procedure PutZPixel(x,y,z,c:integer);}

Procedure InitZGouraud(x,y,z,c:integer);
procedure ScanZGouraud(x,y,z,c:integer);
Procedure DrawZGouraud;

Procedure InitZPoly(x,y,z:integer);
Procedure ScanZPoly(x,y,z:integer);
Procedure DrawZPoly(c:integer);


Implementation

{.$define shift}

Const
{$ifdef shift}
 ColorShift=7;
 HalfColor=1 shl (ColorShift-1);
{$endif}

 PageSize=320*200 div 4;
 LastPage=3*PageSize;

Type
 pZBuffer=^TZBuffer;
 TZBuffer=array[0..MaxY] of integer;
 TZBuffers=array[0..MaxX] of pZBuffer;

Var
 ScanMin:array[0..MaxX] of integer;
 ScanMax:array[0..MaxX] of integer;
 ScanLeft:word;
 ScanRight:word;
 ScanX,ScanY:integer;

 ColorMin:array[0..MaxX] of word;
 ColorMax:array[0..MaxX] of word;
 ScanC:word;

 ScanMinZ:array[0..MaxX] of integer;
 ScanMaxZ:array[0..MaxX] of integer;
 ScanZ:integer;

 ZBuffers:TZBuffers;
 ZBuffer :pZBuffer;

function max(a,b:integer):integer; assembler;
 asm
  mov ax,a
  cmp ax,b
  jae @done
   mov ax,b
  @done:
 end;

Procedure GraphMode; assembler;
 asm
  mov ax,13h
  int 10h
{ Extended Mode X }
  mov dx,$3C4    { port[$3C4]:=4 }
  mov al,4
  out dx,al
  inc dx         { port[$3C5]:=(port[$3C5] and (not 8)) or 4 }
  in  al,dx
  and al,not 8
  or  al,4
  out dx,al

  mov dx,$3D4    { port[$3D4]:=$14 }
  mov al,$14
  out dx,al
  inc dx         { port[$3D5]:=(port[$3D5] and (not 64) }
  in  al,dx
  and al,not 64
  out dx,al

  dec dx         { port[$3D4]:=$17 }
  mov al,$17
  out dx,al
  inc dx
  in  al,dx      { port[$3D5]:=port[$3D5] or 64 }
  or  al,64
  out dx,al

{ clear screen }

  mov ax,$0F02    { port[$3C4]:=2 }
  mov dx,$3C4
  out dx,ax       { port[$3C5]:=1+2+3+4 }

  mov es,SegA000  { FillChar(Mem[$A000:0],(80*200),0) }
  xor di,di
  mov cx,NextLine*200*4
  mov ax,0
  rep stosb

{ init swap page }

  mov Page,0 { PageSize }
 end;

Procedure TextMode; assembler;
 asm
  mov ax,03h
  int 10h
 end;

procedure Clear; assembler;
 asm
  mov ax,$0F02    { port[$3C4]:=2 }
  mov dx,$3C4
  out dx,ax       { port[$3C5]:=1+2+3+4 }

  mov es,SegA000
  mov di,Page
  mov cx,NextLine*200
  mov ax,0
  rep stosb
 end;

procedure Swap; assembler;
 asm
  mov bx,Page
  mov dx,$3D4
  mov al,$C
  mov ah,bh
  out dx,ax
  inc ax
  mov ah,bl
  out dx,ax

  mov ax,Page
  sub ax,PageSize
  jns @ok
   mov ax,LastPage
  @ok:
  mov Page,ax
 end;

Procedure SetRGB(c,r,g,b:word); assembler;
 asm
  mov dx,$3C8
  mov ax,c
  out dx,al
  inc dx
  mov ax,r
  out dx,al
  mov ax,g
  out dx,al
  mov ax,b
  out dx,al
 end;

Procedure SetPalette(Palette:pointer); assembler;
 asm
 push ds
  mov dx,$3C8
  mov al,0
  out dx,al
  inc dx
  lds si,Palette
  mov cx,3*256
  rep outsb
 pop  ds
 end;


Procedure InitZBuffers; assembler;
 asm
  mov ax,32767 { Big Z }
  mov bx,offset ZBuffers
  mov dx,MaxX+1
  @nextbuffer:
  mov di,[bx]; add bx,2
  mov es,[bx]; add bx,2
  mov cx,MaxY+1
  rep stosw
  dec dx
  jnz @nextbuffer
 end;

var
 video:word;

Procedure PutZPixel({x,}y,z,c:integer);
 begin
  if ZBuffer^[y]>z then begin
   ZBuffer^[y]:=z;
   Mem[$A000:Video]:=c{$ifdef shift}shr ColorShift{$endif};
  end;
 end;


Procedure InitZGouraud(x,y,z,c:integer); assembler;
 asm
  mov ax,ds
  mov es,ax

  mov di,offset ScanMin
  mov ax,MaxY+1
  mov cx,MaxX+1
  rep stosw

  mov di,offset ScanMax
  mov ax,0
  mov cx,MaxX+1
  rep stosw

  mov ax,x
  mov bx,y
  mov cx,c
  mov dx,z

 {$ifdef shift}
  shl cx,ColorShift
  add cx,HalfColor
 {$endif}

  mov ScanX,ax
  mov ScanY,bx
  mov ScanC,cx
  mov ScanZ,dx

  mov ScanLeft,MaxX+1
  mov ScanRight,0
 end;

Procedure ScanZGouraud(x,y,z,c:integer);
 var
  dx,dy,dc,dz:integer;
  ix,iy,ic,iz:integer;
  ex,ey,ec,ez:integer;
  ee,ii:integer;
  ValidX:boolean;
 begin
 {$ifdef shift}
  c:=c shl ColorShift;
 {$endif}
  dx:=x-ScanX; if dx<0 then begin dx:=-dx; ix:=-1 end else ix:=1;
  dy:=y-ScanY; if dy<0 then begin dy:=-dy; iy:=-1 end else iy:=1;
  dc:=c-ScanC; if dc<0 then begin dc:=-dc; ic:=-1 end else ic:=1;
  dz:=z-ScanZ; if dz<0 then begin dz:=-dz; iz:=-1 end else iz:=1;

  ex:=dx;
  ey:=dy;
  ec:=dc;
  ez:=dz;

  ee:=max(ex,max(ey,max(ec,ez)));

  if ScanX<ScanLeft then if ScanX<0 then ScanLeft:=0 else ScanLeft:=ScanX;
  if ScanX>ScanRight then if ScanX<=MaxX then ScanRight:=ScanX else ScanRight:=MaxX;

  ValidX:=(ScanX>=0)and(ScanX<=MaxX);

  for ii:=0 to ee do begin

   if ValidX then begin

    if ScanMin[ScanX]>ScanY then begin
      ScanMin[ScanX]:=ScanY;
     ColorMin[ScanX]:=ScanC;
     ScanMinZ[ScanX]:=ScanZ;
    end;

    if ScanMax[ScanX]<ScanY then begin
      ScanMax[ScanX]:=ScanY;
     ColorMax[ScanX]:=ScanC;
     ScanMaxZ[ScanX]:=ScanZ;
    end;

   end;

   if ey>ee then begin
    dec(ey,ee);
    inc(ScanY,iy);
   end;
   inc(ey,dy);

   if ec>ee then begin
    dec(ec,ee);
    inc(ScanC,ic);
   end;
   inc(ec,dc);

   if ex>ee then begin
    dec(ex,ee);
    inc(ScanX,ix);
    ValidX:=(ScanX>=0)and(ScanX<=MaxX);
   end;
   inc(ex,dx);

   if ez>ee then begin
    dec(ez,ee);
    inc(ScanZ,iz);
   end;
   inc(ez,dz);

  end;
 end;

Procedure DrawZGouraud;
 var
  x:word;
  y1,y2:integer;
  c1,c2:integer;
  z1,z2:integer;
  dc,ic,ec:integer;
  dz,iz,ez:integer;
  dy,i ,ey:integer;
  f1,e,f2:integer;
  ee,ii:integer;
 begin
  For x:=ScanLeft to ScanRight do begin
   y1:=ScanMin[x];
   y2:=ScanMax[x];
   if y1<=y2 then begin

    Port[$3c4]:=2;
    Port[$3c5]:=1 shl (x and 3);

    Video:=Page+NextLine*y1+x shr 2;

    ZBuffer:=ZBuffers[x];

    c1:=ColorMin[x];
    c2:=ColorMax[x];
    z1:=ScanMinZ[x];
    z2:=ScanMaxZ[x];

    dy:=y2-y1+1;
    dc:=c2-c1+1; if dc<0 then begin dc:=-dc; ic:=-1 end else ic:=+1;
    dz:=z2-z1+1; if dz<0 then begin dz:=-dz; iz:=-1 end else iz:=+1;

    ey:=dy;
    ec:=dc;
    ez:=dz;

    ee:=max(ey,max(ec,ez));

    for ii:=0 to ee do begin

     if ec>ee then begin
      dec(ec,ee);
      inc(c1,ic);
     end;
     inc(ec,dc);

     if ey>ee then begin
      if (y1>=0)and(y1<=MaxY) then PutZPixel(y1,z1,c1);
      inc(video,NextLine);
      dec(ey,ee);
      inc(y1);
     end;
     inc(ey,dy);

     if ez>ee then begin
      dec(ez,ee);
      inc(z1,iz);
     end;
     inc(ez,dz);

    end;
   end;
  end;

 end;

Procedure InitZPoly(x,y,z:integer); assembler;
 asm
  mov ax,ds
  mov es,ax

  mov di,offset ScanMin
  mov ax,MaxY+1
  mov cx,MaxX+1
  rep stosw

  mov di,offset ScanMax
  mov ax,0
  mov cx,MaxX+1
  rep stosw

  mov ax,x
  mov bx,y
  mov dx,z

  mov ScanX,ax
  mov ScanY,bx
  mov ScanZ,dx

  mov ScanLeft,MaxX+1
  mov ScanRight,0
 end;

Procedure ScanZPoly(x,y,z:integer);
 var
  dx,dy,dz:integer;
  ix,iy,iz:integer;
  ex,ey,ez:integer;
  ee,ii:integer;
  ValidX:boolean;
 begin
  dx:=x-ScanX; if dx<0 then begin dx:=-dx; ix:=-1 end else ix:=1;
  dy:=y-ScanY; if dy<0 then begin dy:=-dy; iy:=-1 end else iy:=1;
  dz:=z-ScanZ; if dz<0 then begin dz:=-dz; iz:=-1 end else iz:=1;

  ex:=dx;
  ey:=dy;
  ez:=dz;

  ee:=max(ex,max(ey,ez));

  if ScanX<ScanLeft then if ScanX<0 then ScanLeft:=0 else ScanLeft:=ScanX;
  if ScanX>ScanRight then if ScanX<=MaxX then ScanRight:=ScanX else ScanRight:=MaxX;

  ValidX:=(ScanX>=0)and(ScanX<=MaxX);

  for ii:=0 to ee do begin

   if ValidX then begin

    if ScanMin[ScanX]>ScanY then begin
      ScanMin[ScanX]:=ScanY;
     ScanMinZ[ScanX]:=ScanZ;
    end;

    if ScanMax[ScanX]<ScanY then begin
      ScanMax[ScanX]:=ScanY;
     ScanMaxZ[ScanX]:=ScanZ;
    end;

   end;

   if ey>ee then begin
    dec(ey,ee);
    inc(ScanY,iy);
   end;
   inc(ey,dy);

   if ex>ee then begin
    dec(ex,ee);
    inc(ScanX,ix);
    ValidX:=(ScanX>=0)and(ScanX<=MaxX);
   end;
   inc(ex,dx);

   if ez>ee then begin
    dec(ez,ee);
    inc(ScanZ,iz);
   end;
   inc(ez,dz);

  end;
 end;

Procedure DrawZPoly(c:integer);
 var
  x:word;
  y1,y2:integer;
  z1,z2:integer;
  dz,iz,ez:integer;
  dy,i ,ey:integer;
  f1,e,f2:integer;
  ee,ii:integer;
 begin
  For x:=ScanLeft to ScanRight do begin
   y1:=ScanMin[x]; if y1<0 then y1:=0;
   y2:=ScanMax[x]; if y2>MaxY then y2:=MaxY;
   if y1<=y2 then begin

    Port[$3c4]:=2;
    Port[$3c5]:=1 shl (x and 3);

    Video:=Page+NextLine*y1+x shr 2;

    ZBuffer:=ZBuffers[x];

    z1:=ScanMinZ[x];
    z2:=ScanMaxZ[x];

    dy:=y2-y1+1;
    dz:=z2-z1+1; if dz<0 then begin dz:=-dz; iz:=-1 end else iz:=+1;

    ey:=dy;
    ez:=dz;

    ee:=max(ey,ez);

    for ii:=0 to ee do begin

     if ey>ee then begin
      PutZPixel(y1,z1,c);
      inc(video,NextLine);
      dec(ey,ee);
      inc(y1);
     end;
     inc(ey,dy);

     if ez>ee then begin
      dec(ez,ee);
      inc(z1,iz);
     end;
     inc(ez,dz);

    end;
   end;
  end;

 end;

var
 i:integer;
begin
 for i:=0 to MaxX do new(ZBuffers[i]);
end.
